/*
 * Copyright (C) 2009, Martin Robinson
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"
#include "DataObjectClutter.h"

#include "markup.h"
#include <gtk/gtk.h>
#include <wtf/gobject/GOwnPtr.h>
#include <wtf/text/StringBuilder.h>

namespace WebCore {

static String replaceNonBreakingSpaceWithSpace(String str)
{
    static const UChar NonBreakingSpaceCharacter = 0xA0;
    static const UChar SpaceCharacter = ' ';
    str.replace(NonBreakingSpaceCharacter, SpaceCharacter);
    return str;
}

ListHashSet<String> DataObjectClutter::types() const
{
    ListHashSet<String> results;

    for (HashMap<String,String>::const_iterator::Keys it = m_data.begin().keys(), end = m_data.end().keys(); it != end; ++it) {
        const String type = *it;
        results.add(type);
        if (type == "text/plain") {
            results.add("text");
            results.add("Text");
        }
        if (type == "text/uri-list")
            results.add("URL");
    }

    if (hasFilenames())
        results.add("Files");

    return results;
}

String DataObjectClutter::data(const String& type) const
{
    return m_data.get(type);
}

String DataObjectClutter::text() const
{
    if (m_range)
        return m_range->text();
    return m_data.get("text/plain");
}

String DataObjectClutter::markup() const
{
    if (m_range)
        return createMarkup(m_range.get(), 0, AnnotateForInterchange, false, ResolveNonLocalURLs);
    return m_data.get("text/html");
}

void DataObjectClutter::setData(const String& type, const String& data)
{
    m_data.set(type, data);
}

void DataObjectClutter::setText(const String& newText)
{
    m_range = 0;
    m_data.set("text/plain", replaceNonBreakingSpaceWithSpace(newText));
}

void DataObjectClutter::setMarkup(const String& newMarkup)
{
    m_range = 0;
    m_data.set("text/html", newMarkup);
}

void DataObjectClutter::setURIList(const String& uriListString)
{
    m_data.set("text/uri-list", uriListString);

    // This code is originally from: platform/chromium/ChromiumDataObject.cpp.
    // FIXME: We should make this code cross-platform eventually.

    // Line separator is \r\n per RFC 2483 - however, for compatibility
    // reasons we also allow just \n here.
    Vector<String> uriList;
    uriListString.split('\n', uriList);

    // Process the input and copy the first valid URL into the url member.
    // In case no URLs can be found, subsequent calls to getData("URL")
    // will get an empty string. This is in line with the HTML5 spec (see
    // "The DragEvent and DataTransfer interfaces"). Also extract all filenames
    // from the URI list.
    bool setURL = false;
    for (size_t i = 0; i < uriList.size(); ++i) {
        String& line = uriList[i];
        line = line.stripWhiteSpace();
        if (line.isEmpty())
            continue;
        if (line[0] == '#')
            continue;

        KURL url = KURL(KURL(), line);
        if (url.isValid()) {
            if (!setURL) {
                m_url = url;
                setURL = true;
            }

            GOwnPtr<GError> error;
            GOwnPtr<gchar> filename(g_filename_from_uri(line.utf8().data(), 0, &error.outPtr()));
            if (!error && filename)
                m_filenames.append(String::fromUTF8(filename.get()));
        }
    }
}

void DataObjectClutter::setURL(const KURL& url, const String& label)
{
    m_url = url;
    m_data.set("text/uri-list", url.string());

    setText(url.string());

    String actualLabel(label);
    if (actualLabel.isEmpty())
        actualLabel = url;

    StringBuilder markup;
    markup.append("<a href=\"");
    markup.append(url.string());
    markup.append("\">");
    GOwnPtr<gchar> escaped(g_markup_escape_text(actualLabel.utf8().data(), -1));
    markup.append(String::fromUTF8(escaped.get()));
    markup.append("</a>");
    setMarkup(markup.toString());
}

void DataObjectClutter::clearText()
{
    m_range = 0;
    m_data.remove("text/plain");
}

void DataObjectClutter::clearMarkup()
{
    m_range = 0;
    m_data.remove("text/html");
}

String DataObjectClutter::urlLabel() const
{
    if (hasText())
        return text();

    if (hasURL())
        return url();

    return String();
}

void DataObjectClutter::clearAllExceptFilenames()
{
    m_data.clear();
    m_range = 0;
}

void DataObjectClutter::clearAll()
{
    clearAllExceptFilenames();
    m_filenames.clear();
}

}
