/*
 * Copyright (C) 2012 Collabora, Ltd.  All rights reserved.
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "PlatformTouchEvent.h"
#include "PlatformTouchPoint.h"
#include "UtilClutter.h"

#include <clutter/clutter.h>
#include <wtf/CurrentTime.h>
#include <wtf/HashMap.h>

#if ENABLE(TOUCH_EVENTS)

namespace WebCore {

// TODO: Share this with KeyEventClutter.cpp
static PlatformEvent::Modifiers modifiersForClutterTouchEvent(const ClutterTouchEvent *event)
{
    unsigned int modifiers = 0;
    if (event->modifier_state & CLUTTER_SHIFT_MASK)
        modifiers |= PlatformEvent::ShiftKey;
    if (event->modifier_state & CLUTTER_CONTROL_MASK)
        modifiers |= PlatformEvent::CtrlKey;
    if (event->modifier_state & CLUTTER_MOD1_MASK)
        modifiers |= PlatformEvent::AltKey;
    if (event->modifier_state & CLUTTER_META_MASK)
        modifiers |= PlatformEvent::MetaKey;
    return static_cast<PlatformEvent::Modifiers>(modifiers);
}

class ClutterTouchEventDetails {
    WTF_MAKE_NONCOPYABLE(ClutterTouchEventDetails);
public:
    ClutterTouchEventDetails(ClutterActor *actor, const ClutterTouchEvent *event)
    {
        m_actor = (ClutterActor*)g_object_ref(actor);
        m_event = (const ClutterTouchEvent*)clutter_event_copy((const ClutterEvent*)event);
    }
    ~ClutterTouchEventDetails()
    {
        g_object_unref(m_actor);
        clutter_event_free((ClutterEvent*)m_event);
    }

    ClutterActor *actor() const { return m_actor; }
    const ClutterTouchEvent *event() const { return m_event; }

private:
    ClutterActor *m_actor;
    const ClutterTouchEvent *m_event;
};

PlatformTouchEvent::PlatformTouchEvent(ClutterActor *actor, const ClutterTouchEvent *event)
    : PlatformEvent()
{
    typedef HashMap<unsigned, ClutterTouchEventDetails* > SequenceTouchesMap;

    // FIXME: Instead of a static variable, this should be stored in an instance
    // with a lifecycle similar to Page
    DEFINE_STATIC_LOCAL(SequenceTouchesMap, touchesBySequence, ());
    unsigned sequence = GPOINTER_TO_UINT(event->sequence);
    // Increment the platform touch id by 1 to avoid storing a key of 0 in the hashmap.
    unsigned id = sequence + 1;
    PlatformTouchPoint::State state;
    IntPoint position, globalPosition;

    m_modifiers = modifiersForClutterTouchEvent(event);
    m_timestamp = currentTime();

    UtilClutter::convertEventCoords(actor, event->x, event->y, position, globalPosition);

    switch (event->type) {
    case CLUTTER_TOUCH_BEGIN:
        state = PlatformTouchPoint::TouchPressed;

        m_type = WebCore::PlatformEvent::TouchStart;
        m_touchPoints.append(PlatformTouchPoint(sequence, position, globalPosition, state));

        break;
    case CLUTTER_TOUCH_UPDATE:
        state = PlatformTouchPoint::TouchMoved;

        m_type = WebCore::PlatformEvent::TouchMove;
        m_touchPoints.append(PlatformTouchPoint(sequence, position, globalPosition, state));

        touchesBySequence.remove(id);
        break;
    case CLUTTER_TOUCH_END:
        state = PlatformTouchPoint::TouchReleased;

        m_type = WebCore::PlatformEvent::TouchEnd;
        m_touchPoints.append(PlatformTouchPoint(sequence, position, globalPosition, state));

        touchesBySequence.remove(id);
        break;
    case CLUTTER_TOUCH_CANCEL:
        state = PlatformTouchPoint::TouchCancelled;

        m_type = WebCore::PlatformEvent::TouchCancel;
        m_touchPoints.append(PlatformTouchPoint(sequence, position, globalPosition, state));

        touchesBySequence.remove(id);
        break;
    default:
        ASSERT_NOT_REACHED();
    }

    for (SequenceTouchesMap::const_iterator::Values it = touchesBySequence.begin().values(), end = touchesBySequence.end().values(); it != end; ++it) {
        const ClutterTouchEventDetails *stationary_event = *it;
        sequence = GPOINTER_TO_UINT(stationary_event->event()->sequence);
        state = PlatformTouchPoint::TouchStationary;
        UtilClutter::convertEventCoords(stationary_event->actor(), stationary_event->event()->x, stationary_event->event()->y, position, globalPosition);
        
        m_touchPoints.append(PlatformTouchPoint(sequence, position, globalPosition, state));
    }

    if (event->type == CLUTTER_TOUCH_BEGIN || event->type == CLUTTER_TOUCH_UPDATE)
        touchesBySequence.set(id, new ClutterTouchEventDetails(actor, event));
}

}

#endif
