/*
 * Copyright (C) 2006 Apple Computer, Inc.  All rights reserved.
 * Copyright (C) 2006 Michael Emmel mike.emmel@gmail.com
 * Copyright (C) 2010 Collabora Ltd.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "PlatformWheelEvent.h"

#include "Scrollbar.h"
#include "UtilClutter.h"
#include <clutter/clutter.h>

namespace WebCore {

// Keep this in sync with the other platform event constructors
PlatformWheelEvent::PlatformWheelEvent(ClutterScrollEvent* event)
    : PlatformEvent(PlatformEvent::Wheel)
    , m_deltaX(0)
    , m_deltaY(0)
    , m_wheelTicksX(0)
    , m_wheelTicksY(0)
    , m_granularity(ScrollByPixelWheelEvent)
    , m_directionInvertedFromDevice(false)
    , m_hasPreciseScrollingDeltas(false)
    , m_phase(PlatformWheelEventPhaseNone)
    , m_momentumPhase(PlatformWheelEventPhaseNone)
    , m_scrollCount(0)
    , m_unacceleratedScrollingDeltaX(0)
    , m_unacceleratedScrollingDeltaY(0)
{
    static const float delta = 1;

    ClutterModifierType state = clutter_event_get_state((ClutterEvent*)event);
    if (state & CLUTTER_SHIFT_MASK)
        m_modifiers |= ShiftKey;
    if (state & CLUTTER_CONTROL_MASK)
        m_modifiers |= CtrlKey;
    if (state & CLUTTER_MOD1_MASK)
        m_modifiers |= AltKey;
    if (state & CLUTTER_META_MASK)
        m_modifiers |= MetaKey;

    m_timestamp = clutter_event_get_time((ClutterEvent*)event) / 1000.0;

    float pixelsPerLineStep = static_cast<float>(PlatformWheelEvent::pixelsPerLineStep());

    // Docs say an upwards scroll (away from the user) has a positive delta
    ClutterScrollDirection direction = clutter_event_get_scroll_direction((ClutterEvent*)event);
    switch (direction) {
    case CLUTTER_SCROLL_SMOOTH: {
        gdouble dX, dY;
        clutter_event_get_scroll_delta((ClutterEvent*)event, &dX, &dY);
        m_wheelTicksX = -dX;
        m_wheelTicksY = -dY;
        break;
    }
    case CLUTTER_SCROLL_UP:
        m_wheelTicksY = delta;
        break;
    case CLUTTER_SCROLL_DOWN:
        m_wheelTicksY = -delta;
        break;
    case CLUTTER_SCROLL_LEFT:
        m_wheelTicksX = delta;
        break;
    case CLUTTER_SCROLL_RIGHT:
        m_wheelTicksX = -delta;
        break;
    }

    m_deltaX = m_wheelTicksX * pixelsPerLineStep;
    m_deltaY = m_wheelTicksY * pixelsPerLineStep;

    gfloat x, y;
    ClutterActor* source = clutter_event_get_source((ClutterEvent*)event);
    clutter_event_get_coords((ClutterEvent*)event, &x, &y);
    UtilClutter::convertEventCoords(source, x, y, m_position, m_globalPosition);

    m_granularity = ScrollByPixelWheelEvent;
}

}
