/*
 * Copyright (C) 2011, 2012 Igalia S.L.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301 USA
 */

#include "config.h"
#include "GraphicsContext3DPrivate.h"

#if USE(3D_GRAPHICS)

#include "HostWindow.h"
#include "NotImplemented.h"
#include "GraphicsLayerActor.h"

#include "OpenGLShims.h"
#include "OpenGLESShims.h"

namespace WebCore {

GraphicsContext3DPrivate* GraphicsContext3DPrivate::s_current_context = 0;

PassOwnPtr<GraphicsContext3DPrivate> GraphicsContext3DPrivate::create(GraphicsContext3D* context)
{
    return adoptPtr(new GraphicsContext3DPrivate(context));
}

bool GraphicsContext3DPrivate::isCOGLES2Supported()
{
    static bool result = false;
    static bool validated = false;
    if (!validated) {
        // We don't need a large size framebuffer here.
        CoglHandle texture = cogl_texture_new_with_size (1, 1, COGL_TEXTURE_NO_SLICING, COGL_PIXEL_FORMAT_ANY);
        ASSERT(texture);

        CoglFramebuffer* framebuffer = (CoglFramebuffer*)cogl_offscreen_new_to_texture(COGL_TEXTURE(texture));
        ASSERT(framebuffer);

        CoglContext* context = cogl_framebuffer_get_context(framebuffer);
        ASSERT(context);

        result = cogl_has_feature(context, COGL_FEATURE_ID_GLES2_CONTEXT);
        if (!result)
            g_warning("Backend doesn't support creating GLES2 contexts.");
        validated = true;

        cogl_handle_unref(texture);
        cogl_object_unref(framebuffer);
    }
    return result;
}

void GraphicsContext3DPrivate::popGraphicsContext()
{
    if (!s_current_context)
        return;

    GraphicsContext3DPrivate* context = s_current_context;
    CoglContext* coglContext = cogl_framebuffer_get_context(context->m_framebuffer);
    cogl_pop_gles2_context(coglContext);

    s_current_context = 0;
}

const CoglGLES2Vtable* GraphicsContext3DPrivate::openGLFunctionTable()
{
    ASSERT_WITH_MESSAGE(s_current_context,
                        "Trying to do a GL call without a current GraphicsContext3D");

    GraphicsContext3DPrivate* context = s_current_context;
    return cogl_gles2_context_get_vtable(context->m_gl_context);
}

GraphicsContext3DPrivate::GraphicsContext3DPrivate(GraphicsContext3D* context)
    : m_context(context)
    , m_texture(0)
    , m_gl_context(0)
    , m_framebuffer(0)
    , m_currentWidth(1)
    , m_currentHeight(1)
    , m_needsClearing(true)
    , m_transformLayer(0)
{
}

GraphicsContext3DPrivate::~GraphicsContext3DPrivate()
{
    g_debug ("GraphicsContext3DPrivate::~GraphicsContext3DPrivate");

    deleteFBOs();

    if (m_gl_context) {
        cogl_object_unref (m_gl_context);
        m_gl_context = 0;
    }
}

void GraphicsContext3DPrivate::markLayerComposited()
{
    this->m_context->markLayerComposited();
}

static void
layerPaintCallback(ClutterActor *layer, GraphicsContext3DPrivate *context)
{
    context->markLayerComposited();
}

void GraphicsContext3DPrivate::ensureFBOs()
{
    if (!m_texture) {
        m_texture = cogl_texture_new_with_size (m_currentWidth, m_currentHeight,
                                                COGL_TEXTURE_NO_SLICING,
                                                COGL_PIXEL_FORMAT_ANY);
        ASSERT(m_texture);
    }

    if (!m_transformLayer) {
        m_transformLayer = adoptGRef(CLUTTER_ACTOR (graphicsLayerActorNew(GraphicsLayerClutter::LayerTypeCanvasLayer)));
        g_signal_connect_after (m_transformLayer.get(), "paint", G_CALLBACK(layerPaintCallback), this);

        graphicsLayerActorSetCoglTexture(GRAPHICS_LAYER_ACTOR (m_transformLayer.get()),
                                         COGL_TEXTURE (m_texture));
    }

    if (!m_framebuffer) {
        m_framebuffer = (CoglFramebuffer *) cogl_offscreen_new_to_texture (COGL_TEXTURE (m_texture));
        ASSERT(m_framebuffer);
        
        m_needsClearing = true;
    }
}

void GraphicsContext3DPrivate::deleteFBOs()
{
    GraphicsContext3DPrivate::popGraphicsContext();

    m_transformLayer = 0;

    if (m_framebuffer) {
        cogl_object_unref (m_framebuffer);
        m_framebuffer = 0;
    }

    if (m_texture) {
        cogl_handle_unref (m_texture);
        m_texture = 0;
    }
}

bool GraphicsContext3DPrivate::ensureGLContext()
{
    GError *error = NULL;

    ensureFBOs();

    if (!m_gl_context) {
        CoglContext* coglContext = cogl_framebuffer_get_context(m_framebuffer);
        m_gl_context = cogl_gles2_context_new(coglContext, &error);
        if (!m_gl_context)
          {
            g_error ("Couldn't create GLES2 context: %s", error->message);
            g_error_free (error);
            return false;
          }
    }

    return true;
}

bool GraphicsContext3DPrivate::makeContextCurrent()
{
    gboolean result;
    GError *error = NULL;

    if (s_current_context == this)
        return true;

    if (s_current_context)
        GraphicsContext3DPrivate::popGraphicsContext();
    //g_debug ("GraphicsContext3DPrivate::makeContextCurrent()");

    result = ensureGLContext();
    if (!result)
        return false;

    CoglContext* coglContext = cogl_framebuffer_get_context(m_framebuffer);
    result = cogl_push_gles2_context(coglContext, m_gl_context,
                                     m_framebuffer, m_framebuffer, &error);
    if (!result)
      {
        g_error ("Couldn't push GLES2 context: %s", error->message);
        g_error_free (error);
        return false;
      }

    s_current_context = this;

    if (m_needsClearing)
      {
        // Initialize renderbuffers to 0.
        GLfloat clearColor[] = {0, 0, 0, 0}, clearDepth = 0;
        GLint clearStencil = 0;
        GLboolean colorMask[] = {GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE}, depthMask = GL_TRUE;
        GLuint stencilMask = 0xffffffff;
        GLbitfield clearMask = GL_COLOR_BUFFER_BIT;
        ::glGetFloatv(GL_COLOR_CLEAR_VALUE, clearColor);
        ::glClearColor(0, 0, 0, 0);
        ::glGetBooleanv(GL_COLOR_WRITEMASK, colorMask);
        ::glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

        ::glGetFloatv(GL_DEPTH_CLEAR_VALUE, &clearDepth);
        m_context->clearDepth(1);
        ::glGetBooleanv(GL_DEPTH_WRITEMASK, &depthMask);
        ::glDepthMask(GL_TRUE);
        clearMask |= GL_DEPTH_BUFFER_BIT;

        ::glGetIntegerv(GL_STENCIL_CLEAR_VALUE, &clearStencil);
        ::glClearStencil(0);
        ::glGetIntegerv(GL_STENCIL_WRITEMASK, reinterpret_cast<GLint*>(&stencilMask));
        ::glStencilMaskSeparate(GL_FRONT, 0xffffffff);
        clearMask |= GL_STENCIL_BUFFER_BIT;

        ::glDisable(GL_SCISSOR_TEST);
        ::glDisable(GL_DITHER);

        ::glClear(clearMask);

        ::glClearColor(clearColor[0], clearColor[1], clearColor[2], clearColor[3]);
        ::glColorMask(colorMask[0], colorMask[1], colorMask[2], colorMask[3]);

        m_context->clearDepth(clearDepth);
        ::glDepthMask(depthMask);

        ::glClearStencil(clearStencil);
        ::glStencilMaskSeparate(GL_FRONT, stencilMask);

        m_needsClearing = false;
      }

    return true;
}

void GraphicsContext3DPrivate::reshape(int width, int height)
{
    if (width == m_currentWidth && height == m_currentHeight)
        return;

    deleteFBOs();

    m_currentWidth = width;
    m_currentHeight = height;
}

PlatformGraphicsContext3D GraphicsContext3DPrivate::platformContext()
{
    g_debug ("GraphicsContext3DPrivate::platformContext(): 0");
    return 0;
}

ClutterActor *GraphicsContext3DPrivate::platformLayer()
{
    if (!m_transformLayer)
        ensureGLContext();
    return m_transformLayer.get();
}

} // namespace WebCore

#endif // USE(3D_GRAPHICS)
