/*
 * Copyright 2011, 2012, 2013 Collabora Limited
 * Copyright (C) 2012 Intel Corporation. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "config.h"

#if USE(ACCELERATED_COMPOSITING)
#include "GraphicsLayerActor.h"

#include "GraphicsContext.h"
#include "GraphicsLayerClutter.h"
#include "Logging.h"
#include "PlatformClutterLayerClient.h"
#include "PlatformContextCairo.h"
#include "RefPtrCairo.h"
#include "TransformationMatrix.h"
#include <wtf/gobject/GOwnPtr.h>
#include <wtf/text/CString.h>

#ifndef NDEBUG
#include <glib/gprintf.h>
#endif

#ifndef g_clear_pointer /* Remove this when we can depend on GLib >= 2.34 */
#define g_clear_pointer(pp, destroy) \
  G_STMT_START {                                                               \
    G_STATIC_ASSERT (sizeof *(pp) == sizeof (gpointer));                       \
    /* Only one access, please */                                              \
    gpointer *_pp = (gpointer *) (pp);                                         \
    gpointer _p;                                                               \
    /* This assignment is needed to avoid a gcc warning */                     \
    GDestroyNotify _destroy = (GDestroyNotify) (destroy);                      \
                                                                               \
    (void) (0 ? (gpointer) *(pp) : 0);                                         \
    do                                                                         \
      _p = g_atomic_pointer_get (_pp);                                         \
    while G_UNLIKELY (!g_atomic_pointer_compare_and_exchange (_pp, _p, NULL)); \
                                                                               \
    if (_p)                                                                    \
      _destroy (_p);                                                           \
  } G_STMT_END
#endif /* g_clear_pointer */

using namespace WebCore;

G_DEFINE_TYPE(GraphicsLayerActor, graphics_layer_actor, CLUTTER_TYPE_ACTOR)

#define GRAPHICS_LAYER_ACTOR_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), GRAPHICS_LAYER_TYPE_ACTOR, GraphicsLayerActorPrivate))

static void createOffScreenFrameBuffer(GraphicsLayerActor*, const CoglMatrix*);

struct _GraphicsLayerActorPrivate {
    GraphicsLayerClutter::LayerType layerType;
    // Guard against changing size while allocating.
    gboolean allocating;

    /* Children */
    ClutterActor* texture; // used by video layer only. we should eventually remove this.
    CoglTexture* coglTexture;
    RefPtr<cairo_surface_t> surface;
    CoglMatrix* childrenMatrix;

    PlatformClutterLayerClient* layerClient;

    bool flatten;
    bool drawsContent;
    bool masksToBounds;

    float scrollX;
    float scrollY;

    float translateX;
    float translateY;

    gboolean sortByDepth;

    gboolean backfaceCulling;

    gboolean offscreen;
    CoglTexture* offscreenTexture;
    CoglFramebuffer* offscreenFramebuffer;
    gfloat expandX;
    gfloat expandY;
    gfloat fboWidth;
    gfloat fboHeight;

    gfloat debugBorderWidth;
    ClutterColor debugBorderColor;

    guint prePaintHandlerID;
};

enum {
    PROP_0,
    PROP_DOUBLE_SIDED,
    PROP_TRANSLATE_X,
    PROP_TRANSLATE_Y,
    PROP_LAST
};

static void graphicsLayerActorSetProperty(GObject* object, guint propID, const GValue* value, GParamSpec *pspec)
{
    GraphicsLayerActor* layer = GRAPHICS_LAYER_ACTOR(object);

    switch (propID) {
    case PROP_TRANSLATE_X:
        graphicsLayerActorSetTranslateX(layer, g_value_get_float(value));
        break;
    case PROP_TRANSLATE_Y:
        graphicsLayerActorSetTranslateY(layer, g_value_get_float(value));
        break;
    case PROP_DOUBLE_SIDED:
        graphicsLayerActorSetDoubleSided(layer, g_value_get_boolean(value));
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, propID, pspec);
    }
}

static void graphicsLayerActorGetProperty(GObject* object, guint propID, GValue* value, GParamSpec* pspec)
{
    GraphicsLayerActor* layer = GRAPHICS_LAYER_ACTOR(object);

    switch (propID) {
    case PROP_TRANSLATE_X:
        g_value_set_float(value, graphicsLayerActorGetTranslateX(layer));
        break;
    case PROP_TRANSLATE_Y:
        g_value_set_float(value, graphicsLayerActorGetTranslateY(layer));
        break;
    case PROP_DOUBLE_SIDED:
        g_value_set_boolean(value, graphicsLayerActorGetDoubleSided(layer));
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, propID, pspec);
    }
}

static void graphicsLayerActorDispose(GObject* object)
{
    GraphicsLayerActor* layer = GRAPHICS_LAYER_ACTOR(object);
    GraphicsLayerActorPrivate* priv = layer->priv;

    g_clear_pointer(&priv->coglTexture, cogl_object_unref);

    priv->surface.clear();

    g_clear_pointer(&priv->childrenMatrix, cogl_matrix_free);
    g_clear_pointer(&priv->offscreenTexture, cogl_object_unref);
    g_clear_pointer(&priv->offscreenFramebuffer, cogl_object_unref);

    G_OBJECT_CLASS(graphics_layer_actor_parent_class)->dispose(object);
}

static void graphicsLayerActorFinalize(GObject* object)
{
    GraphicsLayerActor* layer = GRAPHICS_LAYER_ACTOR(object);

    clutter_threads_remove_repaint_func(layer->priv->prePaintHandlerID);

    G_OBJECT_CLASS(graphics_layer_actor_parent_class)->finalize(object);
}

// Copied from cairo.
#define MAX_IMAGE_SIZE 32767

static void graphicsLayerActorAllocate(ClutterActor* self, const ClutterActorBox *box, ClutterAllocationFlags flags)
{
    GraphicsLayerActor* layer = GRAPHICS_LAYER_ACTOR(self);
    GraphicsLayerActorPrivate* priv = layer->priv;

    LOG(AcceleratedCompositing, "%s actor=%p %s box:%.2fx%.2f:%.2fx%.2f",
        __func__, self, clutter_actor_get_name(self),
        box->x1, box->y1, box->x2, box->y2);

    priv->allocating = TRUE;

    flags = (ClutterAllocationFlags)(flags | CLUTTER_DELEGATE_LAYOUT);

    CLUTTER_ACTOR_CLASS(graphics_layer_actor_parent_class)->allocate(self, box, flags);

    ClutterContent* canvas = clutter_actor_get_content(self);
    if (canvas) {
        float textureWidth = std::min(static_cast<int>(clutter_actor_box_get_width(box)), MAX_IMAGE_SIZE);
        float textureHeight = std::min(static_cast<int>(clutter_actor_box_get_height(box)), MAX_IMAGE_SIZE);
        clutter_canvas_set_size(CLUTTER_CANVAS(canvas), textureWidth, textureHeight);
    }

    if (priv->texture && !priv->coglTexture) {
        gfloat textureWidth = clutter_actor_box_get_width(box);
        gfloat textureHeight = clutter_actor_box_get_height(box);

        // FIXME: some pages make us create very big layers, which
        // exceeds cairo's limits, check for that here.
        if (textureWidth > MAX_IMAGE_SIZE) {
            g_warning("Layer %s has a width that exceeds cairo's limits: %.2f", clutter_actor_get_name(self), textureWidth);
            textureWidth = MAX_IMAGE_SIZE;
        }

        if (textureHeight > MAX_IMAGE_SIZE) {
            g_warning("Layer %s has a height that exceeds cairo's limits: %.2f", clutter_actor_get_name(self), textureHeight);
            textureHeight = MAX_IMAGE_SIZE;
        }

        // The texture occupies the whole area, but is positioned at 0,0
        // relative to the layer actor.
        ClutterActorBox textureBox = CLUTTER_ACTOR_BOX_INIT(0, 0, textureWidth, textureHeight);
        clutter_actor_allocate(priv->texture, &textureBox, flags);
    }

    priv->allocating = FALSE;
}

static bool isIdentityOrTranslation(CoglMatrix* matrix)
{
    return matrix->xx == 1 && !matrix->yx && !matrix->zx && !matrix->wx
           && !matrix->xy && matrix->yy == 1 && !matrix->zy && !matrix->wy
           && !matrix->xz && !matrix->yz && matrix->zz == 1 && !matrix->wz
           && matrix->ww == 1;
}

static void graphicsLayerActorApplyTransform(ClutterActor *actor, CoglMatrix *matrix)
{
    LOG(AcceleratedCompositing, "%s actor=%p %s", __func__, actor, clutter_actor_get_name(actor));

    GraphicsLayerActorPrivate* priv = GRAPHICS_LAYER_ACTOR(actor)->priv;


    // Apply translation and scrolling as a single translation. These
    // need to come before anything else, otherwise they'll be
    // affected by other operations such as scaling, which is not what
    // we want.
    float translateX = priv->translateX - priv->scrollX;
    float translateY = priv->translateY - priv->scrollY;

    if (translateX || translateY)
        cogl_matrix_translate(matrix, translateX, translateY, 0);

    CoglMatrix modelViewTransform = TransformationMatrix();
    CLUTTER_ACTOR_CLASS(graphics_layer_actor_parent_class)->apply_transform(actor, &modelViewTransform);

    if (priv->flatten)
        modelViewTransform = TransformationMatrix(&modelViewTransform).to2dTransform();
    cogl_matrix_multiply(matrix, matrix, &modelViewTransform);
}

// Draw content into the layer.
static void drawLayerContents(ClutterActor* actor, GraphicsContext& context)
{
    LOG(AcceleratedCompositing, "graphics_layer_actor:%s ", __func__);
    GraphicsLayerActorPrivate* priv = GRAPHICS_LAYER_ACTOR(actor)->priv;

    if (!priv->drawsContent || !priv->layerClient)
        return;

    gfloat width = clutter_actor_get_width(actor);
    gfloat height = clutter_actor_get_height(actor);
    IntRect clip(0, 0, width, height);

    // Apply the painted content to the layer.
    priv->layerClient->platformClutterLayerPaintContents(context, clip);
}

static gboolean graphicsLayerActorDraw(ClutterCanvas* texture, cairo_t* cr, gint width, gint height, GraphicsLayerActor* layer)
{
    if (!width || !height) {
        LOG(AcceleratedCompositing, "actor(%p)'s width or height = 0", layer);
        return FALSE;
    }

    GraphicsLayerActorPrivate* priv = layer->priv;
    GraphicsContext context(cr);
    context.clearRect(FloatRect(0, 0, width, height));

    if (priv->surface) {
        int surfaceWidth = cairo_image_surface_get_width(priv->surface.get());
        int surfaceHeight = cairo_image_surface_get_height(priv->surface.get());

        FloatRect srcRect(0.0, 0.0, static_cast<float>(surfaceWidth), static_cast<float>(surfaceHeight));
        FloatRect destRect(0.0, 0.0, width, height);
        context.platformContext()->drawSurfaceToContext(priv->surface.get(), destRect, srcRect, &context);
    }

    if (priv->layerType == GraphicsLayerClutter::LayerTypeWebLayer)
        drawLayerContents(CLUTTER_ACTOR(layer), context);

    return TRUE;
}

static gint sortByDepth(gconstpointer a, gconstpointer b, gpointer data)
{
    gfloat depthA = graphicsLayerActorGetDepth(GRAPHICS_LAYER_ACTOR(a));
    gfloat depthB = graphicsLayerActorGetDepth(GRAPHICS_LAYER_ACTOR(b));

    if (depthA < depthB)
        return -1;

    if (depthA == depthB)
        return 0;

    /* Tell the caller that items order will change. */
    *(gboolean*)data = TRUE;
    return 1;
}

static void graphicsLayerActorSortDepthOrder(GraphicsLayerActor* graphicsLayer)
{
    ClutterActor* self = CLUTTER_ACTOR(graphicsLayer);
    GList* children;
    gboolean refresh = FALSE;

    children = clutter_actor_get_children (self);
    children = g_list_sort_with_data (children, (GCompareDataFunc)sortByDepth, &refresh);
    if (refresh) {
        /* Sort children by resetting their positions in reverse order since
         * setting at index 0 is faster than appending to the list */
        for (GList* c = g_list_last(children); c != NULL; c = g_list_previous(c))
          clutter_actor_set_child_at_index (self, CLUTTER_ACTOR (c->data), 0);
    }
    g_list_free(children);
}

static void graphicsLayerActorPaint(ClutterActor* actor)
{
    GraphicsLayerActor* layer = GRAPHICS_LAYER_ACTOR(actor);
    GraphicsLayerActorPrivate* priv = layer->priv;

    if (priv->masksToBounds)
        clutter_actor_set_clip(actor, priv->scrollX, priv->scrollY, clutter_actor_get_width(actor), clutter_actor_get_height(actor));
    else if (clutter_actor_has_clip(actor))
        clutter_actor_remove_clip(actor);

    cogl_set_backface_culling_enabled(priv->backfaceCulling);

    if (priv->texture)
        clutter_actor_paint(priv->texture);

    if (priv->coglTexture) {
        cogl_set_source_texture (priv->coglTexture);
        cogl_rectangle (0., 0., clutter_actor_get_width(actor),
                        clutter_actor_get_height(actor));
    }

    // Redirects all subsequent drawing to this offscreen buffer
    if (priv->offscreen) {
        CoglColor transparent;

        cogl_push_framebuffer(priv->offscreenFramebuffer);
        cogl_color_init_from_4ub(&transparent, 0, 0, 0, 0);
        cogl_clear(&transparent, COGL_BUFFER_BIT_COLOR | COGL_BUFFER_BIT_DEPTH);
    }

    GOwnPtr<GList> children(clutter_actor_get_children(actor));
    for (GList* child = children.get(); child; child = child->next)
        clutter_actor_paint(CLUTTER_ACTOR(child->data));

    if (priv->debugBorderWidth > 0) {
        ClutterActorBox box;
        gfloat width, height;
        clutter_actor_get_allocation_box(actor, &box);
        clutter_actor_box_get_size(&box, &width, &height);
        cogl_set_source_color4ub(priv->debugBorderColor.red,
                                 priv->debugBorderColor.green,
                                 priv->debugBorderColor.blue,
                                 200);
        cogl_rectangle(priv->debugBorderWidth, 0,
                       width,
                       priv->debugBorderWidth);
        cogl_rectangle(width - priv->debugBorderWidth,
                       priv->debugBorderWidth,
                       width,
                       height);
        cogl_rectangle(0, height - priv->debugBorderWidth,
                       width - priv->debugBorderWidth,
                       height);
        cogl_rectangle(0, 0,
                       priv->debugBorderWidth,
                       height - priv->debugBorderWidth);
    }

    if (priv->offscreen) {
       cogl_pop_framebuffer();

       CoglPipeline *pipeline;
       CoglContext *ctx = clutter_backend_get_cogl_context (clutter_get_default_backend ());

       pipeline = cogl_pipeline_new(ctx);
       cogl_pipeline_set_layer_texture(pipeline, 0, priv->offscreenTexture);

       cogl_set_source(pipeline);
       cogl_rectangle_with_texture_coords (-priv->expandX, -priv->expandY,
                                           priv->fboWidth - priv->expandX,
                                           priv->fboHeight - priv->expandY,
                                           0, 0, 1, 1);
       cogl_object_unref(pipeline);
    }
}

static gboolean
graphicsLayerActorGetPaintVolume(ClutterActor* actor, ClutterPaintVolume* volume)
{
    const ClutterPaintVolume* defaultVolume = clutter_actor_get_default_paint_volume(actor);
    if (!defaultVolume)
        return FALSE;

    clutter_paint_volume_set_width(volume, clutter_paint_volume_get_width(defaultVolume));
    clutter_paint_volume_set_height(volume, clutter_paint_volume_get_height(defaultVolume));
    clutter_paint_volume_set_depth(volume, clutter_paint_volume_get_depth(defaultVolume));

    ClutterVertex vertex;
    clutter_paint_volume_get_origin(defaultVolume, &vertex);
    clutter_paint_volume_set_origin(volume, &vertex);
    return TRUE;
}

static gboolean graphicsLayerActorDepthSorting(gpointer data)
{
    GraphicsLayerActor* layer = GRAPHICS_LAYER_ACTOR(data);
    GraphicsLayerActorPrivate* priv = layer->priv;

    // Set the depth to the actor.
    // This had better be in the apply_transform function, but the apply_transform is called twice 
    // before painting. As a result, the depth value gotten for the first time can be overwritten 
    // by the depth value gotten for the next time. 
    // Therefore, the first depth value will be only used to be sorted by the parent layer.
    ClutterActor* actor = CLUTTER_ACTOR(layer);
    ClutterActor* parent = clutter_actor_get_parent(actor);
    if (GRAPHICS_LAYER_IS_ACTOR(parent) && graphicsLayerActorGetLayerType(GRAPHICS_LAYER_ACTOR(parent)) == GraphicsLayerClutter::LayerTypeTransformLayer) {
       CoglMatrix matrix;
       cogl_get_modelview_matrix(&matrix);

       // FIXME: no case of which returns TRUE.
       if (isIdentityOrTranslation(&matrix))
           clutter_actor_set_z_position(actor, matrix.zw);
       else {
           float width = clutter_actor_get_width(actor);
           float height = clutter_actor_get_height(actor);

           float x = width * 0.5, y = height * 0.5, z = 0, w = 1;
           cogl_matrix_transform_point (&matrix, &x, &y, &z, &w);
           z /= w;

           if (isnan(z))
               z = 0;

           gfloat zpos = clutter_actor_get_z_position(actor);
           if (zpos != z) {
               clutter_actor_set_z_position (actor, z);
               // Clutter tries to paint Actors before sorting, so we need  
               // to sort the siblings by depth when each sibling gets 
               // the depth value for the first time.
               if ((GRAPHICS_LAYER_ACTOR(parent))->priv->sortByDepth == FALSE)
                   graphicsLayerActorSortDepthOrder(GRAPHICS_LAYER_ACTOR(parent));
           }
       }
    }

    // Sort the child layers of the transform layer. 
    if (priv->layerType == GraphicsLayerClutter::LayerTypeTransformLayer) {
        graphicsLayerActorSortDepthOrder(GRAPHICS_LAYER_ACTOR(actor));
        priv->sortByDepth = TRUE;
    }

    return TRUE;
}

static void
graphics_layer_actor_class_init(GraphicsLayerActorClass *klass)
{
    GObjectClass* objectClass = G_OBJECT_CLASS(klass);
    ClutterActorClass* actorClass = CLUTTER_ACTOR_CLASS(klass);
    GParamSpec* pspec;
    
    objectClass->get_property = graphicsLayerActorGetProperty;
    objectClass->set_property = graphicsLayerActorSetProperty;
    objectClass->dispose = graphicsLayerActorDispose;
    objectClass->finalize = graphicsLayerActorFinalize;
    actorClass->apply_transform = graphicsLayerActorApplyTransform;
    actorClass->get_paint_volume = graphicsLayerActorGetPaintVolume;
    actorClass->allocate = graphicsLayerActorAllocate;
    actorClass->paint = graphicsLayerActorPaint;

    g_type_class_add_private(klass, sizeof(GraphicsLayerActorPrivate));

    pspec = g_param_spec_boolean("doubleSided",
                                 "DoubleSided",
                                 "Whether to use backface culling",
                                 FALSE,
                                 static_cast<GParamFlags>(G_PARAM_READWRITE));
    g_object_class_install_property(objectClass, PROP_DOUBLE_SIDED, pspec);

    pspec = g_param_spec_float("translate-x",
                               "Translate X",
                               "Translation value for the X axis",
                               -G_MAXFLOAT,
                               G_MAXFLOAT,
                               0.0,
                               static_cast<GParamFlags>(G_PARAM_READWRITE));
    g_object_class_install_property(objectClass, PROP_TRANSLATE_X, pspec);

    pspec = g_param_spec_float("translate-y",
                               "Translate Y",
                               "Translation value for the Y ayis",
                               -G_MAXFLOAT,
                               G_MAXFLOAT,
                               0.0,
                               static_cast<GParamFlags>(G_PARAM_READWRITE));
    g_object_class_install_property(objectClass, PROP_TRANSLATE_Y, pspec);
}

static void
graphics_layer_actor_init(GraphicsLayerActor *self)
{
    GraphicsLayerActorPrivate* priv = self->priv = GRAPHICS_LAYER_ACTOR_GET_PRIVATE(self);

    clutter_actor_set_reactive(CLUTTER_ACTOR(self), FALSE);

    priv->flatten = true;

    priv->scrollX = 0.0;
    priv->scrollY = 0.0;

    priv->translateX = 0.0;
    priv->translateY = 0.0;

    priv->sortByDepth = FALSE;

    priv->backfaceCulling = FALSE;

    priv->offscreen = FALSE;

    priv->debugBorderWidth = 0.0;
    priv->debugBorderColor = *CLUTTER_COLOR_Red;

    // Default used by GraphicsLayer.
    graphicsLayerActorSetAnchorPoint(self, 0.5, 0.5, 0.0);

    priv->prePaintHandlerID = clutter_threads_add_repaint_func_full(CLUTTER_REPAINT_FLAGS_PRE_PAINT, graphicsLayerActorDepthSorting, self, 0);
}

GraphicsLayerActor* graphicsLayerActorNew(GraphicsLayerClutter::LayerType type)
{
    GraphicsLayerActor* layer = GRAPHICS_LAYER_ACTOR(g_object_new(GRAPHICS_LAYER_TYPE_ACTOR, 0));
    GraphicsLayerActorPrivate* priv = layer->priv;

    priv->layerType = type;
    if (priv->layerType == GraphicsLayerClutter::LayerTypeTransformLayer)
        priv->flatten = false;

    // For video layers we don't want the cairo texture, but a regular one.
    if (priv->layerType == GraphicsLayerClutter::LayerTypeVideoLayer) {
        priv->texture = clutter_texture_new();
        clutter_actor_add_child(CLUTTER_ACTOR(layer), priv->texture);
        graphicsLayerActorSortDepthOrder(layer);
    }

    return layer;
}

GraphicsLayerActor* graphicsLayerActorNewWithClient(GraphicsLayerClutter::LayerType type, PlatformClutterLayerClient* layerClient)
{
    GraphicsLayerActor* layer = graphicsLayerActorNew(type);
    graphicsLayerActorSetClient(layer, layerClient);
    return layer;
}

void
graphicsLayerActorSetClient(GraphicsLayerActor* layer, PlatformClutterLayerClient* client)
{
    layer->priv->layerClient = client;
}

PlatformClutterLayerClient* graphicsLayerActorGetClient(GraphicsLayerActor* layer)
{
    return layer->priv->layerClient;
}

/* 
 * public API
 */
void graphicsLayerActorRemoveAll(GraphicsLayerActor* layer)
{
    LOG(AcceleratedCompositing, "%s\n", __func__);

    g_return_if_fail(GRAPHICS_LAYER_IS_ACTOR(layer));

    ClutterActor *child;
    ClutterActorIter iter;
    clutter_actor_iter_init (&iter, CLUTTER_ACTOR(layer));
    while (clutter_actor_iter_next (&iter, &child)) {
        // We only want to remove sublayers, not our own composite actors.
        if (child == layer->priv->texture)
            continue;

        clutter_actor_iter_remove(&iter);
    }
}

static void graphicsLayerActorUpdateTexture(GraphicsLayerActor* layer)
{
    GraphicsLayerActorPrivate* priv = layer->priv;
    ASSERT(priv->layerType != GraphicsLayerClutter::LayerTypeVideoLayer);
    ASSERT(priv->layerType != GraphicsLayerClutter::LayerTypeCanvasLayer);

    ClutterActor* actor = CLUTTER_ACTOR(layer);
    GRefPtr<ClutterContent> canvas = clutter_actor_get_content(actor);
    if (canvas) {
        // Nothing needs a texture, remove the one we have, if any.
        if (!priv->drawsContent && !priv->surface) {
            g_signal_handlers_disconnect_by_func(canvas.get(), reinterpret_cast<void*>(graphicsLayerActorDraw), layer);
            clutter_actor_set_content(actor, 0);
        }
        return;
    }

    canvas = adoptGRef(clutter_canvas_new());
    clutter_actor_set_content(actor, canvas.get());

    int width = std::max(static_cast<int>(ceilf(clutter_actor_get_width(actor))), 1);
    int height = std::max(static_cast<int>(ceilf(clutter_actor_get_height(actor))), 1);
    clutter_canvas_set_size(CLUTTER_CANVAS(canvas.get()), width, height);

    g_signal_connect(canvas.get(), "draw", G_CALLBACK(graphicsLayerActorDraw), layer);
}

cairo_surface_t* graphicsLayerActorGetSurface(GraphicsLayerActor* layer)
{
    LOG(AcceleratedCompositing, "%s\n", __func__);
    GraphicsLayerActorPrivate* priv = layer->priv;
    if (!priv->surface)
        g_message("surface is null");

    return priv->surface.get();
}

void graphicsLayerActorSetSurface(GraphicsLayerActor* layer, cairo_surface_t* surface)
{
    LOG(AcceleratedCompositing, "%s\n actor= %p", __func__, layer);
    GraphicsLayerActorPrivate* priv = layer->priv;
    priv->surface = surface;
    graphicsLayerActorUpdateTexture(layer);
}

void graphicsLayerActorInvalidateRectangle(GraphicsLayerActor* layer, const FloatRect& dirtyRect)
{
    GraphicsLayerActorPrivate* priv = layer->priv;
    if (priv->texture) {
        if (CLUTTER_IS_CAIRO_TEXTURE (priv->texture)) {
            cairo_rectangle_int_t rect(enclosingIntRect(dirtyRect));
            clutter_cairo_texture_invalidate_rectangle(CLUTTER_CAIRO_TEXTURE(priv->texture), &rect);
        } else
            clutter_actor_queue_redraw(priv->texture);
        return;
    }

    ClutterContent* canvas = clutter_actor_get_content(CLUTTER_ACTOR(layer));
    if (!canvas)
        return;

    ASSERT(CLUTTER_IS_CONTENT(canvas));

    // FIXME: Need to invalidate a specific area?
    clutter_content_invalidate(canvas);
}

static void createOffScreenFrameBuffer(GraphicsLayerActor* layer, const CoglMatrix* matrix)
{
    GraphicsLayerActorPrivate* priv = layer->priv;

    g_clear_pointer(&priv->offscreenTexture, cogl_object_unref);
    g_clear_pointer(&priv->offscreenFramebuffer, cogl_object_unref);
    g_clear_pointer(&priv->childrenMatrix, cogl_matrix_free);

    priv->childrenMatrix = cogl_matrix_copy(matrix);

    gfloat width = clutter_actor_get_width(CLUTTER_ACTOR(layer));
    gfloat height = clutter_actor_get_height(CLUTTER_ACTOR(layer));

    ClutterActorBox box;

    // FIXME: Current implementation has a boudary issue,
    // See http://desandro.github.com/3dtransforms/examples/perspective-03.html.
    // This is because we redirect 3d trasform contents drawing to offscreen buffer
    // in transform layer which has children. So we don't know how large fbo is required here.
    // For a workaround, we don't reduce fbo size than the previous one.
    if (clutter_actor_get_paint_box(CLUTTER_ACTOR(layer), &box)) {
        float boxWidth = 0, boxHeight = 0;
        clutter_actor_box_get_size(&box, &boxWidth, &boxHeight);
        if (priv->fboWidth < boxWidth)
            priv->fboWidth = boxWidth;
        if (priv->fboHeight < boxHeight)
            priv->fboHeight = boxHeight;
        priv->expandX = (priv->fboWidth -  width) / 2.0f;
        priv->expandY = (priv->fboHeight -  height) / 2.0f;
    } else {
        g_warning("Failed to get a paint box");
        float factor = 3.0f; 
        priv->fboWidth =  width * factor;
        priv->fboHeight=  height * factor;
        priv->expandX =  (width * factor - width) / 2.0f;
        priv->expandY =  (height * factor - height) / 2.0f;
    }

    /* 
     * Create an offscreen buffer with texture.
     */
    LOG(AcceleratedCompositing, "Creating texture with size (%f, %f)\n", priv->fboWidth, priv->fboHeight);

    priv->offscreenTexture = cogl_texture_new_with_size(priv->fboWidth, priv->fboHeight,
                                                 COGL_TEXTURE_NO_SLICING,
                                                 COGL_PIXEL_FORMAT_ARGB_8888);
 
    if (!priv->offscreenTexture) {
        g_warning("Failed creating texture with size!\n");
        return;
    }
 

    priv->offscreenFramebuffer = COGL_FRAMEBUFFER(cogl_offscreen_new_to_texture(priv->offscreenTexture));

    if (!priv->offscreenFramebuffer) {
        g_warning("Failed creating offscreen to texture!\n");
        return;
    }

    gfloat zCamera;
    CoglMatrix mvMatrix;
    CoglMatrix projectionMatrix;
    
    cogl_push_framebuffer(priv->offscreenFramebuffer);

    // The constant value needs to be multiplied in order to render like other WebKit browsers.
    gfloat perspectiveOriginX = -matrix->xz / matrix->wz;
    gfloat perspectiveOriginY = -matrix->yz / matrix->wz;

    /* Expand the viewport if the actor is partially off-stage,
     * otherwise the actor will end up clipped to the stage viewport
     */    
    cogl_set_viewport(-perspectiveOriginX, -perspectiveOriginY, priv->fboWidth, priv->fboHeight);

    /* childrenMatrix(childrenTransform) is given by WebCore.
       When CSS3 perspective property value is 400, it has the following matrix value:
       xx,xy,xz,xw=1.000000,0.000000,0.000000,0.000000
       yx,yy,yz,yw=0.000000,1.000000,0.000000,0.000000
       zx,zy,zz,zw=0.000000,0.000000,1.000000,0.000000
       wx,wy,wz,ww=0.000000,0.000000,-0.002500,1.000000
     */
    cogl_matrix_init_identity(&projectionMatrix);
    // Create a perspective projection matrix using the CSS3 perspective property value.

    gfloat css3Perspective = -1.0f / matrix->wz;
    LOG(AcceleratedCompositing, "CSS3 Perspective value is %f", css3Perspective);
    LOG(AcceleratedCompositing, "CSS3 Perspective-origin(x) value is %f", matrix->xz);
    LOG(AcceleratedCompositing, "CSS3 Perspective-origin(y) value is %f", matrix->yz);

    cogl_matrix_perspective(&projectionMatrix, 30, 1.0, 0.1, css3Perspective);
    cogl_set_projection_matrix(&projectionMatrix);

    zCamera = 0.5 * projectionMatrix.xx;
    cogl_matrix_init_identity(&mvMatrix);
    cogl_matrix_translate(&mvMatrix, -0.5f, -0.5f, -zCamera);
    // The last parameter has an effect on the depth.
    cogl_matrix_scale(&mvMatrix, 1.0f / width, -1.0f / height, 1.0f / css3Perspective / 0.55f);
    cogl_matrix_translate(&mvMatrix, 0.0f, -1.0f * height, 0.0f);

    cogl_matrix_translate(&mvMatrix, width / 2.0f, height / 2.0f, 0.0f);
    cogl_matrix_scale(&mvMatrix, width / priv->fboWidth, height / priv->fboHeight, 1.0f);
    cogl_matrix_translate(&mvMatrix, -width / 2.0f, -height / 2.0f, 0.0f);

    cogl_matrix_translate(&mvMatrix, perspectiveOriginX, perspectiveOriginY, 0.0f);
    cogl_set_modelview_matrix(&mvMatrix);

    cogl_pop_framebuffer();
}

void graphicsLayerActorSetAnchorPoint(GraphicsLayerActor* layer, gfloat x, gfloat y, gfloat z)
{
    ClutterActor* actor = CLUTTER_ACTOR(layer);
    clutter_actor_set_pivot_point(actor, x, y);
    clutter_actor_set_pivot_point_z(actor, z);
}

void graphicsLayerActorGetAnchorPoint(GraphicsLayerActor* layer, gfloat* x, gfloat* y, gfloat* z)
{
    ASSERT(x && y);

    ClutterActor* actor = CLUTTER_ACTOR(layer);
    clutter_actor_get_pivot_point(actor, x, y);
    if (z)
        *z = clutter_actor_get_pivot_point_z(actor);
}

void graphicsLayerActorSetChildrenTransform(GraphicsLayerActor* layer, const CoglMatrix* matrix)
{
    GraphicsLayerActorPrivate* priv = layer->priv;

    priv->offscreen = !cogl_matrix_is_identity(matrix);

    ClutterActorIter iter;
    ClutterActor *child;

    // The children of this layer which has a perspective property are not flatten 
    // to make the perspective work.
    clutter_actor_iter_init (&iter, CLUTTER_ACTOR(layer));
    while (clutter_actor_iter_next (&iter, &child)) {
        if (child == priv->texture)
            continue;

        graphicsLayerActorSetFlatten(GRAPHICS_LAYER_ACTOR(child), !priv->offscreen);
    }

    createOffScreenFrameBuffer(layer, matrix);

    clutter_actor_queue_redraw(CLUTTER_ACTOR(layer));
}

gfloat graphicsLayerActorGetDepth(GraphicsLayerActor* layer)
{
    return clutter_actor_get_z_position(CLUTTER_ACTOR(layer));
}

void graphicsLayerActorSetScrollPosition(GraphicsLayerActor* layer, gfloat x, gfloat y)
{
    if (x > 0 || y > 0)
        return;

    GraphicsLayerActorPrivate* priv = layer->priv;
    priv->scrollX = x;
    priv->scrollY = y;

    clutter_actor_queue_redraw(CLUTTER_ACTOR(layer));
}

gint graphicsLayerActorGetCountChildren(GraphicsLayerActor* layer)
{
    ASSERT(GRAPHICS_LAYER_IS_ACTOR(layer));

    return clutter_actor_get_n_children(CLUTTER_ACTOR(layer));
}

void graphicsLayerActorReplaceSublayer(GraphicsLayerActor* layer, ClutterActor* oldChildLayer, ClutterActor* newChildLayer)
{
    LOG(AcceleratedCompositing, "GraphicsLayerActor::%s", __func__);

    ASSERT(GRAPHICS_LAYER_IS_ACTOR(layer));
    ASSERT(CLUTTER_IS_ACTOR(oldChildLayer));
    ASSERT(CLUTTER_IS_ACTOR(newChildLayer));

    clutter_actor_replace_child(CLUTTER_ACTOR(layer), oldChildLayer, newChildLayer);

    graphicsLayerActorSortDepthOrder(layer);
}

void graphicsLayerActorInsertSublayer(GraphicsLayerActor* layer, ClutterActor* childLayer, gint index)
{
    LOG(AcceleratedCompositing, "GraphicsLayerActor::%s", __func__);

    ASSERT(GRAPHICS_LAYER_IS_ACTOR(layer));
    ASSERT(CLUTTER_IS_ACTOR(childLayer));

    clutter_actor_insert_child_at_index(CLUTTER_ACTOR(layer), childLayer, index);
    graphicsLayerActorSortDepthOrder(layer);
}

void graphicsLayerActorSetSublayers(GraphicsLayerActor* layer, GraphicsLayerActorList& subLayers)
{
    if (!subLayers.size()) {
        graphicsLayerActorRemoveAll(layer);
        return;
    }

    bool shouldFlattenChildren = !layer->priv->offscreen && layer->priv->layerType != GraphicsLayerClutter::LayerTypeTransformLayer;
    ClutterActor* newParentActor = CLUTTER_ACTOR(layer);
    for (size_t i = 0; i < subLayers.size(); ++i) {
        ClutterActor* childActor = CLUTTER_ACTOR(subLayers[i].get());
        ClutterActor* oldParentActor = clutter_actor_get_parent(childActor);
        if (oldParentActor) {
            if (oldParentActor == newParentActor)
                continue;
            clutter_actor_remove_child(oldParentActor, childActor);
        }
        clutter_actor_add_child(newParentActor, childActor);
        graphicsLayerActorSetFlatten(GRAPHICS_LAYER_ACTOR(childActor), shouldFlattenChildren);
    }
}

void graphicsLayerActorRemoveFromSuperLayer(GraphicsLayerActor* layer)
{
    ClutterActor* actor = CLUTTER_ACTOR(layer);
    ClutterActor* parentActor = clutter_actor_get_parent(actor);
    if (!parentActor)
        return;

    clutter_actor_remove_child(parentActor, actor);
}

void graphicsLayerActorSetDoubleSided(GraphicsLayerActor* layer, gboolean value)
{
    LOG(AcceleratedCompositing, "GraphicsLayerActor::%s(%d)", __func__, value);

    GraphicsLayerActorPrivate* priv = layer->priv;
    priv->backfaceCulling = !value;
}

gboolean graphicsLayerActorGetDoubleSided(GraphicsLayerActor* layer)
{
    LOG(AcceleratedCompositing, "GraphicsLayerActor::%s", __func__);

    GraphicsLayerActorPrivate* priv = layer->priv;
    return !priv->backfaceCulling;
}

GraphicsLayerClutter::LayerType graphicsLayerActorGetLayerType(GraphicsLayerActor* layer)
{
    LOG(AcceleratedCompositing, "GraphicsLayerActor::%s", __func__);
    GraphicsLayerActorPrivate* priv = layer->priv;
    return priv->layerType;
}

void graphicsLayerActorSetLayerType(GraphicsLayerActor* layer, GraphicsLayerClutter::LayerType layerType)
{
    GraphicsLayerActorPrivate* priv = layer->priv;
    priv->layerType = layerType;
}

void graphicsLayerActorSetTranslateX(GraphicsLayerActor* layer, float value)
{
    LOG(AcceleratedCompositing, "GraphicsLayerActor::%s(%f)", __func__, value);

    GraphicsLayerActorPrivate* priv = layer->priv;
    priv->translateX = value;
    clutter_actor_queue_redraw(CLUTTER_ACTOR(layer));
}

float graphicsLayerActorGetTranslateX(GraphicsLayerActor* layer)
{
    LOG(AcceleratedCompositing, "GraphicsLayerActor::%s", __func__);

    GraphicsLayerActorPrivate* priv = layer->priv;
    return priv->translateX;
}

void graphicsLayerActorSetTranslateY(GraphicsLayerActor* layer, float value)
{
    LOG(AcceleratedCompositing, "GraphicsLayerActor::%s(%f)", __func__, value);

    GraphicsLayerActorPrivate* priv = layer->priv;
    priv->translateY = value;
    clutter_actor_queue_redraw(CLUTTER_ACTOR(layer));
}

float graphicsLayerActorGetTranslateY(GraphicsLayerActor* layer)
{
    LOG(AcceleratedCompositing, "GraphicsLayerActor::%s", __func__);

    GraphicsLayerActorPrivate* priv = layer->priv;
    return priv->translateY;
}

// Only used for video layers. Should move to its own object.
ClutterTexture* graphicsLayerActorGetTexture(GraphicsLayerActor* layer)
{
    GraphicsLayerActorPrivate* priv = layer->priv;
    ASSERT(priv->texture && CLUTTER_IS_TEXTURE(priv->texture) && !CLUTTER_IS_CAIRO_TEXTURE(priv->texture));
    return CLUTTER_TEXTURE (priv->texture);
}

// Only used for WebGL layers. Should move to its own object.
void graphicsLayerActorSetCoglTexture(GraphicsLayerActor* layer,
                                      CoglTexture *texture)
{
    GraphicsLayerActorPrivate* priv = layer->priv;
    
    if (priv->coglTexture == texture)
        return;

    g_clear_pointer(&priv->coglTexture, cogl_object_unref);

    priv->coglTexture = static_cast<CoglTexture*>(cogl_object_ref(texture));
}

void graphicsLayerActorSetDrawsContent(GraphicsLayerActor* layer, bool drawsContent)
{
    GraphicsLayerActorPrivate* priv = layer->priv;

    if (drawsContent == priv->drawsContent)
        return;

    priv->drawsContent = drawsContent;

    graphicsLayerActorUpdateTexture(layer);
}

bool graphicsLayerActorGetDrawsContent(GraphicsLayerActor* layer)
{
    return layer->priv->drawsContent;
}

void graphicsLayerActorSetFlatten(GraphicsLayerActor* layer, bool flatten)
{
    GraphicsLayerActorPrivate* priv = layer->priv;
    if (flatten == priv->flatten)
        return;

    priv->flatten = flatten;
}

void graphicsLayerActorSetMasksToBounds(GraphicsLayerActor* layer, bool masksToBounds)
{
    GraphicsLayerActorPrivate* priv = layer->priv;
    if (priv->masksToBounds == masksToBounds)
        return;

    priv->masksToBounds = masksToBounds;
    clutter_actor_queue_redraw(CLUTTER_ACTOR(layer));
}

void graphicsLayerActorSetDebugBorder(GraphicsLayerActor* layer, gfloat width, ClutterColor* color)
{
    GraphicsLayerActorPrivate* priv = layer->priv;
    priv->debugBorderWidth = width;
    priv->debugBorderColor = color ? *color : *CLUTTER_COLOR_Red;
}

WebCore::PlatformClutterAnimation* graphicsLayerActorGetAnimationForKey(GraphicsLayerActor* layer, const String key)
{
    return static_cast<WebCore::PlatformClutterAnimation*>(g_object_get_data(G_OBJECT(layer), key.utf8().data()));
}

#endif // USE(ACCELERATED_COMPOSITING)
