/*
 * Copyright 2011 Collabora Limited
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "config.h"
#include "GraphicsLayerStage.h"

#if USE(ACCELERATED_COMPOSITING)
#include "GraphicsContext3DPrivate.h"
#include "Logging.h"
#include "cogl/cogl.h"

#ifndef NDEBUG
#include <glib/gprintf.h>
#endif

using namespace WebCore;

G_DEFINE_TYPE(GraphicsLayerStage, graphics_layer_stage, CLUTTER_TYPE_STAGE)

#define GRAPHICS_LAYER_STAGE_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), GRAPHICS_LAYER_TYPE_STAGE, GraphicsLayerStagePrivate))

#ifndef NDEBUG
#define matrixToString(matrix) g_strdup_printf("xx,xy,xz,xw=%f,%f,%f,%f\n" \
                                               "yx,yy,yz,yw=%f,%f,%f,%f\n" \
                                               "zx,zy,zz,zw=%f,%f,%f,%f\n" \
                                               "wx,wy,wz,ww=%f,%f,%f,%f", \
                                               matrix->xx, matrix->xy, matrix->xz, matrix->xw, \
                                               matrix->yx, matrix->yy, matrix->yz, matrix->yw, \
                                               matrix->zx, matrix->zy, matrix->zz, matrix->zw, \
                                               matrix->wx, matrix->wy, matrix->wz, matrix->ww)
#else
#define matrixToString(matrix) matrix
#endif

struct _GraphicsLayerStagePrivate
{
    CoglMatrix* perspectiveMatrix;
    StageProjectionMode projectionMode;
};

enum
{
    PROP_0,

    PROP_PROJECTION_MATRIX,
    PROP_PROJECTION_MODE,

    PROP_LAST
};

static void
graphics_layer_stage_set_property (GObject    *object,
                          guint       prop_id,
                          const GValue     *value,
                          GParamSpec *pspec)
{
    GraphicsLayerStage* stage = GRAPHICS_LAYER_STAGE(object);
    GraphicsLayerStagePrivate* priv = stage->priv;

    switch (prop_id) {
    case PROP_PROJECTION_MATRIX:
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
graphics_layer_stage_get_property (GObject      *object,
                        guint      prop_id,
                        GValue     *value,
                        GParamSpec *pspec)
{
    GraphicsLayerStage* stage = GRAPHICS_LAYER_STAGE(object);
    GraphicsLayerStagePrivate* priv = stage->priv;

    switch (prop_id) {
    case PROP_PROJECTION_MATRIX:
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
graphics_layer_stage_dispose (GObject *object)
{
    GraphicsLayerStage* layer = GRAPHICS_LAYER_STAGE(object);
    GraphicsLayerStagePrivate* priv = layer->priv;

    if (priv->perspectiveMatrix)
        cogl_matrix_free(priv->perspectiveMatrix);

    G_OBJECT_CLASS (graphics_layer_stage_parent_class)->dispose (object);
}

#define APPLY_DEFAULT_PERSPECTIVE_MATRIX_TO_STAGE 0 
static void graphics_layer_stage_apply_transform(ClutterActor *actor, CoglMatrix *matrix)
{
    LOG(AcceleratedCompositing, "%s(%p) %s", __func__, actor, clutter_actor_get_name(actor));

    GraphicsLayerStagePrivate* priv = GRAPHICS_LAYER_STAGE(actor)->priv;

#if APPLY_DEFAULT_PERSPECTIVE_MATRIX_TO_STAGE
    CLUTTER_ACTOR_CLASS(graphics_layer_stage_parent_class)->apply_transform(actor, matrix);
#else
   CoglMatrix projectionMatrix;
    cogl_matrix_init_identity(&projectionMatrix);

    // FIXME: These parameters should be calculated by fomulas.
    //        Refer to cogl_matrix_view_2d_in_perspective.
    cogl_matrix_ortho(&projectionMatrix, -1.0, 1.0, -1.0, 1.0, -1.0, 5);
    cogl_matrix_scale(&projectionMatrix, 0.001 ,0.001 , 0);
    cogl_matrix_translate(&projectionMatrix, -512, 385, 0);

    CoglMatrix reflectionMatrix;
    cogl_matrix_init_identity(&reflectionMatrix);
    reflectionMatrix.yy = -1;
    cogl_matrix_multiply(&projectionMatrix, &projectionMatrix, &reflectionMatrix);

    // Apply a projection matrix.
    cogl_matrix_init_identity(matrix);
    cogl_matrix_multiply(matrix, matrix, &projectionMatrix);
#endif
}

static void
graphics_layer_stage_paint(ClutterActor *actor)
{
#if ENABLE(WEBGL)
    GraphicsContext3DPrivate::popGraphicsContext();
#endif
    CLUTTER_ACTOR_CLASS (graphics_layer_stage_parent_class)->paint (actor);
}

static void
graphics_layer_stage_class_init (GraphicsLayerStageClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    GParamSpec *pspec;
    
    object_class->get_property = graphics_layer_stage_get_property;
    object_class->set_property = graphics_layer_stage_set_property;
    object_class->dispose = graphics_layer_stage_dispose;
    actor_class->apply_transform = graphics_layer_stage_apply_transform;
    actor_class->paint = graphics_layer_stage_paint;
    g_type_class_add_private (klass, sizeof (GraphicsLayerStagePrivate));
}

static void
graphics_layer_stage_init (GraphicsLayerStage *self)
{
    GraphicsLayerStagePrivate* priv = self->priv = GRAPHICS_LAYER_STAGE_GET_PRIVATE(self);

    clutter_actor_set_reactive(CLUTTER_ACTOR(self), TRUE);

    priv->projectionMode = DefaultPerspectiveProjection;
}

GraphicsLayerStage*
graphics_layer_stage_new()
{
    GraphicsLayerStage* layer = GRAPHICS_LAYER_STAGE(g_object_new (GRAPHICS_LAYER_TYPE_STAGE, NULL));
    GraphicsLayerStagePrivate* priv = layer->priv;

    return layer;
}

void graphics_layer_stage_set_projection_mode(GraphicsLayerStage* stage, StageProjectionMode projectionMode)
{
    GraphicsLayerStagePrivate* priv = stage->priv;
    priv->projectionMode = projectionMode;

}

void graphics_layer_stage_set_perspective_matrix(GraphicsLayerStage* stage, const CoglMatrix* matrix) 
{
    LOG(AcceleratedCompositing, "%s actor = %p %s", __func__, stage, clutter_actor_get_name(CLUTTER_ACTOR(stage)));

    GraphicsLayerStagePrivate* priv = stage->priv;

    if (!matrix) {
        priv->perspectiveMatrix = 0;
        priv->projectionMode = OrthographicProjection; 
        return;
    }

    LOG(AcceleratedCompositing, "Matrix: %s", matrixToString(matrix));
    
    CoglMatrix identity;
    cogl_matrix_init_identity(&identity);
    if (cogl_matrix_equal((CoglMatrix*)&identity, (CoglMatrix*)matrix)) {
        LOG(AcceleratedCompositing, "Ignore identity matrix");
        return;
    }

    if (priv->perspectiveMatrix) 
        cogl_matrix_free(priv->perspectiveMatrix);

    priv->perspectiveMatrix = cogl_matrix_copy(matrix);
    priv->projectionMode = CustomPerspectiveProjection; 
}

#endif // USE(ACCELERATED_COMPOSITING)
