/*
 * ClutterSocket: An XEmbed server for Clutter.
 *
 * Copyrigth (C) 2010 Collabora Ltd
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include "clutter-socket.h"

/* Definitions copied from Gtk's xembed.h header. */
/* XEMBED messages */
typedef enum {
  XEMBED_EMBEDDED_NOTIFY        = 0,
  XEMBED_WINDOW_ACTIVATE        = 1,
  XEMBED_WINDOW_DEACTIVATE      = 2,
  XEMBED_REQUEST_FOCUS          = 3,
  XEMBED_FOCUS_IN               = 4,
  XEMBED_FOCUS_OUT              = 5,
  XEMBED_FOCUS_NEXT             = 6,
  XEMBED_FOCUS_PREV             = 7,
  XEMBED_GRAB_KEY               = 8,
  XEMBED_UNGRAB_KEY             = 9,
  XEMBED_MODALITY_ON            = 10,
  XEMBED_MODALITY_OFF           = 11,

/* Non standard messages*/
  XEMBED_GTK_GRAB_KEY           = 108,
  XEMBED_GTK_UNGRAB_KEY         = 109
} XEmbedMessageType;

/* Details for  XEMBED_FOCUS_IN: */
#define XEMBED_FOCUS_CURRENT             0
#define XEMBED_FOCUS_FIRST               1
#define XEMBED_FOCUS_LAST                2
/* From Tk */
#define EMBEDDED_APP_WANTS_FOCUS NotifyNormal+20

/* Flags for XEMBED_FOCUS_IN, XEMBED_FOCUS_NEXT, XEMBED_FOCUS_PREV */
#define XEMBED_FOCUS_WRAPAROUND         (1 << 0)

/* Flags for _XEMBED_INFO */
#define XEMBED_MAPPED                   (1 << 0)

/* Latest version we implement */
#define CLUTTER_XEMBED_PROTOCOL_VERSION 1

/* Forward declarations. */
static void clutter_socket_realize (ClutterActor *socket);
static void clutter_socket_unrealize (ClutterActor *socket);
static void clutter_socket_destroy (ClutterActor *socket);
static ClutterX11FilterReturn clutter_socket_x11_filter (XEvent *xev,
                                                         ClutterEvent *cev,
                                                         gpointer data);
static void handle_xembed_message (ClutterSocket     *socket,
                                   XEmbedMessageType message,
                                   glong             detail,
                                   glong             data1,
                                   glong             data2,
                                   guint32           time);
static void _xembed_send_message (Window recipient,
                                  XEmbedMessageType message,
                                  glong detail,
                                  glong data1,
                                  glong data2);
static void _clutter_socket_windowing_send_configure_event (ClutterSocket *socket);
static void clutter_socket_get_preferred_width (ClutterActor *actor,
                                                gfloat for_height,
                                                gfloat *min_width_p,
                                                gfloat *natural_width_p);
static void clutter_socket_get_preferred_height (ClutterActor *actor,
                                                 gfloat for_width,
                                                 gfloat *min_height_p,
                                                 gfloat *natural_height_p);
static void clutter_socket_allocate (ClutterActor *actor,
                                     const ClutterActorBox *box,
                                     ClutterAllocationFlags flags);
static void clutter_socket_key_focus_in  (ClutterActor *actor);
static void clutter_socket_key_focus_out (ClutterActor *actor);
static gboolean clutter_socket_key_event (ClutterActor *actor,
                                          ClutterKeyEvent *event);

/* Local data. */
struct _ClutterSocketPrivate
{
    Window own_window;
    Window plug_window;

    guint resize_count;

    guint16 request_min_width;
    guint16 request_min_height;
    guint16 request_base_width;
    guint16 request_base_height;
    guint16 current_width;
    guint16 current_height;
    guint16 current_x;
    guint16 current_y;

    gshort xembed_version;
    guint focus_in : 1;
    guint have_size : 1;
    guint need_map : 1;
    guint is_mapped : 1;
    guint active : 1;
    guint created_own_window : 1;
};

enum {
    PLUG_ADDED = 0,
    PLUG_REMOVED,
    LAST_SIGNAL
};

static guint socket_signals[LAST_SIGNAL] = { 0 };

static Atom xembed_atom = None;
static Atom xembed_info_atom = None;
static Atom wm_normal_hints_atom = None;

G_DEFINE_TYPE (ClutterSocket, clutter_socket, CLUTTER_TYPE_ACTOR);

#define CLUTTER_SOCKET_GET_PRIVATE(obj) \
(G_TYPE_INSTANCE_GET_PRIVATE ((obj), CLUTTER_TYPE_SOCKET, ClutterSocketPrivate))


static void
clutter_socket_class_init (ClutterSocketClass *klass)
{
    GObjectClass      *gobject_class = G_OBJECT_CLASS (klass);
    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);

    actor_class->realize = clutter_socket_realize;
    actor_class->unrealize = clutter_socket_unrealize;
    actor_class->destroy = clutter_socket_destroy;
    actor_class->get_preferred_width = clutter_socket_get_preferred_width;
    actor_class->get_preferred_height = clutter_socket_get_preferred_height;
    actor_class->allocate = clutter_socket_allocate;
    actor_class->key_focus_in = clutter_socket_key_focus_in;
    actor_class->key_focus_out = clutter_socket_key_focus_out;
    actor_class->key_press_event = clutter_socket_key_event;
    actor_class->key_release_event = clutter_socket_key_event;

    g_type_class_add_private (gobject_class, sizeof (ClutterSocketPrivate));

    /* Signals */
    /**
     * ClutterSocket::plug-added:
     * @socket: the object which received the signal
     *
     * This signal is emitted when a client is successfully
     * added to the socket.
     */
    socket_signals[PLUG_ADDED] =
      g_signal_new (g_intern_static_string("plug-added"),
                    G_OBJECT_CLASS_TYPE (klass),
                    G_SIGNAL_RUN_LAST,
                    G_STRUCT_OFFSET (ClutterSocketClass, plug_added),
                    NULL, NULL,
                    g_cclosure_marshal_VOID__VOID,
                    G_TYPE_NONE, 0);

    /**
     * ClutterSocket::plug-removed:
     * @socket: the object which received the signal
     *
     * This signal is emitted when a client is removed from the socket.
     */
    socket_signals[PLUG_REMOVED] =
      g_signal_new (g_intern_static_string("plug-removed"),
                    G_OBJECT_CLASS_TYPE (klass),
                    G_SIGNAL_RUN_LAST,
                    G_STRUCT_OFFSET (ClutterSocketClass, plug_removed),
                    NULL, NULL,
                    g_cclosure_marshal_VOID__VOID,
                    G_TYPE_NONE, 0);

    /* Atoms */
    if (xembed_atom == None)
        xembed_atom = XInternAtom(clutter_x11_get_default_display (), "_XEMBED", False);
    if (xembed_info_atom == None)
        xembed_info_atom = XInternAtom(clutter_x11_get_default_display (), "_XEMBED_INFO", False);
    if (wm_normal_hints_atom == None)
        wm_normal_hints_atom = XInternAtom(clutter_x11_get_default_display (), "WM_NORMAL_HINTS", False);
}

static void
clutter_socket_init (ClutterSocket *socket)
{
    ClutterSocketPrivate *priv;

    socket->priv = priv = CLUTTER_SOCKET_GET_PRIVATE (socket);
}

ClutterActor*
clutter_socket_new (void)
{
  return g_object_new (CLUTTER_TYPE_SOCKET, NULL);
}

ClutterActor*
clutter_socket_new_for_window (Window xid)
{
    ClutterSocket *socket = (ClutterSocket *)g_object_new (CLUTTER_TYPE_SOCKET, NULL);
    ClutterSocketPrivate *priv = CLUTTER_SOCKET_GET_PRIVATE (socket);

    priv->own_window = xid;
    priv->created_own_window = FALSE;

    return CLUTTER_ACTOR (socket);
}

static void
_clutter_socket_windowing_size_request (ClutterSocket *socket)
{
    ClutterSocketPrivate *priv = CLUTTER_SOCKET_GET_PRIVATE (socket);
    XSizeHints hints;
    long supplied;

    clutter_x11_trap_x_errors ();

    priv->request_min_width = 1;
    priv->request_min_height = 1;
    priv->request_base_width = 1;
    priv->request_base_height = 1;

    if (XGetWMNormalHints (clutter_x11_get_default_display (),
                           priv->plug_window, &hints, &supplied))
    {
        if (hints.flags & PMinSize)
        {
            priv->request_min_width = MAX (hints.min_width, 1);
            priv->request_min_height = MAX (hints.min_height, 1);
        }
        if (hints.flags & PBaseSize)
        {
            priv->request_base_width = MAX (hints.base_width, 1);
            priv->request_base_height = MAX (hints.base_height, 1);
        }
    }
    priv->have_size = TRUE;
    clutter_x11_untrap_x_errors ();
}

static void
clutter_socket_get_preferred_width (ClutterActor *actor,
                                    gfloat for_height,
                                    gfloat *min_width_p,
                                    gfloat *natural_width_p)
{
    ClutterSocket *socket = CLUTTER_SOCKET (actor);
    ClutterSocketPrivate *priv = CLUTTER_SOCKET_GET_PRIVATE (socket);

    if (priv->is_mapped && !priv->have_size && priv->plug_window)
    {
        _clutter_socket_windowing_size_request (socket);
    }

    if (priv->is_mapped && priv->have_size)
    {
        *min_width_p = priv->request_min_width;
        *natural_width_p = priv->request_base_width;
    }
    else
    {
        *min_width_p = *natural_width_p = 1.0;
    }
}

static void
clutter_socket_get_preferred_height (ClutterActor *actor,
                                     gfloat for_width,
                                     gfloat *min_height_p,
                                     gfloat *natural_height_p)
{
    ClutterSocket *socket = CLUTTER_SOCKET (actor);
    ClutterSocketPrivate *priv = CLUTTER_SOCKET_GET_PRIVATE (socket);

    if (priv->is_mapped && !priv->have_size && priv->plug_window)
    {
        _clutter_socket_windowing_size_request (socket);
    }

    if (priv->is_mapped && priv->have_size)
    {
        *min_height_p = priv->request_min_height;
        *natural_height_p = priv->request_base_height;
    }
    else
    {
        *min_height_p = *natural_height_p = 1.0;
    }
}

static void
clutter_socket_allocate (ClutterActor *actor,
                         const ClutterActorBox *box,
                         ClutterAllocationFlags flags)
{
    ClutterSocket *socket = CLUTTER_SOCKET (actor);
    ClutterSocketPrivate *priv = CLUTTER_SOCKET_GET_PRIVATE (socket);
    Display *display = clutter_x11_get_default_display ();
    gfloat origin_x, origin_y, available_width, available_height;

    CLUTTER_ACTOR_CLASS (clutter_socket_parent_class)->allocate (actor, box, flags);

    if (priv->own_window)
    {
        clutter_actor_box_get_origin (box, &origin_x, &origin_y);
        clutter_actor_box_get_size (box, &available_width, &available_height);

        clutter_x11_trap_x_errors ();

        if (priv->created_own_window)
        {
            ClutterVertex origin = {origin_x, origin_y, 0};
            ClutterVertex relative_origin;
            gfloat clipped_width = available_width;
            gfloat clipped_height = available_height;
            if (origin.x < 0)
                origin.x = 0;
            if (origin.y < 0)
                origin.y = 0;
            clutter_actor_apply_relative_transform_to_point(
                    clutter_actor_get_parent(actor), NULL /* Use the stage */,
                    &origin, &relative_origin);

            if (clutter_actor_has_clip (actor))
            {
                gfloat xoff, yoff;
                clutter_actor_get_clip(actor, &xoff, &yoff,
                                              &clipped_width, &clipped_height);
            }
            clipped_width = (clipped_width < 1)? 1 : clipped_width;
            clipped_height = (clipped_height < 1)? 1 : clipped_height;
            XMoveResizeWindow (display, priv->own_window,
                               relative_origin.x, relative_origin.y,
                               clipped_width, clipped_height);
        }

        if (priv->plug_window)
        {
            if (available_width != priv->current_width ||
                available_height != priv->current_height ||
                origin_x != priv->current_x ||
                origin_y != priv->current_y)
            {
                gfloat plugin_x = (origin_x > 0)? 0 : origin_x;
                gfloat plugin_y = (origin_y > 0)? 0 : origin_y;
                XMoveResizeWindow (display, priv->plug_window,
                                   plugin_x, plugin_y,
                                   available_width, available_height);
                if (priv->resize_count)
                    priv->resize_count--;
                priv->current_width = available_width;
                priv->current_height = available_height;
                priv->current_x = origin_x;
                priv->current_y = origin_y;
            }

            if (priv->need_map)
            {
                XMapRaised (display, priv->plug_window);
                priv->need_map = FALSE;
            }

            while (priv->resize_count)
            {
                _clutter_socket_windowing_send_configure_event (socket);
                priv->resize_count--;
            }
        }

        clutter_x11_untrap_x_errors ();
    }
}

Window
clutter_socket_get_own_window(ClutterSocket *socket)
{
    return socket->priv->own_window;
}

static gboolean
_clutter_socket_embed_get_info (ClutterSocket *socket)
{
    ClutterSocketPrivate *priv;
    priv = CLUTTER_SOCKET_GET_PRIVATE (CLUTTER_SOCKET(socket));
    Display *display = clutter_x11_get_default_display ();
    Atom type;
    int format;
    unsigned long nitems, bytes_after;
    unsigned char *data;
    unsigned long *data_long;
    int status;

    priv->xembed_version = -1;

    status = XGetWindowProperty (display, priv->plug_window, xembed_info_atom,
                                 0, 2, False,
                                 xembed_info_atom, &type, &format,
                                 &nitems, &bytes_after, &data);
    if (status != Success)
        return FALSE;

    if (type == None)
        return FALSE;

    if (type != xembed_info_atom) {
        g_warning ("_XEMBED_INFO property has wrong type\n");
        return FALSE;
    }
    if (nitems < 2) {
        g_warning ("_XEMBED_INFO too short\n");
        XFree (data);
        return FALSE;
    }

    data_long = (unsigned long *)data;
    priv->xembed_version = MIN (CLUTTER_XEMBED_PROTOCOL_VERSION, data_long[0]);
    priv->is_mapped = data_long[1] & XEMBED_MAPPED;

    XFree (data);
    return TRUE;
}

static void
_clutter_socket_add_window (ClutterSocket *socket,
                            Window xid,
                            gboolean need_reparent)
{
    ClutterSocketPrivate *priv;
    priv = CLUTTER_SOCKET_GET_PRIVATE (CLUTTER_SOCKET(socket));
    Display *display = clutter_x11_get_default_display ();

    priv->plug_window = xid;

    if (need_reparent)
    {
        XReparentWindow (display, xid, priv->own_window, 0, 0);
    }

    XSelectInput (display, xid, StructureNotifyMask | PropertyChangeMask);
    priv->have_size = FALSE;
    _clutter_socket_embed_get_info (socket);
    priv->need_map = priv->is_mapped;

    _xembed_send_message (xid, XEMBED_EMBEDDED_NOTIFY, 0, priv->own_window, priv->xembed_version);

    clutter_actor_queue_relayout (CLUTTER_ACTOR (socket));

    g_signal_emit (socket, socket_signals[PLUG_ADDED], 0);
}

static void
_clutter_socket_end_embedding (ClutterSocket *socket,
                               gboolean reparent,
                               gboolean send_signal)
{
    ClutterSocketPrivate *priv = CLUTTER_SOCKET_GET_PRIVATE (socket);
    if (reparent && priv->plug_window)
    {
        clutter_x11_trap_x_errors ();
        XReparentWindow (clutter_x11_get_default_display (),
                         priv->plug_window,
                         clutter_x11_get_root_window (),
                         0, 0);
        XSync (clutter_x11_get_default_display (), False);
        clutter_x11_untrap_x_errors ();
    }
    priv->plug_window = 0;
    priv->current_width = 0;
    priv->current_height = 0;
    priv->resize_count = 0;

    if (send_signal)
        g_signal_emit (socket, socket_signals[PLUG_REMOVED], 0);
}

static void
clutter_socket_realize (ClutterActor *socket)
{
    ClutterSocketPrivate *priv;
    priv = CLUTTER_SOCKET_GET_PRIVATE (CLUTTER_SOCKET(socket));

    if (CLUTTER_ACTOR_CLASS (clutter_socket_parent_class)->realize)
        CLUTTER_ACTOR_CLASS (clutter_socket_parent_class)->realize (socket);

    ClutterStage *stage = CLUTTER_STAGE(clutter_actor_get_stage (socket));
    if (!stage)
        return;
    clutter_x11_trap_x_errors ();
    if (!priv->own_window)
    {
        priv->own_window = XCreateSimpleWindow(clutter_x11_get_default_display (),
                                               clutter_x11_get_stage_window (stage),
                                               10, 10, 100, 100,
                                               0, 0, 0);
        priv->created_own_window = TRUE;
    }
    XSelectInput(clutter_x11_get_default_display (),
                 priv->own_window, FocusChangeMask |
                 SubstructureNotifyMask | SubstructureRedirectMask);
    XMapWindow(clutter_x11_get_default_display (), priv->own_window);
    clutter_x11_untrap_x_errors ();
    clutter_x11_add_filter (clutter_socket_x11_filter, socket);
}

static void
clutter_socket_unrealize (ClutterActor *socket)
{
    clutter_x11_remove_filter (clutter_socket_x11_filter, socket);
    _clutter_socket_end_embedding (CLUTTER_SOCKET (socket), TRUE, FALSE);

    if (CLUTTER_ACTOR_CLASS (clutter_socket_parent_class)->unrealize)
        CLUTTER_ACTOR_CLASS (clutter_socket_parent_class)->unrealize (socket);
}

static void
clutter_socket_destroy (ClutterActor *socket)
{
    ClutterSocketPrivate *priv = CLUTTER_SOCKET_GET_PRIVATE (CLUTTER_SOCKET (socket));
    clutter_x11_remove_filter (clutter_socket_x11_filter, socket);
    _clutter_socket_end_embedding (CLUTTER_SOCKET (socket), TRUE, FALSE);

    if (priv->created_own_window && priv->own_window)
    {
        XDestroyWindow (clutter_x11_get_default_display (),
                        priv->own_window);
    }

    if (CLUTTER_ACTOR_CLASS (clutter_socket_parent_class)->destroy)
        CLUTTER_ACTOR_CLASS (clutter_socket_parent_class)->destroy (socket);
}

static void
_clutter_socket_claim_focus (ClutterSocket *socket, gboolean send_event)
{
    ClutterSocketPrivate *priv = CLUTTER_SOCKET_GET_PRIVATE (socket);

    if (!send_event)
        priv->focus_in = TRUE;

    clutter_actor_grab_key_focus (CLUTTER_ACTOR (socket));
}

static void
clutter_socket_key_focus_in  (ClutterActor *actor)
{
    ClutterSocketPrivate *priv = CLUTTER_SOCKET_GET_PRIVATE (CLUTTER_SOCKET(actor));

    if (!priv->plug_window)
        return;

    if (!priv->focus_in)
    {
        priv->focus_in = TRUE;
        _xembed_send_message(priv->plug_window, XEMBED_FOCUS_IN,
                             XEMBED_FOCUS_CURRENT, 0, 0);
    }
}

static void
clutter_socket_key_focus_out (ClutterActor *actor)
{
    ClutterSocketPrivate *priv = CLUTTER_SOCKET_GET_PRIVATE (CLUTTER_SOCKET(actor));

    if (!priv->plug_window)
        return;

    if (priv->focus_in)
    {
        priv->focus_in = FALSE;
        _xembed_send_message(priv->plug_window, XEMBED_FOCUS_OUT,
                             0, 0, 0);
    }
}

static gboolean
clutter_socket_key_event (ClutterActor *actor,
                          ClutterKeyEvent *event)
{
    ClutterSocketPrivate *priv = CLUTTER_SOCKET_GET_PRIVATE (CLUTTER_SOCKET(actor));

    if (!priv->plug_window || !clutter_actor_has_key_focus (actor))
        return FALSE;

    XKeyEvent xkey;

    memset (&xkey, 0, sizeof(xkey));
    xkey.type = (event->type == CLUTTER_KEY_PRESS)? KeyPress : KeyRelease;
    xkey.window = priv->plug_window;
    xkey.root = clutter_x11_get_root_window ();
    xkey.subwindow = None;
    xkey.time = event->time;
    xkey.x = 0;
    xkey.y = 0;
    xkey.x_root = 0;
    xkey.y_root = 0;
    xkey.state = event->modifier_state;
    xkey.keycode = event->hardware_keycode;
    xkey.same_screen = True;/* FIXME ? */

    clutter_x11_trap_x_errors ();
    XSendEvent (clutter_x11_get_default_display (),
                priv->plug_window,
                False, NoEventMask,
                (XEvent *)&xkey);
    clutter_x11_untrap_x_errors ();

    return TRUE;
}

static ClutterX11FilterReturn
clutter_socket_x11_filter (XEvent *xev,
                           ClutterEvent *cev,
                           gpointer data)
{
    ClutterX11FilterReturn result = CLUTTER_X11_FILTER_CONTINUE;
    ClutterSocket *socket = CLUTTER_SOCKET (data);
    ClutterSocketPrivate *priv = CLUTTER_SOCKET_GET_PRIVATE (socket);
    Display *display = clutter_x11_get_default_display ();

    switch (xev->type)
    {
    case ClientMessage:
    {
        if (xev->xclient.message_type == xembed_atom) {
            handle_xembed_message (socket,
                                   xev->xclient.data.l[1],
                                   xev->xclient.data.l[2],
                                   xev->xclient.data.l[3],
                                   xev->xclient.data.l[4],
                                   xev->xclient.data.l[0]);
            result = CLUTTER_X11_FILTER_REMOVE;
        }
        break;
    }
    case CreateNotify:
    {
        XCreateWindowEvent *xcwe = &xev->xcreatewindow;

        if (!priv->plug_window &&
            priv->own_window == xcwe->parent)
        {
            _clutter_socket_add_window (socket, xcwe->window, FALSE);
        }

        result = CLUTTER_X11_FILTER_REMOVE;
        break;
    }
    case ConfigureRequest:
    {
        XConfigureRequestEvent *xcre = &xev->xconfigurerequest;

        if (!priv->plug_window)
            _clutter_socket_add_window (socket, xcre->window, FALSE);

        if (priv->plug_window)
        {
            if (xcre->value_mask & (CWWidth | CWHeight))
            {
                priv->resize_count++;
                clutter_actor_queue_relayout (CLUTTER_ACTOR (socket));
            }
            else if (xcre->value_mask & (CWX | CWY))
            {
                _clutter_socket_windowing_send_configure_event (socket);
            }

            result = CLUTTER_X11_FILTER_REMOVE;
        }

        break;
    }
    case DestroyNotify:
    {
        XDestroyWindowEvent *xdwe = &xev->xdestroywindow;
        if (priv->plug_window &&
            priv->plug_window == xdwe->window)
        {
            _clutter_socket_end_embedding (socket, FALSE, TRUE);
            result = CLUTTER_X11_FILTER_REMOVE;
        }
        else if (priv->own_window == xdwe->window)
        {
            _clutter_socket_end_embedding (socket, TRUE, FALSE);
        }
        break;
    }
    case FocusIn:
        if (xev->xfocus.window == priv->plug_window)
        {
            if (xev->xfocus.mode == EMBEDDED_APP_WANTS_FOCUS)
            {
                _clutter_socket_claim_focus (socket, TRUE);
            }
            result = CLUTTER_X11_FILTER_REMOVE;
        }
        break;
    case FocusOut:
        if (xev->xfocus.window == priv->plug_window)
            result = CLUTTER_X11_FILTER_REMOVE;
        break;
    case MapRequest:
    {
        if (!priv->plug_window)
        {
            _clutter_socket_add_window (socket, xev->xmaprequest.window, FALSE);
        }

        if (priv->plug_window)
        {
            if (!priv->is_mapped)
            {
                priv->is_mapped = TRUE;
                priv->need_map = TRUE;
                clutter_actor_queue_relayout (CLUTTER_ACTOR (socket));
            }
            result = CLUTTER_X11_FILTER_REMOVE;
        }
        break;
    }
    case PropertyNotify:
        if (priv->plug_window && priv->plug_window == xev->xproperty.window)
        {
            if (xev->xproperty.atom == wm_normal_hints_atom)
            {
                priv->have_size = FALSE;
                clutter_actor_queue_relayout (CLUTTER_ACTOR(socket));
                result = CLUTTER_X11_FILTER_REMOVE;
            }
            else if (xev->xproperty.atom == xembed_info_atom)
            {
                gboolean was_mapped = priv->is_mapped;
                if (_clutter_socket_embed_get_info(socket))
                {
                    if (was_mapped != priv->is_mapped)
                    {
                        if (priv->is_mapped)
                        {
                            priv->need_map = TRUE;
                        }
                        else
                        {
                            clutter_x11_trap_x_errors ();
                            XMapRaised (display, priv->plug_window);
                            clutter_x11_untrap_x_errors ();
                        }
                        clutter_actor_queue_relayout (CLUTTER_ACTOR (socket));
                    }
                }
            }
        }
        break;
    case ReparentNotify:
    {
        XReparentEvent *xre = &xev->xreparent;
        if (!priv->plug_window && xre->parent == priv->own_window)
        {
            _clutter_socket_add_window(socket, xre->window, FALSE);
            result = CLUTTER_X11_FILTER_REMOVE;
        }
        else if (priv->plug_window &&
                 xre->window == priv->plug_window &&
                 xre->parent != priv->own_window)
        {
            _clutter_socket_end_embedding (socket, FALSE, TRUE);
        }
        break;
    }
    case UnmapNotify:
    {
        if (priv->plug_window &&
            xev->xunmap.window == priv->plug_window)
        {
            if (priv->is_mapped)
            {
                priv->is_mapped = FALSE;
                clutter_actor_queue_relayout (CLUTTER_ACTOR (socket));
            }
            result = CLUTTER_X11_FILTER_REMOVE;
        }
        break;
    }
    }
    return result;
}

static const char *
_clutter_xembed_message_name (XEmbedMessageType message)
{
  static char unk[24];

  switch (message)
    {
#define CASE(x) case XEMBED_##x: return "XEMBED_"#x
      CASE (EMBEDDED_NOTIFY);
      CASE (WINDOW_ACTIVATE);
      CASE (WINDOW_DEACTIVATE);
      CASE (REQUEST_FOCUS);
      CASE (FOCUS_IN);
      CASE (FOCUS_OUT);
      CASE (FOCUS_NEXT);
      CASE (FOCUS_PREV);
      CASE (GRAB_KEY);
      CASE (UNGRAB_KEY);
      CASE (MODALITY_ON);
      CASE (MODALITY_OFF);
      CASE (GTK_GRAB_KEY);
      CASE (GTK_UNGRAB_KEY);
#undef CASE
    default:
      snprintf (unk, 24, "UNKNOWN(%d)", message);
      return unk;
    }
}

static void
handle_xembed_message (ClutterSocket     *socket,
                       XEmbedMessageType message,
                       glong             detail,
                       glong             data1,
                       glong             data2,
                       guint32           time)
{
    switch (message)
    {
    case XEMBED_EMBEDDED_NOTIFY:
    case XEMBED_WINDOW_ACTIVATE:
    case XEMBED_WINDOW_DEACTIVATE:
    case XEMBED_MODALITY_ON:
    case XEMBED_MODALITY_OFF:
    case XEMBED_FOCUS_IN:
    case XEMBED_FOCUS_OUT:
        g_warning ("ClutterSocket: Invalid _XEMBED message %s received", _clutter_xembed_message_name (message));
        break;

    case XEMBED_REQUEST_FOCUS:
        _clutter_socket_claim_focus (socket, FALSE);
        break;

    case XEMBED_FOCUS_NEXT:
    case XEMBED_FOCUS_PREV:
    case XEMBED_GTK_GRAB_KEY:
    case XEMBED_GTK_UNGRAB_KEY:
    case XEMBED_GRAB_KEY:
    case XEMBED_UNGRAB_KEY:
        g_warning ("ClutterSocket: Ignoring _XEMBED message %s received", _clutter_xembed_message_name (message));
        break;

    default:
        g_warning ("ClutterSocket: Unknown _XEMBED message of type %d", message);
        break;
    }
}

static void
_xembed_send_message (Window recipient,
                      XEmbedMessageType message,
                      glong detail,
                      glong data1,
                      glong data2)
{
    Display *display = clutter_x11_get_default_display ();
    XClientMessageEvent xclient;

    memset (&xclient, 0, sizeof (xclient));
    xclient.window = recipient;
    xclient.type = ClientMessage;
    xclient.message_type = xembed_atom;
    xclient.format = 32;
    xclient.data.l[0] = clutter_get_current_event_time ();
    xclient.data.l[1] = message;
    xclient.data.l[2] = detail;
    xclient.data.l[3] = data1;
    xclient.data.l[4] = data2;

    clutter_x11_trap_x_errors ();
    XSendEvent (display, recipient, False, NoEventMask, (XEvent *)&xclient);
    clutter_x11_untrap_x_errors ();
}

static void
_clutter_socket_windowing_send_configure_event (ClutterSocket *socket)
{
    ClutterSocketPrivate *priv = CLUTTER_SOCKET_GET_PRIVATE (socket);
    XConfigureEvent xconfigure;

    if (!priv->plug_window)
        return;

    memset (&xconfigure, 0, sizeof (xconfigure));
    xconfigure.type = ConfigureNotify;
    xconfigure.event = priv->plug_window;
    xconfigure.window = priv->plug_window;

    //FIXME: Get the real coordinates relative to root window
    xconfigure.x = 0;
    xconfigure.y = 0;
    xconfigure.width = priv->current_width;
    xconfigure.height = priv->current_height;

    xconfigure.border_width = 0;
    xconfigure.above = None;
    xconfigure.override_redirect = False;

    clutter_x11_trap_x_errors ();
    XSendEvent (clutter_x11_get_default_display (),
                priv->plug_window,
                False, NoEventMask, (XEvent *)&xconfigure);
    clutter_x11_untrap_x_errors ();
}
