/*
 * Copyright (C) 2007, 2008 Holger Hans Peter Freyther
 * Copyright (C) 2007, 2008 Christian Dywan <christian@imendio.com>
 * Copyright (C) 2008 Nuanti Ltd.
 * Copyright (C) 2008 Alp Toker <alp@atoker.com>
 * Copyright (C) 2008 Gustavo Noronha Silva <gns@gnome.org>
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
 * Copyright (C) 2010 Collabora Ltd.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

// X11 include workaround.
#include "webkitclutterinclude.h"

#include "config.h"
#include "ChromeClientGObject.h"

#include "Console.h"
#include "DumpRenderTreeSupportClutter.h"
#include "Element.h"
#include "FileSystem.h"
#include "FileChooser.h"
#include "FloatRect.h"
#include "FrameView.h"
#include "FrameLoadRequest.h"
#include "GraphicsLayerClutter.h"
#include "HTMLNames.h"
#include "IntRect.h"
#include "HitTestResult.h"
#include "Icon.h"
#include "KURL.h"
#include "PopupMenuClutter.h"
#include "SecurityOrigin.h"
#include "NotImplemented.h"
#include "TileClutter.h"
#include "UtilClutter.h"
#include "WindowFeatures.h"
#include "webkitgwebframe.h"
#include "webkitgwebframeprivate.h"
#include "webkitiwebview.h"
#include "webkitiwebviewprivate.h"
#include "webkitwebpageprivate.h"
#include <wtf/text/CString.h>

#if ENABLE(SQL_DATABASE)
#include "DatabaseManager.h"
#endif

#if ENABLE(VIDEO) && USE(NATIVE_FULLSCREEN_VIDEO)
#include "HTMLMediaElement.h"
#endif

#if USE(TILED_BACKING_STORE)
#include "TiledBackingStore.h"
#endif

#if USE(ACCELERATED_COMPOSITING)
#include "GraphicsLayer.h"
#endif


using namespace WebCore;

namespace WebKit {

ChromeClient::ChromeClient(WebKitWebPage* webPage)
    : m_webPage(webPage)
{
    ASSERT(m_webPage);
}

FloatRect ChromeClient::windowRect()
{
    WebKitIWebView* view = webkit_web_page_get_view(m_webPage);
    if (!view)
        return FloatRect();

    ClutterGeometry windowRect;
    webkit_iweb_view_get_window_rect(view, &windowRect);
    return FloatRect(windowRect.x, windowRect.y, windowRect.width, windowRect.height);
}

void ChromeClient::toolkitSetWindowRect(const FloatRect& rect)
{
    WebKitIWebView* view = webkit_web_page_get_view(m_webPage);
    if (!view)
        return;

    ClutterGeometry windowRect;
    windowRect.x = rect.x();
    windowRect.y = rect.y();
    windowRect.width = rect.width();
    windowRect.height = rect.height();
    webkit_iweb_view_set_window_rect(view, &windowRect);
}

FloatRect ChromeClient::pageRect()
{
    int width, height;
    webkit_web_page_get_viewport_size(m_webPage, &width, &height);

    return FloatRect(0, 0, width, height);
}

float ChromeClient::scaleFactor()
{
    notImplemented();
    return 1.0;
}

void ChromeClient::focus()
{
    notImplemented();
}

void ChromeClient::unfocus()
{
    notImplemented();
}

Page* ChromeClient::toolkitCreateWindow(Frame* frame, const FrameLoadRequest& frameLoadRequest, const WindowFeatures& coreFeatures)
{
    WebKitWebPage* newPage = 0;
    g_signal_emit_by_name(m_webPage, "create-window", kit(frame), &newPage);
    return core(newPage);
}

void ChromeClient::toolkitLoadURL(Page* page, const FrameLoadRequest& frameLoadRequest)
{
    notImplemented();
}

WebKitWebWindowFeatures* ChromeClient::toolkitWindowFeatures()
{
    return m_webPage->priv->webWindowFeatures.get();
}

void ChromeClient::toolkitSetWindowFeatures(WebKitWebWindowFeatures* features)
{
    notImplemented();
}

WebKitWebSettings* ChromeClient::toolkitSettings()
{
    return m_webPage->priv->webSettings.get();
}

void ChromeClient::show()
{
    notImplemented();
}

void ChromeClient::closeWindowSoon()
{
    notImplemented();
}

bool ChromeClient::canTakeFocus(FocusDirection)
{
    notImplemented();
    return true;
}

KeyboardUIMode ChromeClient::keyboardUIMode()
{
    return KeyboardAccessDefault;
}

void ChromeClient::invalidateRootView(const IntRect& updateRect, bool immediate)
{
#if USE(TILED_BACKING_STORE)
    if (platformPageClient()) {
        WebCore::TiledBackingStore* backingStore = core(m_webPage)->mainFrame()->tiledBackingStore();
        if (!backingStore)
            return;
        backingStore->invalidate(updateRect);
    }
#endif
}

void ChromeClient::invalidateContentsAndRootView(const IntRect& updateRect, bool immediate)
{
    ClutterGeometry rect = {updateRect.x(), updateRect.y(), updateRect.width(), updateRect.height()};
    if (!updateRect.isEmpty()) {
        WebKitIWebView* view = webkit_web_page_get_view(m_webPage);
        if (view)
            webkit_iweb_view_invalidate_rect(view, &rect);
    }
}

/* FIXME: This is not how scroll should work. This method is for platforms that
 * draw the whole page and then just move the viewport on it, not for platforms
 * that redraw the page when scrolling.
 */
void ChromeClient::scroll(const IntSize& delta, const IntRect& rectToScroll, const IntRect& clipRect)
{
    ClutterActor* actor = CLUTTER_ACTOR(webkit_web_page_get_view(m_webPage));
    if (!actor)
        return;

    ClutterGeometry updateRect;
    clutter_actor_get_geometry(CLUTTER_ACTOR(actor), &updateRect);

    webkit_iweb_view_invalidate_rect(WEBKIT_IWEB_VIEW(actor), &updateRect);
}

IntRect ChromeClient::rootViewToScreen(const IntRect& rect) const
{
    notImplemented();
    return IntRect();
}

IntPoint ChromeClient::screenToRootView(const IntPoint& point) const
{
    notImplemented();
    return IntPoint();
}

PlatformPageClient ChromeClient::platformPageClient() const
{
    return webkit_web_page_get_view(m_webPage);
}

void ChromeClient::contentsSizeChanged(Frame* frame, const IntSize& size) const
{
    WebKitWebFrame* webFrame = kit(frame);

    g_signal_emit_by_name(webFrame, "contents-size-changed", size.width(), size.height(), NULL);
}

void ChromeClient::scrollbarsModeDidChange() const
{
    notImplemented();
}

void ChromeClient::setToolTip(const String& toolTip, TextDirection)
{
    notImplemented();
}

void ChromeClient::print(Frame* frame)
{
    notImplemented();
}

#if ENABLE(SQL_DATABASE)
void ChromeClient::toolkitExceededDatabaseQuota(WebKitWebFrame *webFrame, WebKitWebDatabase* webDatabase)
{
    WebKitWebPage* webPage = webkit_web_frame_get_page(webFrame);
    g_signal_emit_by_name(webPage, "database-quota-exceeded", webFrame, webDatabase);
}
#endif

void ChromeClient::runOpenPanel(Frame*, PassRefPtr<FileChooser> prpFileChooser)
{
    notImplemented();
}

void ChromeClient::loadIconForFiles(const Vector<WTF::String>&, WebCore::FileIconLoader*)
{
    notImplemented();
}

void ChromeClient::setCursorHiddenUntilMouseMoves(bool)
{
    notImplemented();
}

bool ChromeClient::selectItemWritingDirectionIsNatural()
{
    return false;
}

bool ChromeClient::selectItemAlignmentFollowsMenuWritingDirection()
{
    return true;
}

bool ChromeClient::hasOpenedPopup() const
{
    notImplemented();
    return false;
}

PassRefPtr<WebCore::PopupMenu> ChromeClient::createPopupMenu(WebCore::PopupMenuClient* client) const
{
    WebKitIWebView* webView = webkit_web_page_get_view(m_webPage);
    if (webView)
        return adoptRef(new PopupMenuClutter(client, webkit_iweb_view_create_popup_menu(webView)));

    return 0;
}

PassRefPtr<WebCore::SearchPopupMenu> ChromeClient::createSearchPopupMenu(WebCore::PopupMenuClient* client) const
{
    notImplemented();
    return 0;
}

#if ENABLE(VIDEO) && USE(NATIVE_FULLSCREEN_VIDEO)
void ChromeClient::enterFullscreenForNode(Node* node)
{
    if (!node)
        return;

    HTMLElement* element = static_cast<HTMLElement*>(node);
    if (element && element->isMediaElement()) {
        HTMLMediaElement* mediaElement = static_cast<HTMLMediaElement*>(element);
        if (mediaElement->player() && mediaElement->player()->canEnterFullscreen())
            mediaElement->player()->enterFullscreen();
    }
}

void ChromeClient::exitFullscreenForNode(Node* node)
{
    if (!node)
        return;

    HTMLElement* element = static_cast<HTMLElement*>(node);
    if (element && element->isMediaElement()) {
        HTMLMediaElement* mediaElement = static_cast<HTMLMediaElement*>(element);
        if (mediaElement->player())
            mediaElement->player()->exitFullscreen();
    }
}
#endif

#if ENABLE(FULLSCREEN_API)
bool ChromeClient::supportsFullScreenForElement(const WebCore::Element* element, bool withKeyboard)
{
    if (withKeyboard)
        return false;

    return true;
}

void ChromeClient::enterFullScreenForElement(WebCore::Element* element)
{
    element->document()->webkitWillEnterFullScreenForElement(element);
    element->document()->webkitDidEnterFullScreenForElement(element);
}

void ChromeClient::exitFullScreenForElement(WebCore::Element* element)
{
    element->document()->webkitWillExitFullScreenForElement(element);
    element->document()->webkitDidExitFullScreenForElement(element);
}
#endif

#if USE(TILED_BACKING_STORE)
IntRect ChromeClient::visibleRectForTiledBackingStore() const
{
    WebKitWebFrame* webFrame = webkit_web_page_get_main_frame(m_webPage);
    Frame* frame = core(webFrame);
    if (!frame || !frame->view())
        return IntRect();

    int xOffset, yOffset;
    webkit_web_frame_get_scroll_offset(webFrame, &xOffset, &yOffset);

    IntSize size = frame->view()->frameRect().size();
    return IntRect(xOffset, yOffset, size.width(), size.height());
}

void ChromeClient::delegatedScrollRequested(const IntPoint& point)
{
    WebKitIWebView* view = webkit_web_page_get_view(m_webPage);
    webkit_iweb_view_delegated_scroll(view, point.x(), point.y());
}

void ChromeClient::tiledBackingStoreDidCreateTile(Tile* tile)
{
    TileClutter* clutterTile = static_cast<TileClutter*>(tile);

    WebKitIWebView* view = webkit_web_page_get_view(m_webPage);
    webkit_iweb_view_add_tile(view, clutterTile->actor());

    IntRect rect = clutterTile->rect();
    clutter_actor_set_position(clutterTile->actor(), rect.x(), rect.y());
    clutter_actor_set_size(clutterTile->actor(), rect.width(), rect.height());

    clutter_actor_show(clutterTile->actor());
}
#endif

#if USE(ACCELERATED_COMPOSITING)
void ChromeClient::attachRootGraphicsLayer(Frame* frame, GraphicsLayer* graphicsLayer)
{
    WebKitIWebView* view = platformPageClient(); 

    if (graphicsLayer)
        webkit_iweb_view_set_root_graphics_layer(view, static_cast<GraphicsLayerClutter*>(graphicsLayer));
    else
        webkit_iweb_view_set_root_graphics_layer(view, 0);
}

void ChromeClient::setNeedsOneShotDrawingSynchronization()
{
    // we want the layers to synchronize next time we update the screen anyway
    WebKitIWebView* view = platformPageClient(); 
    if (view)
        webkit_iweb_view_mark_for_sync(view);
}

void ChromeClient::scheduleCompositingLayerFlush()
{
    // we want the layers to synchronize ASAP
    if (platformPageClient())
        webkit_iweb_view_mark_for_sync(WEBKIT_IWEB_VIEW(platformPageClient()));
}

ChromeClient::CompositingTriggerFlags ChromeClient::allowedCompositingTriggers() const
{
    if (platformPageClient() && webkit_iweb_view_allows_accelerated_compositing(WEBKIT_IWEB_VIEW(platformPageClient())))
        return AllTriggers;

    return 0;
}
#endif

#if ENABLE(TOUCH_EVENTS)
void ChromeClient::needTouchEvents(bool needed)
{
    notImplemented();
}
#endif

}
