/*
 *  Copyright (C) Research In Motion Limited 2010. All rights reserved.
 *  Copyright (C) 2010 Joone Hur <joone@kldp.org>
 *  Copyright (C) 2009 Google Inc. All rights reserved.
 *  Copyright (C) 2011 Collabora Ltd.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"
#include "DumpRenderTreeSupportClutter.h"

#include "APICast.h"
#include "DumpRenderTreeSupportGObject.h"
#include "CSSComputedStyleDeclaration.h"
#include "Frame.h"
#include "FrameLoaderClientGObject.h"
#include "FrameView.h"
#include "GeolocationClientMock.h"
#include "GeolocationController.h"
#include "GeolocationError.h"
#include "GeolocationPosition.h"
#include "HTMLInputElement.h"
#include "JSCSSStyleDeclaration.h"
#include "JSDOMBinding.h"
#include "JSDocument.h"
#include "JSElement.h"
#include "JSLock.h"
#include "JSNodeList.h"
#include "JSRange.h"
#include "JSCJSValue.h"
#include "JSValueRef.h"
#include "NodeList.h"
#include "Operations.h"
#include "PageGroup.h"
#include "PrintContext.h"
#include "RenderTreeAsText.h"
#include "ScriptSourceCode.h"
#include "ScriptValue.h"
#include "webkit/webkitgwebframe.h"
#include "webkit/webkitgwebframeprivate.h"
#include "webkit/webkitwebpageprivate.h"

#include <stdio.h>

using namespace WebCore;
using namespace WebKit;

DRTSupportGObject* DumpRenderTreeSupportClutter::s_gobject = 0;

bool DumpRenderTreeSupportClutter::s_drtRun = false;
bool DumpRenderTreeSupportClutter::s_linksIncludedInTabChain = true;

WTF::HashMap<unsigned, RefPtr<WebCore::DOMWrapperWorld> > DumpRenderTreeSupportClutter::s_worldMap;

DumpRenderTreeSupportClutter::DumpRenderTreeSupportClutter()
{
}

DumpRenderTreeSupportClutter::~DumpRenderTreeSupportClutter()
{
}

void DumpRenderTreeSupportClutter::setDumpRenderTreeModeEnabled(bool enabled)
{
    s_drtRun = enabled;
}

bool DumpRenderTreeSupportClutter::dumpRenderTreeModeEnabled()
{
    return s_drtRun;
}

DRTSupportGObject* DumpRenderTreeSupportClutter::gobject()
{
    if (!s_gobject)
        s_gobject = drt_support_gobject_new();

    return s_gobject;
}

/**
 * getInnerText:
 * @frame: a #WebKitWebFrame
 *
 * Return value: inner text of @frame
 */
CString DumpRenderTreeSupportClutter::getInnerText(WebKitWebFrame* frame)
{
    g_return_val_if_fail(WEBKIT_IS_WEB_FRAME(frame), CString(""));

    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return CString("");

    FrameView* view = coreFrame->view();

    if (view && view->layoutPending())
        view->layout();

    Element* documentElement = coreFrame->document()->documentElement();
    return documentElement->innerText().utf8();
}

/**
 * getFrameChildren:
 * @frame: a #WebKitWebFrame
 *
 * Return value: child frames of @frame
 */
GSList* DumpRenderTreeSupportClutter::getFrameChildren(WebKitWebFrame* frame)
{
    g_return_val_if_fail(WEBKIT_IS_WEB_FRAME(frame), 0);

    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return 0;

    GSList* children = 0;
    for (Frame* child = coreFrame->tree()->firstChild(); child; child = child->tree()->nextSibling()) {
        FrameLoader* loader = child->loader();
        WebKit::FrameLoaderClient* client = static_cast<WebKit::FrameLoaderClient*>(loader->client());
        if (client)
          children = g_slist_append(children, client->webFrame());
    }

    return children;
}

CString DumpRenderTreeSupportClutter::dumpRenderTree(WebKitWebFrame* frame, bool printingMode)
{
    g_return_val_if_fail(WEBKIT_IS_WEB_FRAME(frame), CString(""));

    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return CString("");

    FrameView* view = coreFrame->view();

    if (view && view->layoutPending())
        view->layout();

    return externalRepresentation(coreFrame, printingMode ? RenderAsTextPrintingMode : RenderAsTextBehaviorNormal).utf8();
}

CString DumpRenderTreeSupportClutter::counterValueForElementById(WebKitWebFrame* frame, const char* id)
{
    g_return_val_if_fail(WEBKIT_IS_WEB_FRAME(frame), CString(""));

    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return CString();

    Element* coreElement = coreFrame->document()->getElementById(AtomicString(id));
    if (!coreElement)
        return CString();

    return counterValueForElement(coreElement).utf8();
}

unsigned int DumpRenderTreeSupportClutter::getPendingUnloadEventCount(WebKitWebFrame* frame)
{
    g_return_val_if_fail(WEBKIT_IS_WEB_FRAME(frame), 0);

    return WebKit::core(frame)->document()->domWindow()->pendingUnloadEventListeners();
}

unsigned int DumpRenderTreeSupportClutter::numberOfActiveAnimations(WebKitWebFrame* frame)
{
    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return 0;

    return coreFrame->animation()->numberOfActiveAnimations(coreFrame->document());
}

bool DumpRenderTreeSupportClutter::pauseAnimation(WebKitWebFrame* frame, const char* name, double time, const char* element)
{
    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return false;

    Element* coreElement = coreFrame->document()->getElementById(AtomicString(element));
    if (!coreElement || !coreElement->renderer())
        return false;

    return coreFrame->animation()->pauseAnimationAtTime(coreElement->renderer(), AtomicString(name), time);
}

bool DumpRenderTreeSupportClutter::pauseTransition(WebKitWebFrame* frame, const char* name, double time, const char* element)
{
    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return false;

    Element* coreElement = coreFrame->document()->getElementById(AtomicString(element));
    if (!coreElement || !coreElement->renderer())
        return false;

    return coreFrame->animation()->pauseTransitionAtTime(coreElement->renderer(), AtomicString(name), time);
}

bool DumpRenderTreeSupportClutter::elementDoesAutoCompleteForElementWithId(WebKitWebFrame* frame, const char* name)
{
    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return false;

    Element* coreElement = coreFrame->document()->getElementById(AtomicString(name));
    if (!coreElement || !coreElement->renderer())
        return false;

    HTMLInputElement* inputElement = static_cast<HTMLInputElement*>(coreElement);
    if (!inputElement)
        return false;

    return inputElement->isTextField() && !inputElement->isPasswordField() && inputElement->shouldAutocomplete();
}

CString DumpRenderTreeSupportClutter::plainText(JSContextRef context, JSValueRef rangeObject)
{
    JSC::ExecState* exec = toJS(context);
    Range* range = toRange(toJS(exec, rangeObject));
    if (!range)
        return CString();

    return range->text().utf8();
}

bool DumpRenderTreeSupportClutter::findString(WebKitWebFrame* frame, const char* target, GList* optionsList)
{
    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return false;

    WebCore::FindOptions options = 0;
    GList* option = g_list_first(optionsList);
    while (option) {
        gchar* optionName = static_cast<gchar*>(option->data);
        if (g_strcmp0(optionName, "CaseInsensitive") == 0)
            options |= WebCore::CaseInsensitive;
        else if (g_strcmp0(optionName, "AtWordStarts") == 0)
            options |= WebCore::AtWordStarts;
        else if (g_strcmp0(optionName, "TreatMedialCapitalAsWordStart") == 0)
            options |= WebCore::TreatMedialCapitalAsWordStart;
        else if (g_strcmp0(optionName, "Backwards") == 0)
            options |= WebCore::Backwards;
        else if (g_strcmp0(optionName, "WrapAround") == 0)
            options |= WebCore::WrapAround;
        else if (g_strcmp0(optionName, "StartInSelection") == 0)
            options |= WebCore::StartInSelection;

        option = option->next;
    }

    return coreFrame->editor()->findString(String(target), options);
}

void DumpRenderTreeSupportClutter::addUserScript(WebKitWebFrame* frame, const char* source, bool runAtStart, bool allFrames)
{
    g_return_if_fail(WEBKIT_IS_WEB_FRAME(frame));

    WebKitWebPage* page = webkit_web_frame_get_page(frame);
    Page* corePage = WebKit::core(page);
    if (!corePage)
        return;

    corePage->group().addUserScriptToWorld(mainThreadNormalWorld(), source, KURL(), Vector<String>(), Vector<String>(),
                                           runAtStart ? InjectAtDocumentStart : InjectAtDocumentEnd,
                                           allFrames ? InjectInAllFrames : InjectInTopFrameOnly);
}

void DumpRenderTreeSupportClutter::addUserStyleSheet(WebKitWebFrame* frame, const char* source, bool allFrames)
{
    g_return_if_fail(WEBKIT_IS_WEB_FRAME(frame));

    WebKitWebPage* page = webkit_web_frame_get_page(frame);
    Page* corePage = WebKit::core(page);
    if (!corePage)
        return;

    corePage->group().addUserStyleSheetToWorld(mainThreadNormalWorld(), source, KURL(), Vector<String>(), Vector<String>(),
                                               allFrames ? InjectInAllFrames : InjectInTopFrameOnly);
}

void DumpRenderTreeSupportClutter::removeAllUserContent(WebKitWebFrame* frame)
{
    g_return_if_fail(WEBKIT_IS_WEB_FRAME(frame));

    WebKitWebPage* page = webkit_web_frame_get_page(frame);
    Page* corePage = WebKit::core(page);
    if (!corePage)
        return;

    corePage->group().removeAllUserContent();
}

CString DumpRenderTreeSupportClutter::markerTextForListItem(WebKitWebFrame* frame, JSContextRef context, JSValueRef nodeObject)
{
    JSC::ExecState* exec = toJS(context);
    Element* element = toElement(toJS(exec, nodeObject));
    if (!element)
        return CString();

    return WebCore::markerTextForListItem(element).utf8();
}

void DumpRenderTreeSupportClutter::setLinksIncludedInFocusChain(bool enabled)
{
    s_linksIncludedInTabChain = enabled;
}

bool DumpRenderTreeSupportClutter::linksIncludedInFocusChain()
{
    return s_linksIncludedInTabChain;
}

JSValueRef DumpRenderTreeSupportClutter::computedStyleIncludingVisitedInfo(JSContextRef context, JSValueRef value)
{
    JSC::ExecState* exec = toJS(context);
    JSC::JSLockHolder lock(exec);

    if (!value)
        return JSValueMakeUndefined(context);

    JSC::JSValue jsValue = toJS(exec, value);
    if (!jsValue.inherits(&JSElement::s_info))
        return JSValueMakeUndefined(context);

    JSElement* jsElement = static_cast<JSElement*>(asObject(jsValue));
    Element* element = jsElement->impl();
    RefPtr<CSSComputedStyleDeclaration> style = CSSComputedStyleDeclaration::create(element, true);
    return toRef(exec, toJS(exec, jsElement->globalObject(), style.get()));
}

int DumpRenderTreeSupportClutter::pageNumberForElementById(WebKitWebFrame* frame, const char* id, float pageWidth, float pageHeight)
{
    g_return_val_if_fail(WEBKIT_IS_WEB_FRAME(frame), 0);

    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return -1;

    Element* coreElement = coreFrame->document()->getElementById(AtomicString(id));
    if (!coreElement)
        return -1;
    return PrintContext::pageNumberForElement(coreElement, FloatSize(pageWidth, pageHeight));
}

int DumpRenderTreeSupportClutter::numberOfPagesForFrame(WebKitWebFrame* frame, float pageWidth, float pageHeight)
{
    g_return_val_if_fail(WEBKIT_IS_WEB_FRAME(frame), 0);

    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return -1;

    return PrintContext::numberOfPages(coreFrame, FloatSize(pageWidth, pageHeight));
}

CString DumpRenderTreeSupportClutter::pageProperty(WebKitWebFrame* frame, const char* propertyName, int pageNumber)
{
    g_return_val_if_fail(WEBKIT_IS_WEB_FRAME(frame), CString());

    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return CString();

    return PrintContext::pageProperty(coreFrame, propertyName, pageNumber).utf8();
}

CString DumpRenderTreeSupportClutter::pageSizeAndMarginsInPixels(WebKitWebFrame* frame, int pageNumber, int width, int height, int marginTop, int marginRight, int marginBottom, int marginLeft)
{
    g_return_val_if_fail(WEBKIT_IS_WEB_FRAME(frame), CString());

    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return CString();

    return PrintContext::pageSizeAndMarginsInPixels(coreFrame, pageNumber, width, height, marginTop, marginRight, marginBottom, marginLeft).utf8();
}

void DumpRenderTreeSupportClutter::clearScriptWorlds()
{
    s_worldMap.clear();
}

void DumpRenderTreeSupportClutter::evaluateScriptInIsolatedWorld(WebKitWebFrame* frame, unsigned worldID, const char* script)
{
    g_return_if_fail(WEBKIT_IS_WEB_FRAME(frame));

    RefPtr<DOMWrapperWorld> world;
    if (!worldID)
        world = ScriptController::createWorld();
    else if (!s_worldMap.contains(worldID)) {
        world = ScriptController::createWorld();
        s_worldMap.add(worldID, world);
    } else
        world = s_worldMap.get(worldID);

    Frame* coreFrame = WebKit::core(frame);
    if (!coreFrame)
        return;

    ScriptSourceCode source(script);
    coreFrame->script()->evaluateInWorld(source, world.get());
}

void DumpRenderTreeSupportClutter::resetGeolocationClientMock(WebKitWebPage* webPage)
{
#if ENABLE(GEOLOCATION)
    GeolocationClientMock* mock = static_cast<GeolocationClientMock*>(GeolocationController::from(core(webPage))->client());
    mock->reset();
#endif
}

void DumpRenderTreeSupportClutter::setMockGeolocationPermission(WebKitWebPage* webPage, bool allowed)
{
#if ENABLE(GEOLOCATION)
    GeolocationClientMock* mock = static_cast<GeolocationClientMock*>(GeolocationController::from(core(webPage))->client());
    mock->setPermission(allowed);
#endif
}

void DumpRenderTreeSupportClutter::setMockGeolocationPosition(WebKitWebPage* webPage, double latitude, double longitude, double accuracy)
{
#if ENABLE(GEOLOCATION)
    GeolocationClientMock* mock = static_cast<GeolocationClientMock*>(GeolocationController::from(core(webPage))->client());

    double timestamp = g_get_real_time() / 1000000.0;
    mock->setPosition(GeolocationPosition::create(timestamp, latitude, longitude, accuracy));
#endif
}

void DumpRenderTreeSupportClutter::setMockGeolocationPositionUnavailableError(WebKitWebPage* webPage, const gchar* errorMessage)
{
#if ENABLE(GEOLOCATION)
    GeolocationClientMock* mock = static_cast<GeolocationClientMock*>(GeolocationController::from(core(webPage))->client());
    mock->setPositionUnavailableError(errorMessage);
#endif
}

int DumpRenderTreeSupportClutter::numberOfPendingGeolocationPermissionRequests(WebKitWebPage* webPage)
{
#if ENABLE(GEOLOCATION)
    GeolocationClientMock* mock = static_cast<GeolocationClientMock*>(GeolocationController::from(core(webPage))->client());
    return mock->numberOfPendingPermissionRequests();
#else
    return 0;
#endif
}
