/*
 *  Copyright (C) 2008 Gustavo Noronha Silva
 *  Copyright (C) 2010,2011 Collabora Ltd.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"
#include "InspectorClientClutter.h"
#undef None

#include "InspectorController.h"
#include "Node.h"
#include "NotImplemented.h"
#include "webkitwebinspectorprivate.h"
#include "webkitwebpageprivate.h"

using namespace WebCore;

namespace WebKit {

namespace {

class InspectorFrontendSettingsClutter : public InspectorFrontendClientLocal::Settings {
public:
    virtual ~InspectorFrontendSettingsClutter() { }

private:
    virtual String getProperty(const String&)
    {
        notImplemented();
        return String();
    }

    virtual void setProperty(const String&, const String&)
    {
        notImplemented();
    }
};

} // namespace

// FIXME - some (most or all?) of this will be shared with GTK+ through InspectorClientGObject
InspectorClient::InspectorClient(WebKitWebPage* webPage)
    : m_inspectedWebPage(webPage)
    , m_frontendPage(0)
    , m_frontendClient(0)
{}

InspectorClient::~InspectorClient()
{
    if (m_frontendClient) {
        m_frontendClient->disconnectInspectorClient();
        m_frontendClient = 0;
    }
}

void InspectorClient::inspectorDestroyed()
{
    delete this;
}

void InspectorClient::closeInspectorFrontend()
{
    if (m_frontendClient)
        m_frontendClient->destroyInspectorWindow(false);
}

void InspectorClient::bringFrontendToFront()
{
    m_frontendClient->bringToFront();
}

WebCore::InspectorFrontendChannel* InspectorClient::openInspectorFrontend(InspectorController*)
{
    // This g_object_get will ref the inspector. We're not doing an
    // unref if this method succeeds because the inspector object must
    // be alive even after the inspected WebPage is destroyed - the
    // close-window and destroy signals still need to be
    // emitted.
    WebKitWebInspector* webInspector = 0;
    g_object_get(m_inspectedWebPage, "web-inspector", &webInspector, NULL);
    ASSERT(webInspector);

    WebKitWebPage* inspectorWebPage = 0;
    g_signal_emit_by_name(m_inspectedWebPage, "inspection-requested", webInspector, &inspectorWebPage);

    if (!inspectorWebPage) {
        g_object_unref(webInspector);
        return 0;
    }

    webkit_web_inspector_set_web_page(webInspector, inspectorWebPage);

    GOwnPtr<gchar> inspectorPath(g_build_filename(inspectorFilesPath(), "inspector.html", NULL));
    GOwnPtr<gchar> inspectorURI(g_filename_to_uri(inspectorPath.get(), 0, 0));
    webkit_web_page_load_uri(inspectorWebPage, inspectorURI.get());

    m_frontendPage = core(inspectorWebPage);
    OwnPtr<InspectorFrontendClient> frontendClient = adoptPtr(new InspectorFrontendClient(m_inspectedWebPage, inspectorWebPage, webInspector, m_frontendPage, this));
    m_frontendClient = frontendClient.get();
    m_frontendPage->inspectorController()->setInspectorFrontendClient(frontendClient.release());

    // The inspector must be in it's own PageGroup to avoid deadlock while debugging.
    m_frontendPage->setGroupName("");

    return this;
}

void InspectorClient::highlight()
{
    notImplemented();
}

void InspectorClient::hideHighlight()
{
    notImplemented();
}

bool InspectorClient::sendMessageToFrontend(const String& message)
{
    return doDispatchMessageOnFrontendPage(m_frontendPage, message);
}

// FrontEndClient
static void inspectorPageDestroyed(ClutterActor* actor, InspectorFrontendClient* inspectorFrontendClient)
{
    inspectorFrontendClient->destroyInspectorWindow(true);
}

InspectorFrontendClient::InspectorFrontendClient(WebKitWebPage* inspectedWebPage, WebKitWebPage* inspectorWebPage, WebKitWebInspector* webInspector, WebCore::Page* inspectorPage, InspectorClient* inspectorClient)
    : InspectorFrontendClientLocal(core(inspectedWebPage)->inspectorController(), inspectorPage, adoptPtr(new InspectorFrontendSettingsClutter()))
    , m_inspectedWebPage(inspectedWebPage)
    , m_inspectorWebPage(inspectorWebPage)
    , m_webInspector(webInspector)
    , m_inspectorClient(inspectorClient)
{
    ClutterActor* inspectorActor = CLUTTER_ACTOR(webkit_web_page_get_view(m_inspectorWebPage.get()));
    g_signal_connect(inspectorActor, "destroy",
                     G_CALLBACK(inspectorPageDestroyed), this);
}

InspectorFrontendClient::~InspectorFrontendClient()
{
    if (m_inspectorClient) {
        m_inspectorClient->disconnectFrontendClient();
        m_inspectorClient = 0;
    }

    ASSERT(!m_webInspector);
}

void InspectorFrontendClient::destroyInspectorWindow(bool notifyInspectorController)
{
    if (!m_webInspector)
        return;

    WebKitWebInspector* webInspector = m_webInspector;
    m_webInspector = 0;

    if (notifyInspectorController)
        core(m_inspectedWebPage)->inspectorController()->disconnectFrontend();

    WebKitIWebView* view = webkit_web_page_get_view(m_inspectorWebPage.get());
    if (view)
        g_signal_handlers_disconnect_by_func(view, (gpointer)inspectorPageDestroyed, (gpointer)this);

    if (m_inspectorClient)
        m_inspectorClient->releaseFrontendPage();

    gboolean handled = FALSE;
    g_signal_emit_by_name(webInspector, "close-window", &handled);

    // Please do not use member variables here because InspectorFrontendClient object pointed by 'this'
    // has been implicitly deleted by "close-window" function.

    // we should now dispose our own references
    m_inspectorWebPage.clear();
    g_object_unref(webInspector);
}

String InspectorFrontendClient::localizedStringsURL()
{
    GOwnPtr<gchar> stringsPath(g_build_filename(m_inspectorClient->inspectorFilesPath(), "localizedStrings.js", NULL));
    GOwnPtr<gchar> stringsURI(g_filename_to_uri(stringsPath.get(), 0, 0));

    // FIXME: support l10n of localizedStrings.js
    return String::fromUTF8(stringsURI.get());
}

String InspectorFrontendClient::hiddenPanels()
{
    notImplemented();
    return String();
}

void InspectorFrontendClient::bringToFront()
{
    if (!m_inspectorWebPage)
        return;

    gboolean handled = FALSE;
    g_signal_emit_by_name(m_webInspector, "show-window", &handled);
}

void InspectorFrontendClient::closeWindow()
{
    destroyInspectorWindow(false);
}

void InspectorFrontendClient::disconnectFromBackend()
{
    destroyInspectorWindow(true);
}

void InspectorFrontendClient::attachWindow()
{
    if (!m_inspectorWebPage)
        return;

    gboolean handled = FALSE;
    g_signal_emit_by_name(m_webInspector, "attach-window", &handled);
}

void InspectorFrontendClient::detachWindow()
{
    if (!m_inspectorWebPage)
        return;

    gboolean handled = FALSE;
    g_signal_emit_by_name(m_webInspector, "detach-window", &handled);
}

void InspectorFrontendClient::setAttachedWindowHeight(unsigned)
{
    notImplemented();
}

void InspectorFrontendClient::inspectedURLChanged(const WTF::String&)
{
    notImplemented();
}

const char* InspectorClient::inspectorFilesPath()
{
    if (m_inspectorFilesPath)
        m_inspectorFilesPath.get();

    // FIXME: we need to get the version at build time instead of hardcoding 0 here.
    const char* environmentPath = getenv("WEBKIT_INSPECTOR_PATH");
    if (environmentPath && g_file_test(environmentPath, G_FILE_TEST_IS_DIR))
        m_inspectorFilesPath.set(g_strdup(environmentPath));
    else
#ifdef INSTALL_PREFIX
        m_inspectorFilesPath.set(g_build_filename(INSTALL_PREFIX, "share", "webkit_clutter-0", "webinspector", NULL));
#else
        m_inspectorFilesPath.set(g_build_filename(DATA_DIR, "webkit_clutter-0", "webinspector", NULL));
#endif

    return m_inspectorFilesPath.get();
}

}
