/*
 * Copyright (C) 2010 Igalia S.L.
 * Copyright (C) 2010 Collabora Ltd.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "test_utils.h"

#include <glib.h>
#include <glib/gstdio.h>
#include <clutter/clutter.h>
#include <webkit/webkitiwebview.h>
#include <webkit/webkitwebpage.h>
#include <webkit/vanilla/webkitwebactor.h>
#include <webkit/webkitdom.h>

extern void webkit_init();

#if GLIB_CHECK_VERSION(2, 16, 0)

#define HTML_DOCUMENT "<html><head><title>This is the title</title></head><body><p id='test'>test</p></body></html>"

typedef struct {
    ClutterActor* webActor;
    ClutterActor* window;
    WebKitWebPage* webPage;
    WebKitDOMDOMWindow* domWindow;
    GMainLoop* loop;

    gboolean loaded;
    gboolean clicked;
    gconstpointer data;
} DomDomviewFixture;

static gboolean finish_loading(DomDomviewFixture* fixture)
{
    if (g_main_loop_is_running(fixture->loop))
        g_main_loop_quit(fixture->loop);

    return FALSE;
}

static void dom_domview_fixture_setup(DomDomviewFixture* fixture, gconstpointer data)
{
    fixture->loop = g_main_loop_new(NULL, TRUE);
    fixture->window = clutter_stage_new();
    fixture->webActor = webkit_web_actor_new();
    fixture->webPage = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(fixture->webActor));
    fixture->data = data;

    clutter_container_add_actor(CLUTTER_CONTAINER(fixture->window), fixture->webActor);
}

static void dom_domview_fixture_teardown(DomDomviewFixture* fixture, gconstpointer data)
{
    clutter_actor_destroy(fixture->window);
    g_main_loop_unref(fixture->loop);
}

static gboolean loadedCallback(WebKitDOMDOMWindow* view, WebKitDOMEvent* event, DomDomviewFixture* fixture)
{
    g_assert(fixture->loaded == FALSE);
    fixture->loaded = TRUE;

    return FALSE;
}

static gboolean clickedCallback(WebKitDOMDOMWindow* view, WebKitDOMEvent* event, DomDomviewFixture* fixture)
{
    WebKitDOMEventTarget* target;
    gushort phase;

    g_assert(event);
    g_assert(WEBKIT_DOM_IS_EVENT(event));

    // We should catch this in the bubbling up phase, since we are connecting to the toplevel object
    phase = webkit_dom_event_get_event_phase(event);
    g_assert_cmpint(phase, ==, 3);

    target = webkit_dom_event_get_current_target(event);
    g_assert(target == WEBKIT_DOM_EVENT_TARGET(view));

    g_assert(fixture->clicked == FALSE);
    fixture->clicked = TRUE;

    finish_loading(fixture);

    return FALSE;
}

gboolean delayed_load_cb(DomDomviewFixture* fixture)
{
    webkit_iweb_view_load_string(WEBKIT_IWEB_VIEW(fixture->webActor), (const char*)fixture->data, NULL, NULL, NULL);

    return FALSE;
}

static void load_event_callback(WebKitWebPage* webPage, GParamSpec* spec, DomDomviewFixture* fixture)
{
    WebKitLoadStatus status = webkit_web_page_get_load_status(webPage);
    if (status == WEBKIT_LOAD_FINISHED) {
        g_signal_connect(fixture->domWindow, "click-event", G_CALLBACK(clickedCallback), fixture);

        g_assert(fixture->clicked == FALSE);

        // Position the event at the middle of the actor.
        gfloat eventX, eventY;
        clutter_actor_get_position(fixture->webActor, &eventX, &eventY);

        eventX += clutter_actor_get_width(fixture->webActor) / 2;
        eventY += clutter_actor_get_height(fixture->webActor) / 2;

        ClutterButtonEvent* event = (ClutterButtonEvent*)clutter_event_new(CLUTTER_BUTTON_PRESS);
        event->button = 1;
        event->x = eventX;
        event->y = eventY;
        event->click_count = 1;
        event->flags = CLUTTER_EVENT_FLAG_SYNTHETIC;
        event->time = clutter_x11_get_current_event_time() + 1;
        event->source = fixture->webActor;
        event->stage = CLUTTER_STAGE(clutter_actor_get_stage(fixture->webActor));

        clutter_actor_event(fixture->webActor, (ClutterEvent*)event, FALSE);
        clutter_event_free((ClutterEvent*)event);

        event = (ClutterButtonEvent*)clutter_event_new(CLUTTER_BUTTON_RELEASE);
        event->button = 1;
        event->x = eventX;
        event->y = eventY;
        event->click_count = 1;
        event->flags = CLUTTER_EVENT_FLAG_SYNTHETIC;
        event->time = clutter_x11_get_current_event_time() + 1;
        event->source = fixture->webActor;
        event->stage = CLUTTER_STAGE(clutter_actor_get_stage(fixture->webActor));

        clutter_actor_event(fixture->webActor, (ClutterEvent*)event, FALSE);
        clutter_event_free((ClutterEvent*)event);
    }

}

static void test_dom_domview_signals(DomDomviewFixture* fixture, gconstpointer data)
{
    g_assert(fixture);
    WebKitWebActor* actor = (WebKitWebActor*)fixture->webActor;
    g_assert(actor);
    WebKitWebPage* page = (WebKitWebPage*)fixture->webPage;
    g_assert(page);
    WebKitDOMDocument* document = webkit_web_page_get_dom_document(page);
    g_assert(document);
    WebKitDOMDOMWindow* domWindow = webkit_dom_document_get_default_view(document);
    g_assert(domWindow);

    fixture->domWindow = domWindow;

    g_signal_connect(domWindow, "load-event", G_CALLBACK(loadedCallback), fixture);
    g_signal_connect(page, "notify::load-status", G_CALLBACK(load_event_callback), fixture);

    clutter_actor_show_all(fixture->window);

    /* FIXME: not ideal, but clutter doesn't seem to have anything like map-event */
    g_idle_add((GSourceFunc)delayed_load_cb, fixture);
    g_main_loop_run(fixture->loop);

    g_assert(fixture->loaded);
    g_assert(fixture->clicked);
}

static gboolean
clicked_cb(WebKitDOMEventTarget* target, WebKitDOMEvent* event, DomDomviewFixture* fixture)
{
    g_assert(fixture->clicked == FALSE);
    fixture->clicked = TRUE;
    finish_loading(fixture);
    return FALSE;
}

static void load_status_callback(WebKitWebPage* webPage, GParamSpec* spec, DomDomviewFixture* fixture)
{
    WebKitLoadStatus status = webkit_web_page_get_load_status(webPage);
    if (status == WEBKIT_LOAD_FINISHED) {
        WebKitDOMDocument* document;
        WebKitDOMDOMWindow* domWindow;
        WebKitDOMElement* element;
        WebKitDOMEvent* event;
        glong clientX, clientY;

        document = webkit_web_page_get_dom_document(webPage);
        g_assert(document);
        domWindow = webkit_dom_document_get_default_view(document);
        g_assert(domWindow);
        fixture->domWindow = domWindow;

        element = webkit_dom_document_get_element_by_id(document, "test");
        g_assert(element);
        event = webkit_dom_document_create_event(document, "MouseEvent", NULL);
        g_assert(event);
        g_assert(WEBKIT_DOM_IS_EVENT(event));
        g_assert(WEBKIT_DOM_IS_MOUSE_EVENT(event));
        clientX = webkit_dom_element_get_client_left(element);
        clientY = webkit_dom_element_get_client_top(element);
        webkit_dom_mouse_event_init_mouse_event(WEBKIT_DOM_MOUSE_EVENT(event),
                                                "click", TRUE, TRUE,
                                                fixture->domWindow, 0, 0, 0, clientX, clientY,
                                                FALSE, FALSE, FALSE, FALSE,
                                                1, WEBKIT_DOM_EVENT_TARGET(element));
        g_signal_connect(element, "click-event", G_CALLBACK(clicked_cb), fixture);
        g_assert(fixture->clicked == FALSE);
        webkit_dom_event_target_dispatch_event(WEBKIT_DOM_EVENT_TARGET(element), event, NULL);
    }

}

static void test_dom_domview_dispatch_event(DomDomviewFixture* fixture, gconstpointer data)
{
    g_signal_connect(fixture->webPage, "notify::load-status", G_CALLBACK(load_status_callback), fixture);

    clutter_actor_show_all(fixture->window);

    /* FIXME: not ideal, but clutter doesn't seem to have anything like map-event */
    g_idle_add((GSourceFunc)delayed_load_cb, fixture);
    g_main_loop_run(fixture->loop);
    g_assert(fixture->clicked);
}

int main(int argc, char** argv)
{
    g_test_init(&argc, &argv, NULL);
    if (clutter_init(&argc, &argv) != CLUTTER_INIT_SUCCESS) {
        g_error ("Failed to initialized clutter.");
        return 1;
    }
    webkit_init();

    g_test_bug_base("https://bugs.webkit.org/");

    g_test_add("/webkit/domdomview/signals",
               DomDomviewFixture, HTML_DOCUMENT,
               dom_domview_fixture_setup,
               test_dom_domview_signals,
               dom_domview_fixture_teardown);

    g_test_add("/webkit/domdomview/dispatch_event",
               DomDomviewFixture, HTML_DOCUMENT,
               dom_domview_fixture_setup,
               test_dom_domview_dispatch_event,
               dom_domview_fixture_teardown);

    return g_test_run();
}

#else
int main(int argc, char** argv)
{
    g_critical("You will need at least glib-2.16.0 to run the unit tests. Doing nothing now.");
    return 0;
}

#endif
