/*
 *  Copyright (C) 2010 Collabora Ltd.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"
#include "mx-render-theme.h"

#include "CSSValueKeywords.h"
#include <mx/mx.h>

using namespace WebCore;

static void webkit_irender_theme_iface_init (WebKitIRenderThemeInterface* iface);

G_DEFINE_TYPE_WITH_CODE(MxRenderTheme, mx_render_theme, G_TYPE_OBJECT,
                        G_IMPLEMENT_INTERFACE(WEBKIT_TYPE_IRENDER_THEME,
                                              webkit_irender_theme_iface_init));

static void mx_render_theme_class_init(MxRenderThemeClass* klass)
{
}

static void mx_render_theme_init(MxRenderTheme* self)
{
}

MxRenderTheme* mx_render_theme_new(void)
{
    return MX_RENDER_THEME(g_object_new(MX_TYPE_RENDER_THEME, NULL));
}

static gboolean get_widget_metrics(WebKitIRenderTheme* theme,
                                   WebKitIRenderThemeWidgetMetrics* metrics)
{
    metrics->blinkInterval = 0.5;

    return TRUE;
}

static gboolean get_scrollbar_metrics(WebKitIRenderTheme* theme,
                                      WebKitIRenderThemeScrollbarMetrics* metrics)
{
    metrics->has_buttons = TRUE;
    metrics->stepper_size = 0;
    metrics->slider_width = 0;
    metrics->trough_border = 0;
    metrics->min_slider_size = 0;
    return TRUE;
}

static void paintScrollbarBackground(cairo_t* cr, cairo_rectangle_int_t* rect)
{
    cairo_save(cr);
    cairo_rectangle(cr, rect->x, rect->y, rect->width, rect->height);
    cairo_clip(cr);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_paint(cr);
    cairo_restore(cr);
}

static void paintTrackBackground(cairo_t* cr, cairo_rectangle_int_t* rect)
{
    cairo_save(cr);
    cairo_rectangle(cr, rect->x, rect->y, rect->width, rect->height);
    cairo_clip(cr);
    cairo_set_source_rgb(cr, 0.6, 0.6, 0.6);
    cairo_paint(cr);
    cairo_restore(cr);
}

static void paintScrollbarThumb(cairo_t* cr, cairo_rectangle_int_t* rect)
{
    cairo_save(cr);
    cairo_rectangle(cr, rect->x, rect->y, rect->width, rect->height);
    cairo_clip(cr);
    cairo_set_source_rgb(cr, 0.0, 0.4, 0.0);
    cairo_paint(cr);
    cairo_restore(cr);
}

static void paintScrollbarButton(cairo_t* cr, cairo_rectangle_int_t* rect, gint flags)
{
    // Background
    cairo_save(cr);
    cairo_rectangle(cr, rect->x, rect->y, rect->width, rect->height);
    cairo_clip(cr);
    cairo_set_source_rgb(cr, 0.2, 0.0, 0.0);
    cairo_paint(cr);
    cairo_restore(cr);

    // Paint the "arrow"
    double start_x, start_y;
    double edge_x, edge_y;
    double end_x, end_y;

    if (flags & stepperVertical) {
        if (flags & stepperBottom) {
            start_x = rect->x;
            start_y = rect->y;
            edge_x = rect->x + (rect->width/2);
            edge_y = rect->y + rect->height;
            end_x = rect->x + rect->width;
            end_y = rect->y;
        } else {
            start_x = rect->x;
            start_y = rect->y + rect->height;
            edge_x = rect->x + (rect->width / 2);
            edge_y = rect->y;
            end_x = rect->x + rect->width;
            end_y = rect->y + rect->height;
        }
    } else {
        if (flags & stepperBottom) {
            start_x = rect->x;
            start_y = rect->y;
            edge_x = rect->x + rect->width;
            edge_y = rect->y + (rect->height / 2);
            end_x = rect->x;
            end_y = rect->y + rect->height;
        } else {
            start_x = rect->x + rect->width;
            start_y = rect->y;
            edge_x = rect->x;
            edge_y = rect->y + (rect->height / 2);
            end_x = rect->x + rect->width;
            end_y = rect->y + rect->height;
        }
    }

    cairo_save(cr);
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_set_line_width(cr, 1.0);
    cairo_move_to(cr, start_x, start_y);
    cairo_line_to(cr, edge_x, edge_y);
    cairo_line_to(cr, end_x, end_y);
    cairo_stroke(cr);
    cairo_restore(cr);
}

static guint32 get_platform_color(WebKitIRenderTheme* theme,
                                  WebKitIRenderThemePlatformColor color)
{
    switch (color) {
    case platformColorActiveSelectionForeground:
    case platformColorActiveListBoxSelectionForeground:
    case platformColorInactiveSelectionForeground:
    case platformColorInactiveListBoxSelectionForeground:
        return 0xFFFFFF00;
    case platformColorActiveSelectionBackground:
    case platformColorInactiveSelectionBackground:
        return 0x00000000;
    default:
        return 0;
    }

    return 0;
}

static guint32 get_system_color(WebKitIRenderTheme* theme,
                                int cssValueId)
{
    // Values copied over and converted to clutter convention from
    // WebCore's RenderTheme.cpp.
    switch (cssValueId) {
        case CSSValueActiveborder:
            return 0xFFFFFF00;
        case CSSValueActivecaption:
            return 0xCCCCCC00;
        case CSSValueAppworkspace:
            return 0xFFFFFF00;
        case CSSValueBackground:
            return 0x6363CE00;
        case CSSValueButtontext:
            return 0xFFFFFF00;
        case CSSValueCaptiontext:
            return 0x00000000;
        case CSSValueGraytext:
            return 0x80808000;
        case CSSValueHighlight:
            return 0xB5D5FF00;
        case CSSValueHighlighttext:
            return 0xFFFFFF00;
        case CSSValueInactiveborder:
            return 0xFFFFFF00;
        case CSSValueInactivecaption:
            return 0xFFFFFF00;
        case CSSValueInactivecaptiontext:
            return 0x7F7F7F00;
        case CSSValueInfobackground:
            return 0xFBFCC500;
        case CSSValueInfotext:
            return 0x00000000;
        case CSSValueMenu:
            return 0xC0C0C000;
        case CSSValueMenulistText:
            return 0xFFFFFF00;
        case CSSValueMenutext:
            return 0x00000000;
        case CSSValueScrollbar:
            return 0xFFFFFF00;
        case CSSValueText:
            return 0xFFFFFF00;
        case CSSValueThreeddarkshadow:
            return 0x66666600;
        case CSSValueThreedface:
            return 0xC0C0C000;
        case CSSValueThreedhighlight:
            return 0xDDDDDD00;
        case CSSValueThreedlightshadow:
            return 0xC0C0C000;
        case CSSValueThreedshadow:
            return 0x88888800;
        case CSSValueWindow:
            return 0xFFFFFF00;
        case CSSValueWindowframe:
            return 0xCCCCCC00;
        case CSSValueWindowtext:
            return 0x00000000;
    }
    return 0;
}

static gboolean paint_widget(WebKitIRenderTheme* theme,
                             WebKitIRenderThemeWidgetType type,
                             WebKitIRenderThemeWidgetState* state,
                             gint flags,
                             cairo_t* cr,
                             cairo_rectangle_int_t* rect)
{
    switch (type) {
    case widgetScrollbarTrackHorizontal:
    case widgetScrollbarTrackVertical:
        paintTrackBackground(cr, rect);
        break;
    case widgetScrolledWindow:
        paintScrollbarBackground(cr, rect);
        break;
    case widgetScrollbarButton:
        paintScrollbarButton(cr, rect, flags);
        break;
    case widgetScrollbarThumbHorizontal:
    case widgetScrollbarThumbVertical:
        paintScrollbarThumb(cr, rect);
        break;
    default:
        return FALSE;
    }

    return TRUE;
}

static void webkit_irender_theme_iface_init (WebKitIRenderThemeInterface* iface)
{
    iface->get_platform_color = get_platform_color;
    iface->get_system_color = get_system_color;
    iface->get_widget_metrics = get_widget_metrics;
    iface->get_scrollbar_metrics = get_scrollbar_metrics;
    iface->paint_widget = paint_widget;
}
