/*
 * Copyright 2010 Collabora Limited
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
 * more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "mx-web-view.h"
#include "config.h"

#include "FrameView.h"
#include "webkitdom.h"
#include "webkit-mx-popup-menu.h"
#include <webkit/webkitiwebview.h>
#include <webkit/webkitwebpageprivate.h>
#include <webkit/webkitgwebframeprivate.h>
#include <clutter/clutter.h>
#include <mx/mx.h>

using namespace WebKit;

static void webkit_iweb_view_iface_init(WebKitIWebViewInterface* iface);
static void mx_scrollable_iface_init(MxScrollableIface *iface);

static void scrollable_set_adjustments (MxScrollable *scrollable,
                                        MxAdjustment *hadjustment,
                                        MxAdjustment *vadjustment);

static void scrollable_get_adjustments (MxScrollable  *scrollable,
                                        MxAdjustment **hadjustment,
                                        MxAdjustment **vadjustment);

static void set_fast_zoom_level (MxWebView* webView, double fastZoomLevel);
static void set_scroll_offset_x (MxWebView* webView, int offset);
static void set_scroll_offset_y (MxWebView* webView, int offset);

G_DEFINE_TYPE_WITH_CODE(MxWebView, mx_web_view, MX_TYPE_WIDGET,
                        G_IMPLEMENT_INTERFACE(WEBKIT_TYPE_IWEB_VIEW,
                                              webkit_iweb_view_iface_init)
                        G_IMPLEMENT_INTERFACE(MX_TYPE_SCROLLABLE,
                                              mx_scrollable_iface_init))

#define MX_WEB_VIEW_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), MX_TYPE_WEB_VIEW, MxWebViewPrivate))

typedef enum {
    GESTURE_PAN_BEGIN,
    GESTURE_PAN_END,
    GESTURE_PAN,
    GESTURE_PAN_INTERPOLATED,
    GESTURE_PAN_STOPPED,
    GESTURE_TAP,
    GESTURE_TAP_CANCEL
} GestureEventType;

typedef struct {
    GestureEventType type;
    gfloat x;
    gfloat y;
    gfloat dx;
    gfloat dy;
    gint64 time;
    ClutterModifierType modifiers;
} GestureEvent;

struct _MxWebViewPrivate
{
    gboolean dirty;
    ClutterGeometry dirtyRect;
    gboolean painting;
    gboolean transparent;

    // Guard against changing size while allocating.
    gboolean allocating;

    MxAdjustment *hadjustment;
    MxAdjustment *vadjustment;

    WebKitWebPage* webPage;

    /* Children */
    ClutterActor *texture;

    double fastZoomLevel;

#if ENABLE(GESTURE_EVENTS)
    ClutterAction *pan_action;
    ClutterAction *tap_action;
#endif

    GSList* plugins;

#if USE(TILED_BACKING_STORE)
    GSList* tiles;
#endif

    ClutterActor *stage;
    gulong stage_captured_event_id;
    GSList* touchesWaitingRelease;

#if ENABLE(GESTURE_EVENTS)
    GArray *queued_gesture_events;

    guint enqueue_gesture_events : 1;
    guint ignore_pan : 1;
    guint ignore_tap : 1;
#endif
};

enum
{
    PROP_0,

    PROP_HADJUST,
    PROP_VADJUST,
    PROP_FAST_ZOOM_LEVEL,
    PROP_SCROLL_OFFSET_X,
    PROP_SCROLL_OFFSET_Y,

    PROP_LAST
};

static void
mx_web_view_get_property (GObject    *object,
                          guint       property_id,
                          GValue     *value,
                          GParamSpec *pspec)
{
    MxWebViewPrivate* priv = MX_WEB_VIEW(object)->priv;
    MxAdjustment* adjustment;

    switch (property_id) {
    case PROP_HADJUST:
        scrollable_get_adjustments (MX_SCROLLABLE (object), &adjustment, NULL);
        g_value_set_object (value, adjustment);
        break;

    case PROP_VADJUST:
        scrollable_get_adjustments (MX_SCROLLABLE (object), NULL, &adjustment);
        g_value_set_object (value, adjustment);
        break;

    case PROP_FAST_ZOOM_LEVEL:
        g_value_set_double (value, priv->fastZoomLevel);
        break;

    case PROP_SCROLL_OFFSET_X: {
        WebKitWebFrame *mainFrame = webkit_web_page_get_main_frame(priv->webPage);
        int offset = webkit_web_frame_get_x_scroll_offset(mainFrame);
        g_value_set_int (value, offset);
        break;
    }

    case PROP_SCROLL_OFFSET_Y: {
        WebKitWebFrame *mainFrame = webkit_web_page_get_main_frame(priv->webPage);
        int offset = webkit_web_frame_get_y_scroll_offset(mainFrame);
        g_value_set_int (value, offset);
        break;
    }

    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
    }
}

static void
mx_web_view_set_property (GObject      *object,
                          guint         property_id,
                          const GValue *value,
                          GParamSpec   *pspec)
{
    MxWebViewPrivate* priv = MX_WEB_VIEW(object)->priv;

    switch (property_id) {
    case PROP_HADJUST:
        scrollable_set_adjustments(MX_SCROLLABLE (object),
                                   static_cast<MxAdjustment*>(g_value_get_object(value)),
                                   priv->vadjustment);
    break;

    case PROP_VADJUST:
        scrollable_set_adjustments(MX_SCROLLABLE (object),
                                   priv->hadjustment,
                                   static_cast<MxAdjustment*>(g_value_get_object(value)));
    break;

    case PROP_FAST_ZOOM_LEVEL:
        set_fast_zoom_level(MX_WEB_VIEW(object), g_value_get_double(value));
        break;

    case PROP_SCROLL_OFFSET_X:
        set_scroll_offset_x(MX_WEB_VIEW(object), g_value_get_int(value));
        break;

    case PROP_SCROLL_OFFSET_Y:
        set_scroll_offset_y(MX_WEB_VIEW(object), g_value_get_int(value));
        break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
    }
}

static void
mx_web_view_dispose (GObject *object)
{
    MxWebView* view = MX_WEB_VIEW(object);
    MxWebViewPrivate* priv = MX_WEB_VIEW_GET_PRIVATE(view);

    if (priv->hadjustment)
      {
        g_object_unref (priv->hadjustment);
        priv->hadjustment = NULL;
      }

    if (priv->vadjustment)
      {
        g_object_unref (priv->vadjustment);
        priv->vadjustment = NULL;
      }

#if USE(TILED_BACKING_STORE)
    if (priv->tiles) {
        g_slist_free(priv->tiles);
        priv->tiles = NULL;
    }
#endif

    if (priv->plugins) {
        g_slist_free(priv->plugins);
        priv->plugins = NULL;
    }

    g_clear_pointer(&priv->touchesWaitingRelease, g_slist_free);

    if (priv->webPage)
    {
        webkit_web_page_set_root_layer(priv->webPage, NULL);
        g_object_unref(priv->webPage);
        priv->webPage = 0;
    }

    /* No need to unref the texture: the only reference we own is the
     * floating reference that got sunk when we made ourselves parent,
     * and it will be decremented when it is unparented.
     */
    if (priv->texture)
    {
        clutter_actor_unparent(priv->texture);
        priv->texture = 0;
    }

#if ENABLE(GESTURE_EVENTS)
    if (g_signal_handler_is_connected (object, priv->stage_captured_event_id))
        g_signal_handler_disconnect (object, priv->stage_captured_event_id);
#endif

    G_OBJECT_CLASS (mx_web_view_parent_class)->dispose (object);
}

static void
mx_web_view_finalize (GObject *object)
{
  G_OBJECT_CLASS (mx_web_view_parent_class)->finalize (object);
}

static void
mx_web_view_allocate (ClutterActor          *self,
                   const ClutterActorBox *box,
                   ClutterAllocationFlags flags)
{
    MxWebViewPrivate* priv = MX_WEB_VIEW_GET_PRIVATE(MX_WEB_VIEW(self));
    priv->allocating = TRUE;

    flags = (ClutterAllocationFlags)(flags | CLUTTER_DELEGATE_LAYOUT);
    CLUTTER_ACTOR_CLASS (mx_web_view_parent_class)->allocate (self, box, flags);

    int childWidth  = clutter_actor_box_get_width(box);
    int childHeight = clutter_actor_box_get_height(box);

    if (priv->texture) {
        ClutterActorBox textureBox;

        textureBox.x1 = 0;
        textureBox.x2 = box->x2;
        textureBox.y1 = 0;
        textureBox.y2 = box->y2;

        /* The texture occupies the whole area.
         */
        clutter_actor_allocate(priv->texture, &textureBox, flags);
    }

    WebKitWebFrame *mainFrame;
    mainFrame = webkit_web_page_get_main_frame(priv->webPage);

    int width, height;
    webkit_web_frame_get_contents_size(mainFrame, &width, &height);

    if (priv->texture)
        clutter_cairo_texture_set_surface_size(CLUTTER_CAIRO_TEXTURE(priv->texture),
                                               childWidth > 0 ? childWidth : 1,
                                               childHeight > 0 ? childHeight : 1);

    webkit_web_page_set_viewport_size(priv->webPage, childWidth, childHeight);

    /* Refresh the adjustments. */
    if (priv->hadjustment) {
        g_object_set(G_OBJECT(priv->hadjustment),
                     "lower", 0.0,
                     "upper", (gdouble)(width < childWidth ? childWidth : width),
                     "page-size", (gdouble)childWidth,
                     "page-increment", (gdouble)childWidth / 3,
                     "step-increment", (gdouble)childWidth / 12,
                     NULL);
    }
    if (priv->vadjustment) {
        g_object_set(G_OBJECT(priv->vadjustment),
                     "lower", 0.0,
                     "upper", (gdouble)(height < childHeight ? childHeight : height),
                     "page-size", (gdouble)childHeight,
                     "page-increment", (gdouble)childHeight / 3,
                     "step-increment", (gdouble)childHeight / 12,
                     NULL);
    }

    priv->allocating = FALSE;
}

static AtkObject* mx_web_view_get_accessible(ClutterActor *actor)
{
    WebKitWebPage* page = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(actor));
    return webkit_web_page_get_accessible(page);
}

static void mx_web_view_get_preferred_width(ClutterActor *actor,
                                            gfloat  forHeight,
                                            gfloat *minWidth,
                                            gfloat *naturalWidth)
{
    *minWidth = 0.0;
    *naturalWidth = 640.0;

    WebKitWebFrame* mainFrame = webkit_web_page_get_main_frame(MX_WEB_VIEW(actor)->priv->webPage);
    if (!mainFrame)
        return;

    int width;
    if (webkit_web_frame_get_contents_size(mainFrame, &width, NULL) && width)
        *naturalWidth = (gfloat)width;

    MxWebViewPrivate* priv = MX_WEB_VIEW(actor)->priv;
    if (priv->hadjustment)
        *minWidth = *naturalWidth;
}

static void mx_web_view_get_preferred_height(ClutterActor* actor,
                                             gfloat  forWidth,
                                             gfloat* minHeight,
                                             gfloat* naturalHeight)
{
    *minHeight = 0.0;
    *naturalHeight = 480.0;

    WebKitWebFrame* mainFrame = webkit_web_page_get_main_frame(MX_WEB_VIEW(actor)->priv->webPage);
    if (!mainFrame)
        return;

    int height;
    if (webkit_web_frame_get_contents_size(mainFrame, NULL, &height) && height)
        *naturalHeight = (gfloat)height;

    MxWebViewPrivate* priv = MX_WEB_VIEW(actor)->priv;
    if (priv->vadjustment)
        *minHeight = *naturalHeight;
}

static void
mx_web_view_paint(ClutterActor* actor)
{
    MxWebViewPrivate* priv = MX_WEB_VIEW(actor)->priv;

    /* First, paint the texture */
    if (priv->texture)
        clutter_actor_paint(priv->texture);

#if USE(TILED_BACKING_STORE)
    if (priv->tiles)
        g_slist_foreach(priv->tiles, (GFunc)clutter_actor_paint, NULL);
#endif

#if USE(ACCELERATED_COMPOSITING)
    ClutterActor* rootLayer = webkit_web_page_get_root_layer_actor(priv->webPage);

    if (rootLayer)
        clutter_actor_paint(rootLayer);
#endif

    g_slist_foreach(priv->plugins, (GFunc)clutter_actor_paint, NULL);
}

static gboolean
mx_web_view_get_paint_volume (ClutterActor* actor, ClutterPaintVolume *volume)
{
    /* With so many modelview transformations, computing the paint volume
     * becomes too expensive, better redraw it whole */
    return false;
}

#if ENABLE(WEBGL)
void
stage_paint (ClutterActor *actor, gpointer user_data)
{
    ClutterActor *self = CLUTTER_ACTOR (user_data);
    webkit_iweb_view_pop_gl_context(self);
}

void
mx_web_view_realize (ClutterActor *self)
{
    ClutterActor *stage = clutter_actor_get_stage (self);
    g_signal_connect (stage, "paint", G_CALLBACK (stage_paint), self);
}
#endif

static gboolean on_captured_event(ClutterActor* actor, ClutterEvent* event, gpointer data)
{
    /* Snoop the event type and set the flag that says that gesture events
     * generated in response of a pointer event should be enqueued and pushed
     * to WebCore after the pointer event that generated them.
     *
     * This may be called more than once, as it is invoked on the actor before any
     * gesture begins and also on the stage when a gesture is running and we
     * should precede the gesture stage handler.
     */
    MxWebViewPrivate* priv = MX_WEB_VIEW(actor)->priv;

    ClutterEventType type = clutter_event_type(event);
    switch (type) {
        case CLUTTER_TOUCH_END:
        {
            // If the user touched the WebView, but moved their finger outside the actor, we
            // will miss the release event, and WebCore will believe that touch point still
            // has a finger on it forever thereafter, so we make sure the touch-event signal
            // is emited on the actor.
            for (GSList* iter = priv->touchesWaitingRelease; iter; iter = iter->next) {
                ClutterEventSequence* sequence = static_cast<ClutterEventSequence*>(iter->data);
                if ((sequence == clutter_event_get_event_sequence (event)) && (clutter_event_get_source(event) != actor)) {
                    gboolean handled;
                    g_signal_emit_by_name (actor, "touch-event", event, &handled);
                }
            }
        }
#if ENABLE(GESTURE_EVENTS)
        case CLUTTER_BUTTON_PRESS:
        case CLUTTER_BUTTON_RELEASE:
        case CLUTTER_MOTION:
        case CLUTTER_TOUCH_BEGIN:
        case CLUTTER_TOUCH_UPDATE:
        case CLUTTER_TOUCH_CANCEL:
            priv->enqueue_gesture_events = TRUE;
            break;
#endif
        default:
            break;
    }
    return CLUTTER_EVENT_PROPAGATE;
}

/*
 * connect_stage_capture_event:
 *
 * Set up a ::captured-event handler called from the current stage to
 * snoop the event type and determine if we should enqueue gesture events
 * before they get emitted.
 */
static void connect_stage_capture_event(MxWebView* self)
{
    MxWebViewPrivate* priv = self->priv;

    if (priv->stage) {
        if (g_signal_handler_is_connected(priv->stage, priv->stage_captured_event_id))
            g_signal_handler_disconnect(priv->stage, priv->stage_captured_event_id);
        g_object_remove_weak_pointer(G_OBJECT(priv->stage), (gpointer*)&priv->stage);
    }

    priv->stage = clutter_actor_get_stage(CLUTTER_ACTOR(self));
    if (!priv->stage)
        return;

    g_object_add_weak_pointer(G_OBJECT(priv->stage), (gpointer*)&priv->stage);
    priv->stage_captured_event_id = g_signal_connect_object(
        G_OBJECT(priv->stage), "captured-event", G_CALLBACK(on_captured_event),
        self, G_CONNECT_SWAPPED);
}

#if ENABLE(GESTURE_EVENTS)
gboolean gesture_is_related_to_event(ClutterGestureAction* gesture,
                                     ClutterEvent* event)
{
    const ClutterEvent* last = clutter_gesture_action_get_last_event(gesture, 0);
    return clutter_event_get_device_id(event) == clutter_event_get_device_id(last) &&
        clutter_event_get_event_sequence(event) == clutter_event_get_event_sequence(last);
}

static gboolean gesture_event_send(MxWebView* self, const GestureEvent* evt)
{
    MxWebViewPrivate* priv = self->priv;
    ClutterActor* actor = CLUTTER_ACTOR(self);

    switch (evt->type)
    {
        case GESTURE_PAN_BEGIN:
            if (priv->ignore_pan)
                break;
            return webkit_iweb_view_gesture_scroll_begin(actor, evt->time, evt->x, evt->y);
        case GESTURE_PAN:
            if (priv->ignore_pan)
                break;
            return webkit_iweb_view_gesture_scroll_update(actor, evt->time, evt->x, evt->y, evt->dx, evt->dy);
        case GESTURE_PAN_END:
            if (priv->ignore_pan)
                break;
            return webkit_iweb_view_gesture_scroll_end(actor, evt->time, evt->x, evt->y) &&
                webkit_iweb_view_gesture_scroll_momentum_begin(actor, evt->time, evt->x, evt->y);
        case GESTURE_PAN_INTERPOLATED:
            if (priv->ignore_pan)
                break;
            return webkit_iweb_view_gesture_scroll_momentum_update(actor, evt->time, evt->x, evt->y, evt->dx, evt->dy);
        case GESTURE_PAN_STOPPED:
            if (priv->ignore_pan)
                break;
            return webkit_iweb_view_gesture_scroll_momentum_end(actor, evt->time, evt->x, evt->y);
        case GESTURE_TAP:
            if (priv->ignore_tap)
                break;
            return webkit_iweb_view_gesture_tap(actor, evt->time, evt->x, evt->y, evt->modifiers);
        case GESTURE_TAP_CANCEL:
            if (priv->ignore_tap)
                break;
            return webkit_iweb_view_gesture_tap_cancel(actor, evt->time, evt->x, evt->y, evt->modifiers);
    }
    return FALSE;
}

static void gesture_event_enqueue_or_send(MxWebView* self, const GestureEvent* gesture)
{
    MxWebViewPrivate* priv = self->priv;
    if (priv->enqueue_gesture_events)
        g_array_append_vals (priv->queued_gesture_events, gesture, 1);
    else
        gesture_event_send(self, gesture);
}

static void gesture_events_queue_flush(ClutterActor* actor, ClutterEvent *event)
{
    MxWebView* self = MX_WEB_VIEW(actor);
    MxWebViewPrivate* priv = self->priv;

    for (guint i = 0; i < priv->queued_gesture_events->len; i++)
      {
          GestureEvent* evt = &g_array_index(priv->queued_gesture_events, GestureEvent, i);
          gesture_event_send(self, evt);
      }

    g_array_set_size(priv->queued_gesture_events, 0);
    priv->enqueue_gesture_events = FALSE;
}
#endif //GESTURE_EVENTS

static gboolean mx_web_view_button_press_event(ClutterActor* actor, ClutterButtonEvent* event)
{
    MxWebViewPrivate* priv = MX_WEB_VIEW(actor)->priv;
    gboolean handled = webkit_iweb_view_button_press_event(actor, event);
#if ENABLE(GESTURE_EVENTS)
    if (gesture_is_related_to_event(CLUTTER_GESTURE_ACTION(priv->pan_action), (ClutterEvent*)event))
        priv->ignore_pan = handled;
    /* Tap gestures caused by button presses are cancelled by on_tap_gesture_begin(),
     * we don't need to ignore them here */
    gesture_events_queue_flush(actor, (ClutterEvent*)event);
#endif
    return handled;
}

static gboolean mx_web_view_button_release_event(ClutterActor* actor, ClutterButtonEvent* event)
{
    gboolean handled = webkit_iweb_view_button_release_event(actor, event);
#if ENABLE(GESTURE_EVENTS)
    gesture_events_queue_flush(actor, (ClutterEvent*)event);
#endif
    return handled;
}

static gboolean mx_web_view_motion_event(ClutterActor* actor, ClutterMotionEvent* event)
{
    gboolean handled = webkit_iweb_view_motion_event(actor, event);
#if ENABLE(GESTURE_EVENTS)
    gesture_events_queue_flush(actor, (ClutterEvent*)event);
#endif
    return handled;
}

static gboolean mx_web_view_touch_event(ClutterActor* actor, ClutterTouchEvent* event)
{
    MxWebView* self = MX_WEB_VIEW(actor);
    MxWebViewPrivate* priv = self->priv;
    gboolean handled = webkit_iweb_view_touch_event(actor, event);
    ClutterEventType eventType = clutter_event_type((ClutterEvent*)event);

    if (eventType == CLUTTER_TOUCH_BEGIN) {
        connect_stage_capture_event(self);

#if ENABLE(GESTURE_EVENTS)
        if (gesture_is_related_to_event(CLUTTER_GESTURE_ACTION(priv->pan_action), reinterpret_cast<ClutterEvent*>(event)))
            priv->ignore_pan = handled;
        if (gesture_is_related_to_event(CLUTTER_GESTURE_ACTION(priv->tap_action), reinterpret_cast<ClutterEvent*>(event)))
            priv->ignore_tap = handled;
#endif

        priv->touchesWaitingRelease = g_slist_append(priv->touchesWaitingRelease, event->sequence);
    }

    if (eventType == CLUTTER_TOUCH_END || eventType == CLUTTER_TOUCH_CANCEL)
        priv->touchesWaitingRelease = g_slist_remove(priv->touchesWaitingRelease, event->sequence);

#if ENABLE(GESTURE_EVENTS)
    gesture_events_queue_flush(actor, (ClutterEvent*)event);
#endif

    return handled;
}

static void
mx_web_view_class_init (MxWebViewClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);

    g_type_class_add_private (klass, sizeof (MxWebViewPrivate));

    object_class->get_property = mx_web_view_get_property;
    object_class->set_property = mx_web_view_set_property;
    object_class->dispose = mx_web_view_dispose;
    object_class->finalize = mx_web_view_finalize;

    actor_class->allocate = mx_web_view_allocate;
    actor_class->get_accessible = mx_web_view_get_accessible;
    actor_class->get_preferred_width = mx_web_view_get_preferred_width;
    actor_class->get_preferred_height = mx_web_view_get_preferred_height;
    actor_class->get_paint_volume = mx_web_view_get_paint_volume;
    actor_class->paint = mx_web_view_paint;
    actor_class->button_press_event = mx_web_view_button_press_event;
    actor_class->button_release_event = mx_web_view_button_release_event;
    actor_class->motion_event = mx_web_view_motion_event;
    actor_class->scroll_event = webkit_iweb_view_scroll_event;
    actor_class->key_press_event = webkit_iweb_view_key_event;
    actor_class->key_release_event = webkit_iweb_view_key_event;
#if ENABLE(TOUCH_EVENTS)
    actor_class->touch_event = mx_web_view_touch_event;
#endif
    actor_class->key_focus_in = webkit_iweb_view_key_focus_in;
    actor_class->key_focus_out = webkit_iweb_view_key_focus_out;

#if ENABLE(WEBGL)
    actor_class->realize = mx_web_view_realize;
#endif

    g_object_class_override_property(object_class,
                                     PROP_HADJUST,
                                     "horizontal-adjustment");

    g_object_class_override_property(object_class,
                                     PROP_VADJUST,
                                     "vertical-adjustment");

    /**
     * MxWebView:fast-zoom-level:
     *
     * The zoom level to scale without having to re-render the content, making
     * it fast enough to be used in animations.
     *
     * The #MxWebView:fast-zoom-level property is animatable.
     */
    g_object_class_install_property(object_class,
        PROP_FAST_ZOOM_LEVEL,
        g_param_spec_double ("fast-zoom-level",
                             "Fast zoom level",
                             "Level to quickly zoom, without re-rendering",
                             -G_MAXDOUBLE, G_MAXDOUBLE,
                             1.,
                             static_cast<GParamFlags>(G_PARAM_READWRITE |
                                                      G_PARAM_STATIC_STRINGS)));

    /**
     * MxWebView:scroll-offset-x:
     *
     * The horizontal offset to which to scroll.
     *
     * The #MxWebView:scroll-offset-x property is animatable.
     */
    g_object_class_install_property(object_class,
        PROP_SCROLL_OFFSET_X,
        g_param_spec_int ("scroll-offset-x",
                          "Scroll offset X",
                          "Horizontal scroll position",
                          G_MININT, G_MAXINT,
                          0,
                          static_cast<GParamFlags>(G_PARAM_READWRITE |
                                                   G_PARAM_STATIC_STRINGS)));

    /**
     * MxWebView:scroll-offset-y:
     *
     * The vertical offset to which to scroll.
     *
     * The #MxWebView:scroll-offset-y property is animatable.
     */
    g_object_class_install_property(object_class,
        PROP_SCROLL_OFFSET_Y,
        g_param_spec_int ("scroll-offset-y",
                          "Scroll offset Y",
                          "Vertical scroll position",
                          G_MININT, G_MAXINT,
                          0,
                          static_cast<GParamFlags>(G_PARAM_READWRITE |
                                                   G_PARAM_STATIC_STRINGS)));
}

static void contents_size_changed_cb(WebKitWebFrame* frame,
                                     int width, int height,
                                     MxWebView* webView)
{
    MxWebViewPrivate* priv = webView->priv;
    if (!priv->allocating)
        clutter_actor_queue_relayout(CLUTTER_ACTOR(webView));
}

static void update_child_transform(MxWebView* webView)
{
  ClutterMatrix matrix;
  int xScroll, yScroll;
  MxWebViewPrivate *priv = webView->priv;
  WebKitWebFrame *mainFrame;

  clutter_matrix_init_identity(&matrix);

  mainFrame = webkit_web_page_get_main_frame(priv->webPage);
  if (core(mainFrame)->view() && core(mainFrame)->view()->delegatesScrolling()) {
      webkit_web_frame_get_scroll_offset(mainFrame, &xScroll, &yScroll);

      xScroll *= webkit_web_frame_get_contents_scale(mainFrame);
      yScroll *= webkit_web_frame_get_contents_scale(mainFrame);

      xScroll *= priv->fastZoomLevel;
      yScroll *= priv->fastZoomLevel;

      cogl_matrix_translate(&matrix, -xScroll, -yScroll, 0.0f);
  }

  cogl_matrix_scale(&matrix, priv->fastZoomLevel, priv->fastZoomLevel, 1.0);

  clutter_actor_set_child_transform(CLUTTER_ACTOR(webView), &matrix);
}

static void scroll_offset_changed(WebKitWebFrame* frame,
                                  MxWebView* webView)
{
    MxWebViewPrivate* priv = webView->priv;
    gint x, y;

    webkit_web_frame_get_scroll_offset(frame, &x, &y);
    if (priv->hadjustment)
        mx_adjustment_set_value(priv->hadjustment,
                (gdouble)x);
    if (priv->vadjustment)
        mx_adjustment_set_value(priv->vadjustment,
                (gdouble)y);
    
    update_child_transform(webView);
}

static gboolean mx_web_view_draw(ClutterCairoTexture* texture, cairo_t* cr, MxWebView* webView)
{
    MxWebViewPrivate* priv = webView->priv;

    WebKitWebPage* page = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(webView));
    if (!page)
        return FALSE;

    WebKitWebFrame* mainFrame = webkit_web_page_get_main_frame(page);
    if (!mainFrame)
        return FALSE;

    priv->painting = TRUE;

    guint width, height;
    clutter_cairo_texture_get_surface_size(texture, &width, &height);

    webkit_web_frame_render(mainFrame, cr, 0, 0, (int)width, (int)height);

    priv->painting = FALSE;

    return TRUE;
}

static void settingsChanged(GObject* object, GParamSpec* pspec, MxWebView* webView)
{
    MxWebViewPrivate* priv = webView->priv;

    WebKitWebSettings* settings = webkit_web_page_get_settings(priv->webPage);

    gboolean tiledBackingStoreEnabled;
    g_object_get(settings, "enable-tiled-backing-store", &tiledBackingStoreEnabled, NULL);

#if USE(TILED_BACKING_STORE)
    if (tiledBackingStoreEnabled) {
        if (priv->texture)
            clutter_actor_remove_child(CLUTTER_ACTOR(webView), priv->texture);
        priv->texture = 0;
        return;
    }
#endif

    priv->texture = clutter_cairo_texture_new(0, 0);
    clutter_actor_add_child(CLUTTER_ACTOR(webView), priv->texture);

    g_signal_connect(priv->texture, "draw",
                     G_CALLBACK(mx_web_view_draw), webView);
}

#if ENABLE(GESTURE_EVENTS)
static gboolean
on_pan_gesture_begin(ClutterAction   *action,
                     ClutterActor    *actor,
                     gpointer         data)
{
    MxWebView* self = MX_WEB_VIEW(actor);

    const ClutterEvent* event = clutter_gesture_action_get_last_event(CLUTTER_GESTURE_ACTION(action), 0);
    GestureEvent gesture;
    gesture.type = GESTURE_PAN_BEGIN;
    gesture.time = clutter_event_get_time(event);
    gesture.modifiers = clutter_event_get_state(event);
    clutter_gesture_action_get_press_coords(CLUTTER_GESTURE_ACTION(action), 0, &gesture.x, &gesture.y);
    gesture_event_enqueue_or_send(self, &gesture);

    connect_stage_capture_event(MX_WEB_VIEW(actor));

    return TRUE;
}

static void
on_pan_gesture_end(ClutterAction   *action,
                   ClutterActor    *actor,
                   gpointer         data)
{
    MxWebView* self = MX_WEB_VIEW(actor);

    const ClutterEvent* event = clutter_gesture_action_get_last_event(CLUTTER_GESTURE_ACTION(action), 0);
    GestureEvent gesture;
    gesture.type = GESTURE_PAN_END;
    gesture.time = clutter_event_get_time(event);
    gesture.modifiers = clutter_event_get_state(event);
    clutter_gesture_action_get_release_coords(CLUTTER_GESTURE_ACTION(action), 0, &gesture.x, &gesture.y);
    gesture_event_enqueue_or_send(self, &gesture);
}

static gboolean
on_pan(ClutterPanAction   *action,
       ClutterActor       *actor,
       gboolean            is_interpolated,
       gpointer            data)
{
    MxWebView* self = MX_WEB_VIEW(actor);

    GestureEvent gesture;
    if (!is_interpolated) {
        const ClutterEvent* event = clutter_gesture_action_get_last_event(CLUTTER_GESTURE_ACTION(action), 0);
        gesture.modifiers = clutter_event_get_state(event);
        gesture.type = GESTURE_PAN;
        gesture.time = clutter_event_get_time(event);
        clutter_gesture_action_get_motion_coords(CLUTTER_GESTURE_ACTION(action), 0, &gesture.x, &gesture.y);
        clutter_gesture_action_get_motion_delta(CLUTTER_GESTURE_ACTION(action), 0, &gesture.dx, &gesture.dy);
    }
    else {
        gesture.type = GESTURE_PAN_INTERPOLATED;
        gesture.time = g_get_monotonic_time();
        clutter_pan_action_get_interpolated_coords(action, &gesture.x, &gesture.y);
        clutter_pan_action_get_interpolated_delta(action, &gesture.dx, &gesture.dy);
    }
    gesture_event_enqueue_or_send(self, &gesture);

    return FALSE;
}

static void
on_pan_stopped(ClutterPanAction    *action,
               ClutterActor        *actor,
               gpointer             data)
{
    MxWebView* self = MX_WEB_VIEW(actor);

    GestureEvent gesture;
    gesture.type = GESTURE_PAN_STOPPED;
    gesture.time = g_get_monotonic_time();
    clutter_pan_action_get_interpolated_coords(action, &gesture.x, &gesture.y);
    gesture_event_enqueue_or_send(self, &gesture);
}

static gboolean on_tap_gesture_begin(ClutterTapAction *action,
                                     ClutterActor     *actor,
                                     gpointer          data)
{
    const ClutterEvent* event = clutter_gesture_action_get_last_event(CLUTTER_GESTURE_ACTION(action), 0);

    /* Cancel tap gestures caused by button presses, WebCore expect them only from touch events */
    if (clutter_event_type(event) == CLUTTER_BUTTON_PRESS)
        return FALSE;

    /* Tap begin gesture events are not enqueued but delivered immediately as they
     * are needed to update the focused frame */

    gint64 time = clutter_event_get_time(event);
    ClutterModifierType modifiers = clutter_event_get_state(event);
    gfloat x, y;
    clutter_gesture_action_get_press_coords(CLUTTER_GESTURE_ACTION(action), 0, &x, &y);
    webkit_iweb_view_gesture_tap_begin(actor, time, x, y, modifiers);

    connect_stage_capture_event(MX_WEB_VIEW(actor));

    return TRUE;
}

static void
on_tap(ClutterTapAction *action,
       ClutterActor     *actor,
       gpointer          data)
{
    MxWebView* self = MX_WEB_VIEW(actor);

    const ClutterEvent* event = clutter_gesture_action_get_last_event(CLUTTER_GESTURE_ACTION(action), 0);
    GestureEvent gesture;
    gesture.type = GESTURE_TAP;
    gesture.time = clutter_event_get_time(event);
    gesture.modifiers = clutter_event_get_state(event);
    clutter_gesture_action_get_press_coords(CLUTTER_GESTURE_ACTION(action), 0, &gesture.x, &gesture.y);
    gesture_event_enqueue_or_send(self, &gesture);
}

static void on_tap_gesture_cancel(ClutterTapAction *action,
                                  ClutterActor     *actor,
                                  gpointer          data)
{
    MxWebView* self = MX_WEB_VIEW(actor);

    /* Check that we're not cancelling an already cancelled gesture */
    if (clutter_gesture_action_get_n_current_points(CLUTTER_GESTURE_ACTION(action)) == 0)
        return;

    const ClutterEvent* event = clutter_gesture_action_get_last_event(CLUTTER_GESTURE_ACTION(action), 0);

    /* Ignore taps being cancelled because they are caused by mouse button presses */
    if (clutter_event_type(event) == CLUTTER_BUTTON_PRESS)
        return;

    GestureEvent gesture;
    gesture.type = GESTURE_TAP_CANCEL;
    gesture.time = clutter_event_get_time(event);
    gesture.modifiers = clutter_event_get_state(event);
    clutter_gesture_action_get_press_coords(CLUTTER_GESTURE_ACTION(action), 0, &gesture.x, &gesture.y);
    gesture_event_enqueue_or_send(self, &gesture);
}

static void mx_web_view_cancel_gestures(WebKitIWebView* webView)
{
    MxWebViewPrivate* priv = MX_WEB_VIEW(webView)->priv;
    clutter_gesture_action_cancel(CLUTTER_GESTURE_ACTION(priv->pan_action));
    clutter_gesture_action_cancel(CLUTTER_GESTURE_ACTION(priv->tap_action));
}
#endif // GESTURE_EVENTS

static void
mx_web_view_init (MxWebView *self)
{
    MxWebViewPrivate* priv = self->priv = MX_WEB_VIEW_GET_PRIVATE(self);
    ClutterActor* actor = CLUTTER_ACTOR(self);

    // We sure want events!
    clutter_actor_set_reactive(actor, TRUE);

    // The WebKitWebPage
    priv->webPage = WEBKIT_WEB_PAGE(g_object_new(WEBKIT_TYPE_WEB_PAGE, "web-view", WEBKIT_IWEB_VIEW(self), 0));

    g_signal_connect(priv->webPage, "notify::settings",
                     G_CALLBACK(settingsChanged), self);

    g_signal_connect(webkit_web_page_get_settings(priv->webPage), "notify::enable-tiled-backing-store",
                     G_CALLBACK(settingsChanged), self);

    WebKitWebFrame* mainFrame = webkit_web_page_get_main_frame(priv->webPage);
    if (mainFrame) {
        g_signal_connect(mainFrame, "contents-size-changed",
                         G_CALLBACK(contents_size_changed_cb), self);
        g_signal_connect(mainFrame, "scroll-offset-changed",
                         G_CALLBACK(scroll_offset_changed), self);
    }

    // The contained ClutterCairoTexture
    priv->texture = clutter_cairo_texture_new(0, 0);
    clutter_actor_set_parent(priv->texture, CLUTTER_ACTOR(self));

    g_signal_connect(priv->texture, "draw",
                     G_CALLBACK(mx_web_view_draw), self);

#if ENABLE(GESTURE_EVENTS)
    priv->queued_gesture_events = g_array_new (FALSE, FALSE, sizeof(GestureEvent));
    g_signal_connect(self, "captured-event", G_CALLBACK (on_captured_event), NULL);

    priv->tap_action = clutter_tap_action_new();
    clutter_actor_add_action(CLUTTER_ACTOR(self), priv->tap_action);

    // Make sure the tap gesture acts before the others so that it can update
    // the focused element as soon as possible
    g_signal_connect(priv->tap_action, "gesture-begin", G_CALLBACK (on_tap_gesture_begin), NULL);
    g_signal_connect(priv->tap_action, "tap", G_CALLBACK (on_tap), NULL);
    g_signal_connect(priv->tap_action, "gesture-cancel", G_CALLBACK (on_tap_gesture_cancel), NULL);

    priv->pan_action = clutter_pan_action_new();
    clutter_pan_action_set_interpolate(CLUTTER_PAN_ACTION(priv->pan_action), TRUE);
    clutter_actor_add_action(CLUTTER_ACTOR(self), priv->pan_action);

    g_signal_connect(priv->pan_action, "gesture-begin", G_CALLBACK (on_pan_gesture_begin), NULL);
    g_signal_connect(priv->pan_action, "pan", G_CALLBACK (on_pan), NULL);
    g_signal_connect(priv->pan_action, "gesture-end", G_CALLBACK (on_pan_gesture_end), NULL);
    g_signal_connect(priv->pan_action, "pan-stopped", G_CALLBACK (on_pan_stopped), NULL);
#endif

#if USE(ACCELERATED_COMPOSITING)
    ClutterActor* rootLayer = webkit_web_page_get_root_layer_actor(priv->webPage);
    if (rootLayer)
        clutter_actor_add_child(CLUTTER_ACTOR(self), rootLayer);
#endif

    clutter_actor_set_clip_to_allocation(CLUTTER_ACTOR(self), true);

    priv->fastZoomLevel = 1.0;
}

ClutterActor *
mx_web_view_new (void)
{
    return CLUTTER_ACTOR(g_object_new (MX_TYPE_WEB_VIEW, NULL));
}

static gboolean mx_web_view_late_invalidate(gpointer data)
{
    MxWebView* view = MX_WEB_VIEW(data);
    MxWebViewPrivate* priv = view->priv;

    if (!priv->texture)
        return FALSE;

    /* Try again later. */
    if (priv->painting)
        return TRUE;

    cairo_rectangle_int_t rect = { priv->dirtyRect.x, priv->dirtyRect.y,
                                   priv->dirtyRect.width, priv->dirtyRect.height };

    priv->dirty = false;
    priv->dirtyRect.x = priv->dirtyRect.y = 0;
    priv->dirtyRect.width = priv->dirtyRect.height = 0;

    clutter_cairo_texture_invalidate_rectangle(CLUTTER_CAIRO_TEXTURE(priv->texture), &rect);
    return FALSE;
}

static void mx_web_view_invalidate_rect(WebKitIWebView* webView,
                                        ClutterGeometry* updateRect)
{
    MxWebView* view = MX_WEB_VIEW(webView);
    MxWebViewPrivate* priv = view->priv;

    if (!priv->texture)
        return;

    if (priv->dirtyRect.width <= 0 || priv->dirtyRect.height <= 0) {
        priv->dirtyRect.x = updateRect->x;
        priv->dirtyRect.y = updateRect->y;
        priv->dirtyRect.width = updateRect->width;
        priv->dirtyRect.height = updateRect->height;
    } else
        clutter_geometry_union(updateRect, &priv->dirtyRect, &priv->dirtyRect);

    if (priv->dirty)
        return;

    priv->dirty = true;

    g_object_ref(view);
    clutter_threads_add_idle_full(CLUTTER_PRIORITY_REDRAW, mx_web_view_late_invalidate, view, g_object_unref);
}

static WebKitWebPage* mx_web_view_get_page(WebKitIWebView* webView)
{
    g_return_val_if_fail(MX_IS_WEB_VIEW(webView), 0);
    return MX_WEB_VIEW(webView)->priv->webPage;
}

static WebKitIPopupMenu* mx_web_view_create_popup_menu(WebKitIWebView* webView)
{
    WebKitIPopupMenu* popupMenu = (WebKitIPopupMenu*)webkit_mx_popup_menu_new();
    mx_dialog_set_transient_parent(MX_DIALOG(popupMenu), clutter_actor_get_stage(CLUTTER_ACTOR(webView)));
    return popupMenu;
}

static void mx_web_view_post_commit_setup(WebKitIWebView* view, gboolean resetScrolling)
{
    MxWebViewPrivate* priv = MX_WEB_VIEW(view)->priv;

    if (!resetScrolling || !priv->hadjustment)
        return;

    g_object_set(priv->hadjustment,
                 "lower", 0.0,
                 "upper", 0.0,
                 "value", 0.0,
                 NULL);

    g_object_set(priv->vadjustment,
                 "lower", 0.0,
                 "upper", 0.0,
                 "value", 0.0,
                 NULL);
}

#if ENABLE(VIDEO)
static guint controlsBoxTimeoutId = 0;
static ClutterActor* stage = 0;
static ClutterActor* controlsBox = 0;
static ClutterActor* videoTexture = 0;

// We cannot include webkitwebpageprivate.h because this is a .c file. Should it?
void webkit_web_page_request_exit_fullscreen_media(WebKitWebPage*);

static void exitFullscreen(WebKitIWebView* view)
{
    if (controlsBoxTimeoutId) {
        g_source_remove(controlsBoxTimeoutId);
        controlsBoxTimeoutId = 0;
    }

    webkit_web_page_request_exit_fullscreen_media(webkit_iweb_view_get_page(view));
}

static gboolean keyReleaseEvent(ClutterActor* actor, ClutterKeyEvent* event, WebKitIWebView* view)
{
    switch (event->keyval) {
    case CLUTTER_KEY_Escape:
        exitFullscreen(view);
        break;
    default:
        return FALSE;
    }

    return TRUE;
}

static void placeControlsBox()
{
    float width, height;
    clutter_actor_get_size(stage, &width, &height);

    float x, y;

    x = width / 2 - (clutter_actor_get_width(controlsBox) / 2);
    y = height - clutter_actor_get_height(controlsBox) - 40;

    clutter_actor_set_position(controlsBox, x, y);
}

static void placeTexture()
{
    float width, height;
    clutter_actor_get_size(stage, &width, &height);
    clutter_actor_set_size(videoTexture, width, height);
}

static void allocationChanged(ClutterActor* stage, GParamSpec* pspec, gpointer data)
{
    placeControlsBox();
    placeTexture();
}

static gboolean hideControlsBox(gpointer data)
{
    ClutterActor* controlsBox = CLUTTER_ACTOR(data);
    clutter_actor_animate(controlsBox, CLUTTER_LINEAR, 500,
                          "opacity", 0, NULL);

    controlsBoxTimeoutId = 0;

    return FALSE;
}

static void motionEvent(ClutterActor* stage, ClutterEvent* event, ClutterActor* controlsBox)
{
    if (controlsBoxTimeoutId)
        g_source_remove(controlsBoxTimeoutId);
    else
        clutter_actor_animate(controlsBox, CLUTTER_LINEAR, 500,
                              "opacity", 255, NULL);

    controlsBoxTimeoutId = g_timeout_add_seconds(3, hideControlsBox, controlsBox);
}

static void mediaPlayingChanged(WebKitDOMHTMLMediaElement* element, GParamSpec* pspec, MxButton* button)
{
    if (webkit_dom_html_media_element_get_paused(element))
        mx_button_set_label(button, "Play");
    else
        mx_button_set_label(button, "Pause");
}

static void mediaProgressChanged(WebKitDOMHTMLMediaElement* element, GParamSpec* pspec, MxSlider* slider)
{
    float current, duration;
    g_object_get(G_OBJECT(element), "current-time", &current, "duration", &duration, NULL);
    g_object_set(G_OBJECT(slider), "value", current / duration, NULL);
}

static void playPauseButtonClicked(MxButton* button, WebKitDOMHTMLMediaElement* element)
{
    if (webkit_dom_html_media_element_get_paused(element))
        webkit_dom_html_media_element_play(element);
    else
        webkit_dom_html_media_element_pause(element);
}

static void exitFullscreenButtonClicked(MxButton* button, WebKitIWebView* view)
{
    exitFullscreen(view);
}

static gboolean mx_web_view_handle_fullscreen_media(WebKitIWebView* view, WebKitDOMHTMLVideoElement* element, ClutterActor* texture)
{
    // First we give the application a shot at handling this.
    gboolean handled;
    g_signal_emit_by_name(view, "media-fullscreen", element, texture, &handled);
    if (handled)
        return TRUE;

    g_assert(!stage);
    g_assert(!videoTexture);
    g_assert(!controlsBox);

    videoTexture = texture;

    stage = clutter_stage_new();
    clutter_container_add_actor(CLUTTER_CONTAINER(stage), texture);

    controlsBox = mx_box_layout_new();
    clutter_container_add_actor(CLUTTER_CONTAINER(stage), controlsBox);

    ClutterActor* button = mx_button_new();
    clutter_container_add_actor(CLUTTER_CONTAINER(controlsBox), button);

    if (webkit_dom_html_media_element_get_paused(WEBKIT_DOM_HTML_MEDIA_ELEMENT(element)))
        mx_button_set_label(MX_BUTTON(button), "Play");
    else
        mx_button_set_label(MX_BUTTON(button), "Pause");

    g_signal_connect(button, "clicked",
                     G_CALLBACK(playPauseButtonClicked), element);

    ClutterActor* slider = mx_slider_new();
    clutter_container_add_actor(CLUTTER_CONTAINER(controlsBox), slider);

    g_signal_connect(element, "notify::current-time",
                     G_CALLBACK(mediaProgressChanged), slider);

    ClutterActor* exitButton = mx_button_new_with_label("Exit Fullscreen");
    clutter_container_add_actor(CLUTTER_CONTAINER(controlsBox), exitButton);

    g_signal_connect(exitButton, "clicked",
                     G_CALLBACK(exitFullscreenButtonClicked), view);

    // Media events.
    g_signal_connect(element, "notify::paused",
                     G_CALLBACK(mediaPlayingChanged), button);


    g_signal_connect(stage, "notify::allocation",
                     G_CALLBACK(allocationChanged), 0);

    g_signal_connect(stage, "key-release-event",
                     G_CALLBACK(keyReleaseEvent), view);

    g_signal_connect(stage, "motion-event",
                     G_CALLBACK(motionEvent), controlsBox);

    controlsBoxTimeoutId = g_timeout_add_seconds(3, hideControlsBox, controlsBox);

    clutter_actor_show_all(stage);
    clutter_stage_set_fullscreen(CLUTTER_STAGE(stage), TRUE);

    return TRUE;
}

gboolean mx_web_view_handle_exit_fullscreen_media(WebKitIWebView* view,
                                                  ClutterActor* texture)
{
    gboolean handled;
    g_signal_emit_by_name(view, "media-exit-fullscreen", texture, &handled);
    if (handled)
        return TRUE;

    clutter_actor_remove_child(stage, texture);
    clutter_actor_destroy(stage);

    stage = videoTexture = controlsBox = 0;

    return TRUE;
}
#endif

static void pluginDestroyed(gpointer data, GObject* object)
{
    MxWebViewPrivate* priv = MX_WEB_VIEW(data)->priv;
    priv->plugins = g_slist_remove(priv->plugins, object);
}

static void mx_web_view_add_plugin_actor(WebKitIWebView* view, ClutterActor* plugin)
{
    MxWebViewPrivate* priv = MX_WEB_VIEW(view)->priv;
    priv->plugins = g_slist_prepend(priv->plugins, plugin);

    g_object_weak_ref(G_OBJECT(plugin), pluginDestroyed, view);

    clutter_actor_add_child(CLUTTER_ACTOR(view), plugin);
}

#if USE(TILED_BACKING_STORE)
static void tileDestroyed(gpointer data, GObject* object)
{
    MxWebViewPrivate* priv = MX_WEB_VIEW(data)->priv;
    priv->tiles = g_slist_remove(priv->tiles, object);
}

static void mx_web_view_add_tile(WebKitIWebView* view, ClutterActor* tile)
{
    MxWebViewPrivate* priv = MX_WEB_VIEW(view)->priv;
    priv->tiles = g_slist_prepend(priv->tiles, tile);

    g_object_weak_ref(G_OBJECT(tile), tileDestroyed, view);

    clutter_actor_add_child(CLUTTER_ACTOR(view), tile);
}

static void mx_web_view_delegated_scroll(WebKitIWebView* view, int x, int y)
{
    MxWebViewPrivate* priv = MX_WEB_VIEW(view)->priv;

    if (priv->hadjustment) {
        mx_adjustment_set_value(priv->hadjustment, x);
        mx_adjustment_set_value(priv->vadjustment, y);
        return;
    }

    WebKitWebFrame *mainFrame = webkit_web_page_get_main_frame(priv->webPage);

    webkit_web_frame_set_scroll_offset(mainFrame, x, y);
}
#endif

static void webkit_iweb_view_iface_init(WebKitIWebViewInterface* iface)
{
    iface->get_page = mx_web_view_get_page;
    iface->invalidate_rect = mx_web_view_invalidate_rect;
    iface->create_popup_menu = mx_web_view_create_popup_menu;
    iface->post_commit_setup = mx_web_view_post_commit_setup;

#if ENABLE(VIDEO)
    iface->handle_fullscreen_media = mx_web_view_handle_fullscreen_media;
    iface->handle_exit_fullscreen_media = mx_web_view_handle_exit_fullscreen_media;
#endif

#if USE(TILED_BACKING_STORE)
    iface->delegated_scroll = mx_web_view_delegated_scroll;
    iface->add_tile = mx_web_view_add_tile;
#endif

#if ENABLE(GESTURE_EVENTS)
    iface->cancel_gestures = mx_web_view_cancel_gestures;
#endif

    iface->add_plugin_actor = mx_web_view_add_plugin_actor;
}

static void mx_web_view_hadjustment_value_notify_cb(MxAdjustment *adjustment,
                                                    GParamSpec   *pspec,
                                                    MxWebView   *webView)
{
    MxWebViewPrivate *priv = webView->priv;
    WebKitWebFrame *mainFrame = webkit_web_page_get_main_frame(priv->webPage);
    gdouble value;

    value = mx_adjustment_get_value (adjustment);
    webkit_web_frame_set_x_scroll_offset(mainFrame, (int)value);
}

static void vadjustment_value_notify_cb(MxAdjustment *adjustment,
                                        GParamSpec   *arg1,
                                        MxWebView   *webView)
{
    MxWebViewPrivate *priv = webView->priv;
    WebKitWebFrame *mainFrame = webkit_web_page_get_main_frame(priv->webPage);
    gdouble value;

    value = mx_adjustment_get_value (adjustment);
    webkit_web_frame_set_y_scroll_offset(mainFrame, (int)value);
}

static void scrollable_set_adjustments(MxScrollable *scrollable,
                                       MxAdjustment *hadjustment,
                                       MxAdjustment *vadjustment)
{
    MxWebViewPrivate *priv;

    g_return_if_fail (MX_IS_WEB_VIEW (scrollable));

    priv = ((MxWebView *) scrollable)->priv;

    if (hadjustment != priv->hadjustment)
    {
      if (priv->hadjustment)
        {
          g_signal_handlers_disconnect_by_func (priv->hadjustment,
                                                reinterpret_cast<gpointer>(mx_web_view_hadjustment_value_notify_cb),
                                                scrollable);
          g_object_unref (priv->hadjustment);
        }

      if (hadjustment)
        {
          g_object_ref (hadjustment);
          g_signal_connect (hadjustment, "notify::value",
                            G_CALLBACK (mx_web_view_hadjustment_value_notify_cb),
                            scrollable);
        }

      priv->hadjustment = hadjustment;
    }

    if (vadjustment != priv->vadjustment)
    {
      if (priv->vadjustment)
        {
          g_signal_handlers_disconnect_by_func (priv->vadjustment,
                                                reinterpret_cast<gpointer>(vadjustment_value_notify_cb),
                                                scrollable);
          g_object_unref (priv->vadjustment);
        }

      if (vadjustment)
        {
          g_object_ref (vadjustment);
          g_signal_connect (vadjustment, "notify::value",
                            G_CALLBACK (vadjustment_value_notify_cb),
                            scrollable);
        }

      priv->vadjustment = vadjustment;
    }
}

static void scrollable_get_adjustments(MxScrollable  *scrollable,
                                       MxAdjustment **hadjustment,
                                       MxAdjustment **vadjustment)
{
  MxWebViewPrivate *priv;

  g_return_if_fail (MX_IS_WEB_VIEW (scrollable));

  priv = ((MxWebView *) scrollable)->priv;

  if (hadjustment)
    {
      if (priv->hadjustment)
        *hadjustment = priv->hadjustment;
      else
        {
          MxAdjustment *adjustment;

          /* create an initial adjustment. this is filled with correct values
           * as soon as allocate() is called */

          adjustment = mx_adjustment_new ();

          scrollable_set_adjustments (scrollable,
                                      adjustment,
                                      priv->vadjustment);

          g_object_unref (adjustment);

          *hadjustment = adjustment;
        }
    }

  if (vadjustment)
    {
      if (priv->vadjustment)
        *vadjustment = priv->vadjustment;
      else
        {
          MxAdjustment *adjustment;

          /* create an initial adjustment. this is filled with correct values
           * as soon as allocate() is called */

          adjustment = mx_adjustment_new ();

          scrollable_set_adjustments (scrollable,
                                      priv->hadjustment,
                                      adjustment);

          g_object_unref (adjustment);

          *vadjustment = adjustment;
        }
    }
}

static void mx_scrollable_iface_init(MxScrollableIface *iface)
{
  iface->set_adjustments = scrollable_set_adjustments;
  iface->get_adjustments = scrollable_get_adjustments;
}

static ClutterTransition* ensure_transition(ClutterActor* actor, const char *prop_name)
{
    ClutterTransition* transition;
    ClutterTimeline* timeline;
    guint easingDuration = clutter_actor_get_easing_duration(actor);
    ClutterAnimationMode mode = clutter_actor_get_easing_mode(actor);

    transition = clutter_actor_get_transition(actor, prop_name);
    if (transition == NULL) {
        transition = clutter_property_transition_new(prop_name);
        clutter_actor_add_transition(actor, prop_name, transition);
    }

    clutter_transition_set_remove_on_complete(transition, TRUE);

    timeline = CLUTTER_TIMELINE(transition);
    clutter_timeline_set_delay(timeline, clutter_actor_get_easing_delay(actor));
    clutter_timeline_set_duration(timeline, easingDuration);
    clutter_timeline_set_progress_mode(timeline, mode);

    clutter_timeline_rewind(timeline);
    clutter_timeline_start(timeline);
    
    return transition;
}

static void set_fast_zoom_level(MxWebView *webView, double zoomLevel)
{
    MxWebViewPrivate *priv = webView->priv;

    priv->fastZoomLevel = zoomLevel;

    update_child_transform(webView);
}

/**
 * mx_web_view_set_fast_zoom_level:
 * @webView: a #MxWebView
 * @zoomLevel: the zoom scale factor
 *
 * Scales the content without having to re-render it, making it fast enough to
 * be used in animations.
 *
 * The #MxWebView:fast-zoom-level property is animatable.
 */
void mx_web_view_set_fast_zoom_level(MxWebView *webView, double zoomLevel)
{
    MxWebViewPrivate* priv = webView->priv;
    ClutterActor *actor = CLUTTER_ACTOR(webView);
    guint easingDuration = clutter_actor_get_easing_duration(actor);
    ClutterTransition* transition;
    ClutterInterval* interval;
 
    if (priv->fastZoomLevel == zoomLevel)
        return;

    if (easingDuration == 0) {
        clutter_actor_remove_transition(actor, "fast-zoom-level");

        set_fast_zoom_level(webView, zoomLevel);

        return;
    }

    transition = ensure_transition(actor, "fast-zoom-level");

    interval = clutter_interval_new(G_TYPE_DOUBLE,
                                    priv->fastZoomLevel,
                                    zoomLevel);

    clutter_transition_set_interval(transition, interval);
}

static void set_scroll_offset_x(MxWebView *webView, int offset)
{
    MxWebViewPrivate *priv = webView->priv;
    WebKitWebFrame *mainFrame = webkit_web_page_get_main_frame(priv->webPage);

    webkit_web_frame_set_x_scroll_offset(mainFrame, offset);
}

/**
 * mx_web_view_set_x_scroll_offset:
 * @webView: a #MxWebView
 * @offset: the horizontal scroll offset
 *
 * Sets the horizontal scroll offset of the #MxWebView.
 */
void mx_web_view_set_x_scroll_offset(MxWebView* webView, int offset)
{
    MxWebViewPrivate* priv = webView->priv;
    WebKitWebFrame *mainFrame = webkit_web_page_get_main_frame(priv->webPage);
    int currentOffset = webkit_web_frame_get_x_scroll_offset(mainFrame);
    ClutterActor *actor = CLUTTER_ACTOR(webView);
    guint easingDuration = clutter_actor_get_easing_duration(actor);
    ClutterTransition* transition;
    ClutterInterval* interval;
 
    if (currentOffset == offset)
        return;

    if (easingDuration == 0) {
        clutter_actor_remove_transition(actor, "scroll-offset-x");

        set_scroll_offset_x(webView, offset);

        return;
    }

    transition = ensure_transition(actor, "scroll-offset-x");

    interval = clutter_interval_new(G_TYPE_INT, currentOffset, offset);

    clutter_transition_set_interval(transition, interval);
}

static void set_scroll_offset_y(MxWebView *webView, int offset)
{
    MxWebViewPrivate *priv = webView->priv;
    WebKitWebFrame *mainFrame = webkit_web_page_get_main_frame(priv->webPage);

    webkit_web_frame_set_y_scroll_offset(mainFrame, offset);
}

/**
 * mx_web_view_set_y_scroll_offset:
 * @webView: a #MxWebView
 * @offset: the vertical scroll offset
 *
 * Sets the vertical scroll offset of the #MxWebView.
 */
void mx_web_view_set_y_scroll_offset(MxWebView* webView, int offset)
{
    MxWebViewPrivate* priv = webView->priv;
    WebKitWebFrame *mainFrame = webkit_web_page_get_main_frame(priv->webPage);
    int currentOffset = webkit_web_frame_get_y_scroll_offset(mainFrame);
    ClutterActor *actor = CLUTTER_ACTOR(webView);
    guint easingDuration = clutter_actor_get_easing_duration(actor);
    ClutterTransition* transition;
    ClutterInterval* interval;
 
    if (currentOffset == offset)
        return;

    if (easingDuration == 0) {
        clutter_actor_remove_transition(actor, "scroll-offset-y");

        set_scroll_offset_y(webView, offset);

        return;
    }

    transition = ensure_transition(actor, "scroll-offset-y");

    interval = clutter_interval_new(G_TYPE_INT, currentOffset, offset);

    clutter_transition_set_interval(transition, interval);
}
