/*
 *  Copyright (C) 2007, 2008 Holger Hans Peter Freyther
 *  Copyright (C) 2007, 2008, 2009 Christian Dywan <christian@imendio.com>
 *  Copyright (C) 2007 Xan Lopez <xan@gnome.org>
 *  Copyright (C) 2007, 2008 Alp Toker <alp@atoker.com>
 *  Copyright (C) 2008 Jan Alonzo <jmalonzo@unpluggable.com>
 *  Copyright (C) 2008 Gustavo Noronha Silva <gns@gnome.org>
 *  Copyright (C) 2008 Nuanti Ltd.
 *  Copyright (C) 2008, 2009, 2010 Collabora Ltd.
 *  Copyright (C) 2009, 2010 Igalia S.L.
 *  Copyright (C) 2009 Movial Creative Technologies Inc.
 *  Copyright (C) 2009 Bobby Powers
 *  Copyright (C) 2010 Collabora Ltd.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"
#include "webkitwebactor.h"

#include "webkitiwebview.h"
#include "webkitnetworkresponse.h"
#include "webkitwebbackforwardlist.h"
#include "webkitwebinspector.h"
#include "webkitwebwindowfeatures.h"
#include <clutter/clutter.h>
#include <glib/gi18n-lib.h>

struct _WebKitWebActorPrivate {
    gboolean dirty;
    ClutterGeometry dirtyRect;
    gboolean transparent;

    WebKitWebPage* webPage;

#if ENABLE(GESTURE_EVENTS)
    ClutterAction *pan_action;
    ClutterAction *tap_action;
#endif
};

static void webkit_iweb_view_iface_init (WebKitIWebViewInterface* iface);
static void webkit_ipopup_menu_iface_init (WebKitIPopupMenuInterface* iface);

G_DEFINE_TYPE_WITH_CODE(WebKitWebActor, webkit_web_actor, CLUTTER_TYPE_CAIRO_TEXTURE,
                        G_IMPLEMENT_INTERFACE(WEBKIT_TYPE_IWEB_VIEW,
                                              webkit_iweb_view_iface_init)
                        G_IMPLEMENT_INTERFACE(WEBKIT_TYPE_IPOPUP_MENU,
                                              webkit_ipopup_menu_iface_init));

#define WEBKIT_WEB_ACTOR_GET_PRIVATE(obj)    (G_TYPE_INSTANCE_GET_PRIVATE((obj), WEBKIT_TYPE_WEB_ACTOR, WebKitWebActorPrivate))

static void webkit_web_actor_dispose(GObject* object)
{
    WebKitWebActor* webActor = WEBKIT_WEB_ACTOR(object);
    WebKitWebActorPrivate* priv = webActor->priv;

    if (priv->webPage) {
        g_object_unref(priv->webPage);
        priv->webPage = 0;
    }

    G_OBJECT_CLASS(webkit_web_actor_parent_class)->dispose(object);
}

static void webkit_web_actor_finalize(GObject* object)
{
    WEBKIT_WEB_ACTOR(object)->priv->~WebKitWebActorPrivate();
    G_OBJECT_CLASS(webkit_web_actor_parent_class)->finalize(object);
}

static void webkit_web_actor_allocate(ClutterActor *actor,
                                      const ClutterActorBox *box,
                                      ClutterAllocationFlags flags)
{
    WebKitWebActor *webActor = WEBKIT_WEB_ACTOR(actor);

    /* chain up to store the allocation of the actor */
    CLUTTER_ACTOR_CLASS (webkit_web_actor_parent_class)->allocate (actor, box, flags);

    gfloat available_width, available_height;
    clutter_actor_box_get_size(box, &available_width, &available_height);
    clutter_cairo_texture_set_surface_size(CLUTTER_CAIRO_TEXTURE(webActor), available_width, available_height);

    webkit_web_page_set_viewport_size(webActor->priv->webPage,
                                      available_width, available_height);
}

static gboolean webkit_web_actor_get_paint_volume (ClutterActor *actor,
                                                   ClutterPaintVolume *volume)
{
    return clutter_paint_volume_set_from_allocation(volume, actor);
}

static AtkObject* mx_web_view_get_accessible(ClutterActor *actor)
{
    WebKitWebPage* page = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(actor));
    return webkit_web_page_get_accessible(page);
}

static void webkit_web_actor_get_preferred_width(ClutterActor *actor,
                                                gfloat  forHeight,
                                                gfloat *minWidth,
                                                gfloat *naturalWidth)
{
    *minWidth = 0;
    *naturalWidth = 0;
    int width;

    WebKitWebFrame* mainFrame = webkit_web_page_get_main_frame(WEBKIT_WEB_ACTOR(actor)->priv->webPage);
    if (mainFrame)
        if (webkit_web_frame_get_contents_size(mainFrame, &width, NULL))
            *minWidth = *naturalWidth = (gfloat)width;
}

static void webkit_web_actor_get_preferred_height(ClutterActor* actor,
                                                  gfloat  forWidth,
                                                  gfloat* minHeight,
                                                  gfloat* naturalHeight)
{
    *minHeight = 0;
    *naturalHeight = 0;
    int height;

    WebKitWebFrame* mainFrame = webkit_web_page_get_main_frame(WEBKIT_WEB_ACTOR(actor)->priv->webPage);
    if (mainFrame)
        if (webkit_web_frame_get_contents_size(mainFrame, NULL, &height))
            *minHeight = *naturalHeight = (gfloat)height;
}

#if ENABLE(GESTURE_EVENTS)
static gboolean on_pan_gesture_begin(ClutterAction   *action,
                                     ClutterActor    *actor,
                                     gpointer         data)
{
    gfloat x, y;

    clutter_gesture_action_get_press_coords(CLUTTER_GESTURE_ACTION(action), 0, &x, &y);
    webkit_iweb_view_gesture_scroll_begin(actor,
        g_get_monotonic_time(),
        x, y);

    return TRUE;
}

static void on_pan_gesture_end(ClutterAction   *action,
                               ClutterActor    *actor,
                               gpointer         data)
{
    gfloat x, y;

    clutter_gesture_action_get_release_coords(CLUTTER_GESTURE_ACTION(action), 0, &x, &y);

    webkit_iweb_view_gesture_scroll_end(actor,
        g_get_monotonic_time(),
        x, y);
    webkit_iweb_view_gesture_scroll_momentum_begin(actor,
        g_get_monotonic_time(),
        x, y);
}

static gboolean on_pan(ClutterPanAction   *action,
                       ClutterActor       *actor,
                       gboolean            is_interpolated,
                       gpointer            data)
{
    gfloat x, y, dx, dy;
    gint64 time;

    time = g_get_monotonic_time();

    if (!is_interpolated) {
        clutter_gesture_action_get_motion_coords(CLUTTER_GESTURE_ACTION(action), 0, &x, &y);
        clutter_gesture_action_get_motion_delta(CLUTTER_GESTURE_ACTION(action), 0, &dx, &dy);
        webkit_iweb_view_gesture_scroll_update(actor,
            time,
            x, y,
            dx, dy);
    }
    else {
        clutter_pan_action_get_interpolated_coords(action, &x, &y);
        clutter_pan_action_get_interpolated_delta(action, &dx, &dy);
        webkit_iweb_view_gesture_scroll_momentum_update(actor,
            time,
            x, y,
            dx, dy);
    }

    return FALSE;
}

static void on_pan_stopped(ClutterPanAction    *action,
                           ClutterActor        *actor,
                           gpointer             data)
{
    gfloat x, y;

    clutter_pan_action_get_interpolated_coords(action, &x, &y);
    webkit_iweb_view_gesture_scroll_momentum_end(actor,
        g_get_monotonic_time(),
        x, y);
}

static gboolean on_tap_begin(ClutterTapAction *action,
                         ClutterActor     *actor,
                         gpointer          data)
{
    gfloat x, y;

    clutter_gesture_action_get_press_coords(CLUTTER_GESTURE_ACTION(action), 0, &x, &y);

    webkit_iweb_view_gesture_tap_begin(actor,
        g_get_monotonic_time(),
        x, y,
        (ClutterModifierType)0); //FIXME get press eventstate

    return TRUE;
}

static void on_tap(ClutterTapAction *action,
                   ClutterActor     *actor,
                   gpointer          data)
{
    gfloat x, y;

    clutter_gesture_action_get_press_coords(CLUTTER_GESTURE_ACTION(action), 0, &x, &y);

    webkit_iweb_view_gesture_tap(actor,
        g_get_monotonic_time(),
        x, y,
        (ClutterModifierType)0); //FIXME get press eventstate
}

static void on_tap_cancel(ClutterTapAction *action,
                          ClutterActor     *actor,
                          gpointer          data)
{
    gfloat x, y;

    clutter_gesture_action_get_press_coords(CLUTTER_GESTURE_ACTION(action), 0, &x, &y);

    webkit_iweb_view_gesture_tap_cancel(actor,
        g_get_monotonic_time(),
        x, y,
        (ClutterModifierType)0); //FIXME get press eventstate
}

static void webkit_web_actor_cancel_gestures(WebKitIWebView* webView)
{
    WebKitWebActorPrivate* priv = WEBKIT_WEB_ACTOR(webView)->priv;
    clutter_gesture_action_cancel(CLUTTER_GESTURE_ACTION(priv->pan_action));
    clutter_gesture_action_cancel(CLUTTER_GESTURE_ACTION(priv->tap_action));
}
#endif // GESTURE_EVENTS

gboolean gesture_is_related_to_event(ClutterGestureAction* gesture,
                                     ClutterEvent* event)
{
    ClutterEventSequence *evt_seq = clutter_event_get_event_sequence(event);
    gint evt_dev_id = clutter_event_get_device_id(event);
    ClutterEventSequence *gesture_seq = clutter_gesture_action_get_sequence(gesture, 0);
    ClutterInputDevice *gesture_dev = clutter_gesture_action_get_device(gesture, 0);

    return evt_seq == gesture_seq &&
        evt_dev_id == clutter_input_device_get_device_id(gesture_dev);
}

gboolean button_press_event(ClutterActor* actor, ClutterButtonEvent* event)
{
    WebKitWebActorPrivate* priv = WEBKIT_WEB_ACTOR_GET_PRIVATE(actor);
    gboolean handled = webkit_iweb_view_button_press_event(actor, event);
#if ENABLE(GESTURE_EVENTS)
    if (handled &&
        gesture_is_related_to_event(CLUTTER_GESTURE_ACTION(priv->pan_action), (ClutterEvent*)event)) {
        clutter_gesture_action_cancel(CLUTTER_GESTURE_ACTION(priv->pan_action));
    }

    /* Ignore taps caused by button presses */
    clutter_gesture_action_cancel(CLUTTER_GESTURE_ACTION(priv->tap_action));
#endif
    return handled;
}

gboolean touch_event(ClutterActor* actor, ClutterTouchEvent* event)
{
    WebKitWebActorPrivate* priv = WEBKIT_WEB_ACTOR_GET_PRIVATE(actor);
    gboolean handled = webkit_iweb_view_touch_event(actor, event);
#if ENABLE(GESTURE_EVENTS)
    if (handled && clutter_event_type((ClutterEvent*)event) == CLUTTER_TOUCH_BEGIN) {
        if (gesture_is_related_to_event(CLUTTER_GESTURE_ACTION(priv->pan_action), (ClutterEvent*)event))
            clutter_gesture_action_cancel(CLUTTER_GESTURE_ACTION(priv->pan_action));
        if (gesture_is_related_to_event(CLUTTER_GESTURE_ACTION(priv->tap_action), (ClutterEvent*)event))
            clutter_gesture_action_cancel(CLUTTER_GESTURE_ACTION(priv->tap_action));
    }
#endif
    return handled;
}

static void webkit_web_actor_class_init(WebKitWebActorClass* webActorClass)
{
    GObjectClass* objectClass = G_OBJECT_CLASS(webActorClass);
    objectClass->dispose = webkit_web_actor_dispose;
    objectClass->finalize = webkit_web_actor_finalize;

    ClutterActorClass* actorClass = CLUTTER_ACTOR_CLASS(webActorClass);
    actorClass->allocate = webkit_web_actor_allocate;
    actorClass->get_accessible = mx_web_view_get_accessible;
    actorClass->get_paint_volume = webkit_web_actor_get_paint_volume;
    actorClass->get_preferred_width = webkit_web_actor_get_preferred_width;
    actorClass->get_preferred_height = webkit_web_actor_get_preferred_height;
    actorClass->button_press_event = button_press_event;
    actorClass->button_release_event = webkit_iweb_view_button_release_event;
    actorClass->motion_event = webkit_iweb_view_motion_event;
    actorClass->scroll_event = webkit_iweb_view_scroll_event;
    actorClass->key_press_event = webkit_iweb_view_key_event;
    actorClass->key_release_event = webkit_iweb_view_key_event;
#if ENABLE(TOUCH_EVENTS)
    actorClass->touch_event = touch_event;
#endif
    actorClass->key_focus_in = webkit_iweb_view_key_focus_in;
    actorClass->key_focus_out = webkit_iweb_view_key_focus_out;

    g_type_class_add_private(webActorClass, sizeof(WebKitWebActorPrivate));
}

static void webkit_web_actor_init(WebKitWebActor* webActor)
{
    // We sure want events!
    clutter_actor_set_reactive(CLUTTER_ACTOR(webActor), TRUE);

    WebKitWebActorPrivate* priv = WEBKIT_WEB_ACTOR_GET_PRIVATE(webActor);
    webActor->priv = priv;
    new (priv) WebKitWebActorPrivate();

    // The WebKitWebPage
    priv->webPage = WEBKIT_WEB_PAGE(g_object_new(WEBKIT_TYPE_WEB_PAGE, "web-view", WEBKIT_IWEB_VIEW(webActor), 0));

#if ENABLE(GESTURE_EVENTS)
    priv->tap_action = clutter_tap_action_new();
    clutter_actor_add_action(CLUTTER_ACTOR(webActor), priv->tap_action);

    g_signal_connect(priv->tap_action, "gesture-begin", G_CALLBACK (on_tap_begin), NULL);
    g_signal_connect(priv->tap_action, "tap", G_CALLBACK (on_tap), NULL);
    g_signal_connect(priv->tap_action, "gesture-cancel", G_CALLBACK (on_tap_cancel), NULL);

    priv->pan_action = clutter_pan_action_new();
    clutter_pan_action_set_interpolate(CLUTTER_PAN_ACTION(priv->pan_action), TRUE);
    clutter_actor_add_action(CLUTTER_ACTOR(webActor), priv->pan_action);

    g_signal_connect(priv->pan_action, "gesture-begin", G_CALLBACK (on_pan_gesture_begin), NULL);
    g_signal_connect(priv->pan_action, "pan", G_CALLBACK (on_pan), NULL);
    g_signal_connect(priv->pan_action, "gesture-end", G_CALLBACK (on_pan_gesture_end), NULL);
    g_signal_connect(priv->pan_action, "pan-stopped", G_CALLBACK (on_pan_stopped), NULL);
#endif
}

ClutterActor* webkit_web_actor_new(void)
{
    WebKitWebActor* webActor = WEBKIT_WEB_ACTOR(g_object_new(WEBKIT_TYPE_WEB_ACTOR, NULL));

    return CLUTTER_ACTOR(webActor);
}

static gboolean webkit_web_actor_draw(gpointer data)
{
    WebKitWebActor* webActor = WEBKIT_WEB_ACTOR(data);

    // Do not try to paint if we are not visible, otherwise we crash
    // when running "headless".
    if (!clutter_actor_get_paint_visibility(CLUTTER_ACTOR(webActor)))
        return FALSE;

    WebKitWebActorPrivate* priv = webActor->priv;

    if (!priv->dirty || !priv->webPage)
        return FALSE;

    ClutterCairoTexture *cairoTexture = CLUTTER_CAIRO_TEXTURE(webActor);
    ClutterGeometry rect = {0, 0, 0, 0};

    if (priv->dirtyRect.width <= 0 || priv->dirtyRect.height <= 0)
        clutter_cairo_texture_get_surface_size(cairoTexture, &rect.width, &rect.height);
    else {
        rect.x = priv->dirtyRect.x;
        rect.y = priv->dirtyRect.y;
        rect.width = priv->dirtyRect.width;
        rect.height = priv->dirtyRect.height;
    }

    cairo_t* cr = clutter_cairo_texture_create_region(cairoTexture,
                    rect.x, rect.y, rect.width, rect.height);

    WebKitWebFrame* mainFrame = webkit_web_page_get_main_frame(priv->webPage);
    if (mainFrame)
        webkit_web_frame_render(mainFrame, cr,
            rect.x, rect.y, rect.width, rect.height);

    cairo_destroy(cr);

    priv->dirty = false;
    priv->dirtyRect.x = priv->dirtyRect.y = 0;
    priv->dirtyRect.width = priv->dirtyRect.height = 0;
    return FALSE;
}

static void webkit_web_actor_invalidate_rect(WebKitIWebView* webActor,
                                             ClutterGeometry* updateRect)
{
    WebKitWebActorPrivate* priv = WEBKIT_WEB_ACTOR(webActor)->priv;

    if (priv->dirtyRect.width <= 0 || priv->dirtyRect.height <= 0) {
        priv->dirtyRect.x = updateRect->x;
        priv->dirtyRect.y = updateRect->y;
        priv->dirtyRect.width = updateRect->width;
        priv->dirtyRect.height = updateRect->height;
    } else
        clutter_geometry_union(updateRect, &priv->dirtyRect, &priv->dirtyRect);

    if (priv->dirty)
        return;

    priv->dirty = true;

    g_object_ref(webActor);
    clutter_threads_add_idle_full(CLUTTER_PRIORITY_REDRAW, webkit_web_actor_draw, webActor, g_object_unref);
}

static WebKitWebPage* webkit_web_actor_get_page(WebKitIWebView* webActor)
{
    g_return_val_if_fail(WEBKIT_IS_WEB_ACTOR(webActor), 0);
    return WEBKIT_WEB_ACTOR(webActor)->priv->webPage;
}

static WebKitIPopupMenu* webkit_web_actor_create_popup_menu(WebKitIWebView* view)
{
    WebKitWebActor* actor = WEBKIT_WEB_ACTOR(view);
    return WEBKIT_IPOPUP_MENU(actor);
}

static gboolean webkit_web_actor_handle_fullscreen_media(WebKitIWebView* webView, WebKitDOMHTMLVideoElement* element, ClutterActor* texture)
{
    // First we give the application a shot at handling this. Returning FALSE here will trigger the
    // default implementation in WebKitIWebView.
    gboolean handled;
    g_signal_emit_by_name(webView, "media-fullscreen", element, texture, &handled);
    return handled;
}

static gboolean webkit_web_actor_handle_exit_fullscreen_media(WebKitIWebView* webView, ClutterActor* texture)
{
    gboolean handled;
    g_signal_emit_by_name(webView, "media-exit-fullscreen", texture, &handled);
    return handled;
}

static void webkit_iweb_view_iface_init(WebKitIWebViewInterface* iface)
{
    iface->get_page = webkit_web_actor_get_page;
    iface->invalidate_rect = webkit_web_actor_invalidate_rect;
    iface->create_popup_menu = webkit_web_actor_create_popup_menu;

#if ENABLE(VIDEO)
    iface->handle_fullscreen_media = webkit_web_actor_handle_fullscreen_media;
    iface->handle_exit_fullscreen_media = webkit_web_actor_handle_exit_fullscreen_media;
#endif

#if ENABLE(GESTURE_EVENTS)
    iface->cancel_gestures = webkit_web_actor_cancel_gestures;
#endif
}

void webkit_web_actor_popup_show(WebKitIPopupMenu* popup, guint x, guint y)
{
    printf("%s: (%d, %d)\n", __FUNCTION__, x, y);
    return;
}

void webkit_web_actor_popup_add_item(WebKitIPopupMenu* popup,
                                     gint index,
                                     WebKitIPopupMenuItemData* item)
{
    printf("%s: %d %s\n", __FUNCTION__, index,
            (item->type == popupMenuSeparator)? "separator" : item->text);
    return;
}

void webkit_web_actor_popup_hide(WebKitIPopupMenu* popup)
{
    printf("%s\n", __FUNCTION__);
    g_signal_emit_by_name(popup, "hidden", NULL);
    return;
}

static void webkit_ipopup_menu_iface_init(WebKitIPopupMenuInterface* iface)
{
    iface->show = webkit_web_actor_popup_show;
    iface->add_item = webkit_web_actor_popup_add_item;
    iface->hide = webkit_web_actor_popup_hide;
}
