/*
 *  Copyright (C) 2010 Collabora Ltd.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"
#include "webkitdefaultrendertheme.h"

#include "CSSValueKeywords.h"
#include <math.h>

using namespace WebCore;

static void webkit_irender_theme_iface_init (WebKitIRenderThemeInterface* iface);

G_DEFINE_TYPE_WITH_CODE(WebKitDefaultRenderTheme, webkit_default_render_theme, G_TYPE_OBJECT,
                        G_IMPLEMENT_INTERFACE(WEBKIT_TYPE_IRENDER_THEME,
                                              webkit_irender_theme_iface_init));

#define HIGHLIGHT_VALUE 0.44

static void webkit_default_render_theme_class_init(WebKitDefaultRenderThemeClass* klass)
{
}

static void webkit_default_render_theme_init(WebKitDefaultRenderTheme* self)
{
}

WebKitDefaultRenderTheme* webkit_default_render_theme_new(void)
{
    return WEBKIT_DEFAULT_RENDER_THEME(g_object_new(WEBKIT_TYPE_DEFAULT_RENDER_THEME, NULL));
}

static guint32 get_platform_color(WebKitIRenderTheme* theme,
                                  WebKitIRenderThemePlatformColor color)
{
    switch (color) {
    case platformColorActiveSelectionForeground:
    case platformColorActiveListBoxSelectionForeground:
    case platformColorInactiveSelectionForeground:
    case platformColorInactiveListBoxSelectionForeground:
        return 0xFFFFFF00;
    case platformColorActiveSelectionBackground:
    case platformColorInactiveSelectionBackground:
        return 0x00000000;
    default:
        return 0;
    }

    return 0;
}

static guint32 get_system_color(WebKitIRenderTheme* theme,
                                int cssValueId)
{
    // Values copied over and converted to clutter convention from
    // WebCore's RenderTheme.cpp.
    switch (cssValueId) {
        case CSSValueActiveborder:
            return 0xFFFFFF00;
        case CSSValueActivecaption:
            return 0xCCCCCC00;
        case CSSValueAppworkspace:
            return 0xFFFFFF00;
        case CSSValueBackground:
            return 0x6363CE00;
        case CSSValueButtontext:
            return 0xFFFFFF00;
        case CSSValueCaptiontext:
            return 0x00000000;
        case CSSValueGraytext:
            return 0x80808000;
        case CSSValueHighlight:
            return 0xB5D5FF00;
        case CSSValueHighlighttext:
            return 0xFFFFFF00;
        case CSSValueInactiveborder:
            return 0xFFFFFF00;
        case CSSValueInactivecaption:
            return 0xFFFFFF00;
        case CSSValueInactivecaptiontext:
            return 0x7F7F7F00;
        case CSSValueInfobackground:
            return 0xFBFCC500;
        case CSSValueInfotext:
            return 0x00000000;
        case CSSValueMenu:
            return 0xC0C0C000;
        case CSSValueMenulistText:
            return 0xFFFFFF00;
        case CSSValueMenutext:
            return 0x00000000;
        case CSSValueScrollbar:
            return 0xFFFFFF00;
        case CSSValueText:
            return 0xFFFFFF00;
        case CSSValueThreeddarkshadow:
            return 0x66666600;
        case CSSValueThreedface:
            return 0xC0C0C000;
        case CSSValueThreedhighlight:
            return 0xDDDDDD00;
        case CSSValueThreedlightshadow:
            return 0xC0C0C000;
        case CSSValueThreedshadow:
            return 0x88888800;
        case CSSValueWindow:
            return 0xFFFFFF00;
        case CSSValueWindowframe:
            return 0xCCCCCC00;
        case CSSValueWindowtext:
            return 0x00000000;
    }
    return 0;
}

static gboolean get_widget_metrics(WebKitIRenderTheme* theme,
                                   WebKitIRenderThemeWidgetMetrics* metrics)
{
    metrics->blinkInterval = 0.5;
    metrics->menulistButtonWidth = 13;
    metrics->menulistButtonHeight = 13;

    return TRUE;
}

static gboolean get_scrollbar_metrics(WebKitIRenderTheme* theme,
                                      WebKitIRenderThemeScrollbarMetrics* metrics)
{
    metrics->has_buttons = TRUE;
    metrics->stepper_size = 13;
    metrics->slider_width = 13;
    metrics->trough_border = 0;
    metrics->min_slider_size = 13;

    return TRUE;
}

static void paintHighlightedBox(cairo_t* cr, cairo_rectangle_int_t* rect)
{
    cairo_save(cr);
    cairo_rectangle(cr, rect->x, rect->y, rect->width, rect->height);
    cairo_clip(cr);
    cairo_set_source_rgb(cr, HIGHLIGHT_VALUE, HIGHLIGHT_VALUE, HIGHLIGHT_VALUE);
    cairo_paint(cr);
    cairo_restore(cr);
}

static void paintBlackBox(cairo_t* cr, cairo_rectangle_int_t* rect)
{
    cairo_save(cr);
    cairo_rectangle(cr, rect->x, rect->y, rect->width, rect->height);
    cairo_clip(cr);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_paint(cr);
    cairo_restore(cr);
}

static void paintScrollbarBackground(cairo_t* cr, cairo_rectangle_int_t* rect)
{
    paintBlackBox(cr, rect);
}

static void paintTrackBackground(cairo_t* cr, cairo_rectangle_int_t* rect)
{
    cairo_save(cr);
    cairo_rectangle(cr, rect->x, rect->y, rect->width, rect->height);
    cairo_clip(cr);
    cairo_set_source_rgb(cr, 0.6, 0.6, 0.6);
    cairo_paint(cr);
    cairo_restore(cr);
}

static void paintBox(cairo_t* cr, cairo_rectangle_int_t* rect, WebKitIRenderThemeWidgetState* state)
{
    if (!state->disabled && state->inHover)
        paintHighlightedBox(cr, rect);
    else
        paintBlackBox(cr, rect);
}

static void paintMediaSliderThumb(cairo_t* cr, cairo_rectangle_int_t* rect, WebKitIRenderThemeWidgetState* state)
{
    cairo_rectangle_int_t thumbRect = *rect;

    int yOffset = (thumbRect.height / 12);

    if (yOffset < 1)
        yOffset = 1;

    thumbRect.y += yOffset;
    thumbRect.height -= 2 * yOffset;

    paintBox(cr, &thumbRect, state);
}

static void paintScrollbarThumb(cairo_t* cr, cairo_rectangle_int_t* rect, WebKitIRenderThemeWidgetState* state)
{
    paintBox(cr, rect, state);
}

static void paintArrowButton(cairo_t* cr, cairo_rectangle_int_t* rect, gint flags, WebKitIRenderThemeWidgetState* state)
{
    // Background
    paintBox(cr, rect, state);

    // Paint the "arrow"
    double start_x, start_y;
    double edge_x, edge_y;
    double end_x, end_y;

    double xOffset = rect->width / 8.0;
    double yOffset = rect->height / 8.0;

    if (flags & stepperVertical) {
        if (flags & stepperBottom) {
            start_x = rect->x + xOffset;
            start_y = rect->y + yOffset;
            edge_x = rect->x + (rect->width/2);
            edge_y = rect->y + rect->height - yOffset;
            end_x = rect->x + rect->width - xOffset;
            end_y = rect->y + yOffset;
        } else {
            start_x = rect->x + xOffset;
            start_y = rect->y + rect->height - yOffset;
            edge_x = rect->x + (rect->width / 2);
            edge_y = rect->y + yOffset;
            end_x = rect->x + rect->width - xOffset;
            end_y = rect->y + rect->height - yOffset;
        }
    } else {
        if (flags & stepperBottom) {
            start_x = rect->x + xOffset;
            start_y = rect->y + yOffset;
            edge_x = rect->x + rect->width - xOffset;
            edge_y = rect->y + (rect->height / 2);
            end_x = rect->x + xOffset;
            end_y = rect->y + rect->height - yOffset;
        } else {
            start_x = rect->x + rect->width - xOffset;
            start_y = rect->y + xOffset;
            edge_x = rect->x + xOffset;
            edge_y = rect->y + (rect->height / 2);
            end_x = rect->x + rect->width - xOffset;
            end_y = rect->y + rect->height - xOffset;
        }
    }

    cairo_save(cr);

    if (state->disabled)
        cairo_set_source_rgb(cr, 0.4, 0.4, 0.4);
    else
        cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);

    cairo_move_to(cr, start_x, start_y);
    cairo_line_to(cr, edge_x, edge_y);
    cairo_line_to(cr, end_x, end_y);
    cairo_line_to(cr, start_x, start_y);
    cairo_fill(cr);
    cairo_restore(cr);
}

static void paintScrollbarButton(cairo_t* cr, cairo_rectangle_int_t* rect, gint flags, WebKitIRenderThemeWidgetState* state)
{
    paintArrowButton(cr, rect, flags, state);
}

static void paintFlatButton(cairo_t* cr, cairo_rectangle_int_t* rect,
                            gint flags, WebKitIRenderThemeWidgetState* state)
{
    cairo_save(cr);

    cairo_rectangle(cr, rect->x, rect->y, rect->width, rect->height);
    cairo_clip(cr);

    // Highlight the button when it is hovered.
    gfloat greyTone = 0.0;
    if (!state->disabled && state->inHover)
        greyTone = HIGHLIGHT_VALUE;

    cairo_set_source_rgb(cr, greyTone, greyTone, greyTone);

    cairo_paint(cr);
    cairo_restore(cr);
}

static void paintButton(cairo_t* cr, cairo_rectangle_int_t* rect,
                        gint flags, WebKitIRenderThemeWidgetState* state)
{
    cairo_rectangle_int_t buttonRect = *rect;

    buttonRect.width -= 2;
    buttonRect.height -= 2;

    if (state->active) {
        buttonRect.x += 2;
        buttonRect.y += 2;
    }

    paintFlatButton(cr, &buttonRect, flags, state);
}

static void paintSpinButton(cairo_t* cr, cairo_rectangle_int_t* rect, WebKitIRenderThemeWidgetState* state)
{
    paintBlackBox(cr, rect);
}

static void paintSpinButtonUp(cairo_t* cr, cairo_rectangle_int_t* rect, WebKitIRenderThemeWidgetState* state)
{
    paintArrowButton(cr, rect, stepperVertical, state);
}

static void paintSpinButtonDown(cairo_t* cr, cairo_rectangle_int_t* rect, WebKitIRenderThemeWidgetState* state)
{
    paintArrowButton(cr, rect, stepperVertical | stepperBottom, state);
}

static void paintSearchFieldResultsButton(cairo_t* cr, cairo_rectangle_int_t* rect, WebKitIRenderThemeWidgetState* state)
{
    double lineWidth = 2.;
    double padding = 2.;
    cairo_rectangle_t r = { rect->x + padding, rect->y + padding,
                            rect->width - padding*2, rect->height - padding*2};
    cairo_save(cr);
    cairo_rectangle(cr, r.x, r.y, r.width, r.height);
    cairo_clip(cr);
    cairo_set_line_width(cr, lineWidth);
    cairo_set_source_rgb(cr, 0., 0., 0.);
    cairo_arc(cr, r.x + r.width/3. + lineWidth/2, r.y + r.height/3. + lineWidth/2,
              r.width/3., 0., 2 * M_PI);
    cairo_move_to(cr, r.x + 2*r.width/3., r.y + 2*r.height/3.);
    cairo_line_to(cr, r.x + r.width, r.y + r.height);
    cairo_stroke(cr);
    cairo_restore(cr);
}

static void paintSearchFieldCancelButton(cairo_t* cr, cairo_rectangle_int_t* rect, WebKitIRenderThemeWidgetState* state)
{
    double lineWidth = 2.;
    double padding = 2.;
    cairo_rectangle_t r = { rect->x+padding, rect->y+padding, rect->width-padding*2, rect->height-padding*2};
    cairo_save(cr);
    cairo_rectangle(cr, r.x, r.y, r.width, r.height);
    cairo_clip(cr);
    cairo_set_line_width(cr, lineWidth);
    cairo_set_source_rgb(cr, 0., 0., 0.);
    cairo_move_to(cr, r.x, r.y);
    cairo_line_to(cr, r.x + r.width, r.y + r.height);
    cairo_move_to(cr, r.x, r.y + r.height);
    cairo_line_to(cr, r.x + r.width, r.y);
    cairo_stroke(cr);
    cairo_restore(cr);
}

static void paintCheckButton(cairo_t* cr, cairo_rectangle_int_t* rect,
                             gint isChecked, WebKitIRenderThemeWidgetState* state)
{
    cairo_save(cr);
    cairo_rectangle(cr, rect->x, rect->y, rect->width, rect->height);
    cairo_clip(cr);
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_paint(cr);
    cairo_restore(cr);

    float greyTone = 0.0;
    if (state->inHover)
        greyTone = HIGHLIGHT_VALUE;

    cairo_save(cr);
    cairo_rectangle(cr, rect->x, rect->y, rect->width, rect->height);
    cairo_set_line_width(cr, 2.0);
    cairo_set_source_rgb(cr, greyTone, greyTone, greyTone);
    cairo_stroke(cr);
    cairo_restore(cr);

    if (isChecked) {
        cairo_save(cr);
        cairo_set_line_width(cr, 2.0);
        cairo_set_source_rgb(cr, greyTone, greyTone, greyTone);
        cairo_move_to(cr, rect->x + 3, rect->y + rect->height / 2);
        cairo_line_to(cr, rect->x + rect->width / 2, rect->y + rect->height - 3);
        cairo_line_to(cr, rect->x + rect->width - 3, rect->y + 3);
        cairo_stroke(cr);
        cairo_restore(cr);
    }
}

static void paintRadioButton(cairo_t* cr, cairo_rectangle_int_t* rect,
                             gint isChecked, WebKitIRenderThemeWidgetState* state)
{
    float greyTone = 0.0;
    if (state->inHover)
        greyTone = HIGHLIGHT_VALUE;

    cairo_save(cr);
    cairo_arc(cr, rect->x + rect->width / 2, rect->y + rect->height / 2, rect->width / 2, 0.0, 2 * M_PI);
    cairo_set_line_width(cr, 2.0);
    cairo_set_source_rgb(cr, greyTone, greyTone, greyTone);
    cairo_stroke(cr);
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_fill(cr);
    cairo_restore(cr);

    if (isChecked) {
        cairo_save(cr);
        cairo_arc(cr, rect->x + rect->width / 2, rect->y + rect->height / 2, rect->width / 2 - 3, 0.0, 2 * M_PI);
        cairo_set_source_rgb(cr, greyTone, greyTone, greyTone);
        cairo_fill(cr);
        cairo_restore(cr);
    }
}

static void paintMenuList(cairo_t* cr, cairo_rectangle_int_t* rect, WebKitIRenderThemeWidgetState* state)
{
    paintBox(cr, rect, state);
}

static void paintMenuListArrow(cairo_t* cr, cairo_rectangle_int_t* rect, WebKitIRenderThemeWidgetState* state)
{
    paintArrowButton(cr, rect, stepperVertical | stepperBottom, state);
}

static void paintMediaFullscreenButton(cairo_t* cr, cairo_rectangle_int_t* rect,
                                       gint flags, WebKitIRenderThemeWidgetState* state)
{
    paintFlatButton(cr, rect, 0, state);

    cairo_rectangle_int_t innerRect;

    int xOffset = rect->width / 12;
    if (xOffset < 2)
        xOffset = 2;

    innerRect.x = rect->x + xOffset;
    innerRect.width = rect->width - (2 * xOffset);

    int yOffset = rect->height / 12;
    if (yOffset < 2)
        yOffset = 2;

    innerRect.y = rect->y + yOffset;
    innerRect.height = rect->height - (2 * yOffset);

    // Top-left.
    cairo_save(cr);
    cairo_set_line_width(cr, 1.5);
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_move_to(cr, innerRect.x, innerRect.y + (innerRect.height / 2) - yOffset);
    cairo_line_to(cr, innerRect.x, innerRect.y);
    cairo_line_to(cr, innerRect.x + (innerRect.width / 2) - xOffset, innerRect.y);
    cairo_stroke(cr);

    // Top-right.
    cairo_save(cr);
    cairo_set_line_width(cr, 1.5);
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_move_to(cr, innerRect.x + (innerRect.width / 2) + xOffset, innerRect.y);
    cairo_line_to(cr, innerRect.x + innerRect.width, innerRect.y);
    cairo_line_to(cr, innerRect.x + innerRect.width, innerRect.y + (innerRect.height / 2) - yOffset);
    cairo_stroke(cr);

    // Bottom-left.
    cairo_save(cr);
    cairo_set_line_width(cr, 1.5);
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_move_to(cr, innerRect.x, innerRect.y + (innerRect.height / 2) + yOffset);
    cairo_line_to(cr, innerRect.x, innerRect.y + innerRect.height);
    cairo_line_to(cr, innerRect.x + (innerRect.width / 2) - xOffset, innerRect.y + innerRect.height);
    cairo_stroke(cr);

    // Bottom-right.
    cairo_save(cr);
    cairo_set_line_width(cr, 1.5);
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_move_to(cr, innerRect.x + (innerRect.width / 2) + xOffset, innerRect.y + innerRect.height);
    cairo_line_to(cr, innerRect.x + innerRect.width, innerRect.y + innerRect.height);
    cairo_line_to(cr, innerRect.x + innerRect.width, innerRect.y + (innerRect.height / 2) + yOffset);
    cairo_stroke(cr);
}

static void paintMediaMuteButton(cairo_t* cr, cairo_rectangle_int_t* rect,
                                 gint flags, WebKitIRenderThemeWidgetState* state)
{
    paintFlatButton(cr, rect, flags, state);

    double positionModifier = rect->width / 8.0;
    double radiusModifier = 0.2;

    double position = positionModifier;
    double radius = 0.2;

    for (int i = 0; i < 5; i++) {
        cairo_save(cr);
        cairo_translate(cr, rect->x + position, rect->y + (rect->height / 2.0));
        cairo_scale(cr, rect->width / 4.0, rect->height / 2);
        cairo_set_line_width(cr, 0.2);
        cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
        cairo_arc(cr, 0, 0, radius, - M_PI / 2, M_PI / 2);
        cairo_stroke(cr);
        cairo_restore(cr);

        position += positionModifier;
        radius += radiusModifier;
    }

    if (!(flags & mediaMuted))
        return;

    cairo_save(cr);
    cairo_set_line_width(cr, 1.5);
    cairo_set_source_rgb(cr, 1.0, 0.0, 0.0);
    cairo_move_to(cr, rect->x, rect->y);
    cairo_line_to(cr, rect->x + rect->width, rect->y + rect->height);
    cairo_move_to(cr, rect->x, rect->y + rect->height);
    cairo_line_to(cr, rect->x + rect->width, rect->y);
    cairo_stroke(cr);
    cairo_restore(cr);
}

static void paintMediaPlayButton(cairo_t* cr, cairo_rectangle_int_t* rect,
                                 gint flags, WebKitIRenderThemeWidgetState* state)
{
    if (!(flags & mediaPlaying)) {
        paintArrowButton(cr, rect, stepperBottom, state);
        return;
    }

    paintFlatButton(cr, rect, 0, state);

    cairo_rectangle_int_t pausePosition = *rect;
    pausePosition.x = rect->x + ((rect->width / 2) - (rect->width / 4));
    pausePosition.y = rect->y + (rect->height / 10);
    pausePosition.height = rect->height - (2 * (rect->height / 10));

    cairo_save(cr);
    cairo_set_line_width(cr, 2.0);
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_move_to(cr, pausePosition.x, pausePosition.y);
    cairo_line_to(cr, pausePosition.x, pausePosition.y + pausePosition.height);
    cairo_stroke(cr);

    pausePosition.x = pausePosition.x + (rect->width / 2);
    cairo_move_to(cr, pausePosition.x, pausePosition.y);
    cairo_line_to(cr, pausePosition.x, pausePosition.y + pausePosition.height);
    cairo_stroke(cr);
    cairo_restore(cr);
}

static void paintDoubleArrowButton(cairo_t* cr, cairo_rectangle_int_t* rect, gint flags, WebKitIRenderThemeWidgetState* state)
{
    // Background
    paintBox(cr, rect, state);

    // Paint the "arrow"
    double start_x, start_y;
    double edge_x, edge_y;
    double end_x, end_y;

    double xOffset = rect->width / 8.0;
    double yOffset = rect->height / 8.0;

    int halfWidth = (rect->width - xOffset) / 2;

    if (flags & stepperBottom) {
        start_x = rect->x + xOffset;
        start_y = rect->y + yOffset;
        edge_x = rect->x + halfWidth;
        edge_y = rect->y + (rect->height / 2);
        end_x = rect->x + xOffset;
        end_y = rect->y + rect->height - yOffset;
    } else {
        start_x = rect->x + rect->width - xOffset;
        start_y = rect->y + xOffset;
        edge_x = rect->x + xOffset + halfWidth;
        edge_y = rect->y + (rect->height / 2);
        end_x = rect->x + rect->width - xOffset;
        end_y = rect->y + rect->height - xOffset;
    }

    cairo_save(cr);

    if (state->disabled)
        cairo_set_source_rgb(cr, 0.4, 0.4, 0.4);
    else
        cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);

    cairo_move_to(cr, start_x, start_y);
    cairo_line_to(cr, edge_x, edge_y);
    cairo_line_to(cr, end_x, end_y);
    cairo_line_to(cr, start_x, start_y);
    cairo_fill(cr);
    cairo_restore(cr);

    if (flags & stepperBottom) {
        start_x += halfWidth;
        edge_x += halfWidth;
        end_x += halfWidth;
    } else {
        start_x -= halfWidth;
        edge_x -= halfWidth;
        end_x -= halfWidth;
    }

    cairo_save(cr);

    if (state->disabled)
        cairo_set_source_rgb(cr, 0.4, 0.4, 0.4);
    else
        cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);

    cairo_move_to(cr, start_x, start_y);
    cairo_line_to(cr, edge_x, edge_y);
    cairo_line_to(cr, end_x, end_y);
    cairo_line_to(cr, start_x, start_y);
    cairo_fill(cr);

    cairo_restore(cr);
}

static void paintMediaSeekBackButton(cairo_t* cr, cairo_rectangle_int_t* rect,
                                     gint flags, WebKitIRenderThemeWidgetState* state)
{
    paintDoubleArrowButton(cr, rect, 0, state);
    return;
}

static void paintMediaSeekForwardButton(cairo_t* cr, cairo_rectangle_int_t* rect,
                                     gint flags, WebKitIRenderThemeWidgetState* state)
{
    paintDoubleArrowButton(cr, rect, stepperBottom, state);
    return;
}

static gboolean paint_widget(WebKitIRenderTheme* theme,
                             WebKitIRenderThemeWidgetType type,
                             WebKitIRenderThemeWidgetState* state,
                             gint flags,
                             cairo_t* cr,
                             cairo_rectangle_int_t* rect)
{
    switch (type) {
    case widgetDropDown:
        paintMenuList(cr, rect, state);
        break;
    case widgetDropDownArrow:
        paintMenuListArrow(cr, rect, state);
        break;
    case widgetButton:
        paintButton(cr, rect, flags, state);
        break;
    case widgetCheckButton:
        paintCheckButton(cr, rect, flags, state);
        break;
    case widgetRadioButton:
        paintRadioButton(cr, rect, flags, state);
        break;
    case widgetSpinButton:
        paintSpinButton(cr, rect, state);
        break;
    case widgetSpinButtonUp:
        paintSpinButtonUp(cr, rect, state);
        break;
    case widgetSpinButtonDown:
        paintSpinButtonDown(cr, rect, state);
        break;
    case widgetSearchFieldDecoration:
    case widgetSearchFieldResultsButton:
        paintSearchFieldResultsButton(cr, rect, state);
        break;
    case widgetSearchFieldCancelButton:
        paintSearchFieldCancelButton(cr, rect, state);
        break;
    case widgetScrollbarTrackHorizontal:
    case widgetScrollbarTrackVertical:
        paintTrackBackground(cr, rect);
        break;
    case widgetScrolledWindow:
        paintScrollbarBackground(cr, rect);
        break;
    case widgetScrollbarButton:
        paintScrollbarButton(cr, rect, flags, state);
        break;
    case widgetScrollbarThumbHorizontal:
    case widgetScrollbarThumbVertical:
        paintScrollbarThumb(cr, rect, state);
        break;
    case widgetMediaFullscreenButton:
        paintMediaFullscreenButton(cr, rect, flags, state);
        break;
    case widgetMediaMuteButton:
        paintMediaMuteButton(cr, rect, flags, state);
        break;
    case widgetMediaPlayButton:
        paintMediaPlayButton(cr, rect, flags, state);
        break;
    case widgetMediaSeekBackButton:
        paintMediaSeekBackButton(cr, rect, flags, state);
        break;
    case widgetMediaSeekForwardButton:
        paintMediaSeekForwardButton(cr, rect, flags, state);
        break;
    case widgetMediaSliderTrack:
        paintTrackBackground(cr, rect);
        break;
    case widgetMediaSliderThumb:
        paintMediaSliderThumb(cr, rect, state);
        break;
    default:
        return FALSE;
    }

    return TRUE;
}

static void webkit_irender_theme_iface_init (WebKitIRenderThemeInterface* iface)
{
    iface->get_platform_color = get_platform_color;
    iface->get_system_color = get_system_color;
    iface->get_widget_metrics = get_widget_metrics;
    iface->get_scrollbar_metrics = get_scrollbar_metrics;
    iface->paint_widget = paint_widget;
}
