/*
 * Copyright (C) 2010 Collabora Ltd.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include <clutter/clutter.h>
#include <webkit/webkitdefines.h>
#include <webkit/webkithittestresult.h>
#include <webkit/webkitiwebview.h>
#include <webkit/vanilla/webkitwebactor.h>

extern void webkit_init();

static ClutterActor* stage = 0;

static void backCallback(ClutterActor* actor, ClutterEvent* event, WebKitIWebView* webView)
{
    webkit_web_page_go_back(webkit_iweb_view_get_page(webView));
}

static void locationChangeCallback(ClutterActor* actor, WebKitIWebView* webView)
{
    webkit_iweb_view_load_uri(webView, clutter_text_get_text(CLUTTER_TEXT(actor)));
}

static WebKitWebActor* createWindowCallback(WebKitWebActor* actor, WebKitWebFrame* frame)
{
    ClutterActor* window = clutter_stage_new();
    clutter_actor_set_size(window, 800, 600);

    ClutterActor* newActor = webkit_web_actor_new();
    clutter_container_add_actor(CLUTTER_CONTAINER(window), newActor);
    clutter_actor_set_size(newActor, 800, 600);

    clutter_actor_show_all(window);

    return WEBKIT_WEB_ACTOR(newActor);
}

static void get_absolute_offset_for_element(WebKitDOMElement* element, glong *x, glong *y)
{
    WebKitDOMElement* offset_parent;
    glong offset_top, offset_left;

    g_object_get(element,
                 "offset-left", &offset_left,
                 "offset-top", &offset_top,
                 "offset-parent", &offset_parent,
                 NULL);

    *x = offset_left;
    *y = offset_top;

    if (!offset_parent)
        return;

    /* If there's a parent, we keep going. */
    glong parent_x, parent_y;
    get_absolute_offset_for_element(offset_parent, &parent_x, &parent_y);

    *x += parent_x;
    *y += parent_y;
}

static gboolean is_zoomed = FALSE;
static glong original_width = 0;

static void load_status_changed_cb(WebKitWebPage* webPage, GParamSpec* pspec, gpointer data)
{
    if (webkit_web_page_get_load_status(webPage) == WEBKIT_LOAD_COMMITTED) {
        webkit_web_page_set_zoom_level(webPage, 0.8);
        is_zoomed = FALSE;
        original_width = 0;
    }
}

static void title_changed_cb(WebKitWebPage* webPage, GParamSpec* pspec, ClutterStage* stage)
{
    clutter_stage_set_title(stage, webkit_web_page_get_title(webPage));
}

static void uri_changed_cb(WebKitWebPage* webPage, GParamSpec* pspec, ClutterText* location_bar)
{
    clutter_text_set_text(location_bar, webkit_web_page_get_uri(webPage));
}

static void progress_changed_cb(WebKitWebPage* webPage, GParamSpec* pspec, ClutterStage* stage)
{
    char* title = g_strdup_printf("%s (%.2f%%)", webkit_web_page_get_title(webPage), webkit_web_page_get_progress(webPage) * 100);
    clutter_stage_set_title(stage, title);
    g_free(title);
}

typedef struct {
    ClutterActor* actor;
    ClutterButtonEvent* event;
} DelayedEventData;

static guint delayed_event_id = 0;
static gboolean firing_synthetic_event = FALSE;

static gboolean delayed_button_press_event(gpointer user_data)
{
    DelayedEventData* data = (DelayedEventData*)user_data;

    firing_synthetic_event = TRUE;

    data->event->time = clutter_get_current_event_time() + 1;
    clutter_actor_event(data->actor, (ClutterEvent*)data->event, FALSE);

    data->event->time += 1;
    data->event->type = CLUTTER_BUTTON_RELEASE;
    clutter_actor_event(data->actor, (ClutterEvent*)data->event, FALSE);

    firing_synthetic_event = FALSE;

    clutter_event_free((ClutterEvent*)data->event);
    g_slice_free(DelayedEventData, data);
    delayed_event_id = 0;

    return FALSE;
}

static gboolean button_release_event_cb(ClutterActor* actor, ClutterEvent* event, gpointer data)
{
    if (delayed_event_id && !firing_synthetic_event)
        return TRUE;

    return FALSE;
}

static gboolean button_press_event_cb(ClutterActor* actor, ClutterEvent* event, gpointer data)
{
    ClutterButtonEvent* buttonEvent = (ClutterButtonEvent*)event;

    /* Delay the mouse press event - we want to make sure double
     * clicks on something WebKit would normally handle as a single click at
     * cause the context zoom to happen.
     */
    if (buttonEvent->click_count == 1 && !firing_synthetic_event) {
        DelayedEventData* dedata = g_slice_new0(DelayedEventData);
        dedata->actor = actor;
        dedata->event = (ClutterButtonEvent*)clutter_event_copy(event);
        delayed_event_id = g_timeout_add(500, delayed_button_press_event, dedata);
        return TRUE;
    }

    if (buttonEvent->click_count == 1 && firing_synthetic_event) {
        gfloat actorX, actorY;
        clutter_actor_transform_stage_point(actor, buttonEvent->x, buttonEvent->y, &actorX, &actorY);

        WebKitWebPage* webPage = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(actor));
        WebKitHitTestResult* result = webkit_web_page_hit_test(webPage, (int)actorX, (int)actorY);

        guint context;
        g_object_get (result, "context", &context, NULL);
        if (!(context & WEBKIT_HIT_TEST_RESULT_CONTEXT_IMAGE))
            return FALSE;

        WebKitDOMNode* node;
        g_object_get(result, "inner-node", &node, NULL);
        if (!node)
            return FALSE;

        // TODO: Are we sure this is an <img>?
        WebKitDOMHTMLImageElement *image = WEBKIT_DOM_HTML_IMAGE_ELEMENT(node);
        webkit_web_page_load_image(webPage, webkit_dom_html_image_element_get_src(image));

        return TRUE;
    }

    if (buttonEvent->click_count != 2)
        return FALSE;

    if (delayed_event_id) {
        g_source_remove(delayed_event_id);
        delayed_event_id = 0;
    }

    WebKitWebPage* webPage = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(actor));

    if (is_zoomed) {
        /* Restore original body width if needed */
        if (original_width != 0) {
            WebKitDOMDocument* document = webkit_web_page_get_dom_document(webPage);
            WebKitDOMHTMLElement* body = webkit_dom_document_get_body(document);

            WebKitDOMCSSStyleDeclaration* style;
            g_object_get(body, "style", &style, NULL);

            char* width_str = g_strdup_printf("%ldpx", original_width);
            webkit_dom_css_style_declaration_set_property(style, "width", width_str, "", NULL);
            g_free(width_str);
        }

        webkit_web_page_set_zoom_level(webPage, 0.8);
        is_zoomed = FALSE;
        return TRUE;
    }

    is_zoomed = TRUE;

    gfloat actorX, actorY;
    clutter_actor_transform_stage_point(actor, buttonEvent->x, buttonEvent->y, &actorX, &actorY);

    WebKitHitTestResult* result = webkit_web_page_hit_test(webPage, (int)actorX, (int)actorY);
    WebKitDOMNode* node;
    g_object_get(result, "inner-node", &node, NULL);

    if (!node)
        return FALSE;

    /* Use the immediate parent that seems to make sense. */
    WebKitDOMElement* parent = webkit_dom_node_get_parent_element(node);
    if (!parent)
        return FALSE;

    while (g_strcmp0("P", webkit_dom_element_get_tag_name(parent)) &&
           g_strcmp0("DIV", webkit_dom_element_get_tag_name(parent)) &&
           g_strcmp0("TD", webkit_dom_element_get_tag_name(parent))) {
        WebKitDOMElement* tmp;
        tmp = webkit_dom_node_get_parent_element(WEBKIT_DOM_NODE(parent));

        /* We'll use what we have */
        if (!tmp)
            break;

        parent = tmp;
    }

    /* First we zoom */
    WebKitDOMDocument* document = webkit_web_page_get_dom_document(webPage);

    float actor_width = clutter_actor_get_width(actor);
    glong width = webkit_dom_element_get_offset_width(parent);

    float proportion = actor_width / width;
    float zoomLevel = proportion;

    if (proportion < 2) {
        WebKitDOMHTMLElement* body = webkit_dom_document_get_body(document);

        /* Save original width */
        glong body_width = webkit_dom_element_get_offset_width(WEBKIT_DOM_ELEMENT(body));
        if ((body_width > actor_width) && ((body_width - actor_width) > 50)) {
            original_width = body_width;

            WebKitDOMCSSStyleDeclaration* style;
            g_object_get(body, "style", &style, NULL);

            char* width_str = g_strdup_printf("%ldpx", (glong)actor_width);
            webkit_dom_css_style_declaration_set_property(style, "width", width_str, "", NULL);
            g_free(width_str);
        }
    } else
        webkit_web_page_set_zoom_level(webPage, zoomLevel);

    /* Then we scroll to focus the element */
    glong x, y;
    get_absolute_offset_for_element(parent, &x, &y);

    /* Do not scroll too close to the node, for better readability */
    if (x > 10)
        x -= 10;

    if (y > 10)
        y -= 10;

    WebKitDOMDOMWindow* dom_window = webkit_dom_document_get_default_view(document);
    webkit_dom_dom_window_scroll(dom_window, x, y);

    g_object_unref(result);

    return TRUE;
}

typedef struct {
    WebKitWebFrame* frame;
    float original_level;
    float delta;
    gboolean in;
} AnimationData;

static void animate_zoom_new_frame_cb(ClutterTimeline* timeline, gint msecs, AnimationData* data)
{
    gdouble progress = clutter_timeline_get_progress(timeline);
    gfloat new_delta = data->delta * progress;

    gfloat new_level;
    if (data->in)
        new_level = data->original_level + new_delta;
    else
        new_level = data->original_level - new_delta;

    webkit_web_frame_set_zoom_level(data->frame, new_level);
}

static void animation_completed_cb(ClutterTimeline* timeline, AnimationData* data)
{
    g_slice_free(AnimationData, data);
}

static void animate_zoom(WebKitWebFrame* frame, gboolean in)
{
    ClutterTimeline* timeline = clutter_timeline_new(300);
    AnimationData* data = g_slice_new(AnimationData);

    data->frame = frame;
    data->original_level = webkit_web_frame_get_zoom_level(frame);

    /* We'll go for 10% zoom - this should take into account the
     * configuration in WebKitWebSettings.
     */
    data->delta = 0.1;
    data->in = in;

    g_signal_connect(timeline, "new-frame",
                     G_CALLBACK(animate_zoom_new_frame_cb),
                     data);

    g_signal_connect(timeline, "completed",
                     G_CALLBACK(animation_completed_cb),
                     data);

    clutter_timeline_start(timeline);
}

static void animate_zoom_in(WebKitWebFrame* frame)
{
    animate_zoom(frame, TRUE);
}

static void animate_zoom_out(WebKitWebFrame* frame)
{
    animate_zoom(frame, FALSE);
}

static gboolean scroll_event_cb(ClutterActor* actor, ClutterEvent* event, WebKitWebPage* page)
{
    WebKitWebFrame* frame = webkit_web_page_get_focused_frame(page);
    if (!frame)
        return FALSE;

    ClutterScrollEvent* scroll = (ClutterScrollEvent*)event;

    if (scroll->direction != CLUTTER_SCROLL_DOWN && scroll->direction != CLUTTER_SCROLL_UP)
        return FALSE;

    if (scroll->modifier_state & CLUTTER_CONTROL_MASK) {
        if (scroll->direction == CLUTTER_SCROLL_UP)
            animate_zoom_in(frame);
        else if (scroll->direction == CLUTTER_SCROLL_DOWN)
            animate_zoom_out(frame);

        return TRUE;
    }

    if (scroll->modifier_state & CLUTTER_SHIFT_MASK) {
        if (scroll->direction == CLUTTER_SCROLL_UP)
            webkit_web_frame_set_text_zoom_level(frame, webkit_web_frame_get_text_zoom_level(frame) + 0.05);
        else if (scroll->direction == CLUTTER_SCROLL_DOWN)
            webkit_web_frame_set_text_zoom_level(frame, webkit_web_frame_get_text_zoom_level(frame) - 0.05);

        return TRUE;
    }

    return FALSE;
}

static void input_method_state_set_cb(WebKitWebPage* page, WebKitDOMNode* node, gboolean enabled)
{
    printf("Input method state set to %s\n", enabled?"enabled":"disabled");
}

#define WINDOW_WIDTH 360.0
#define WINDOW_HEIGHT 600.0

static gboolean buttonReleaseOnTexture(ClutterActor* texture, ClutterEvent* event, WebKitDOMHTMLVideoElement* element)
{
#ifdef ENABLE_VIDEO
    webkit_dom_html_video_element_webkit_exit_fullscreen(element);
#endif
    return TRUE;
}

gboolean media_exit_fullscreen_cb(WebKitWebActor* webActor, ClutterActor* texture)
{
    return TRUE;
}

gboolean media_fullscreen_cb(WebKitWebActor* webActor, WebKitDOMHTMLElement* element, ClutterActor* texture)
{
    clutter_actor_set_reactive(texture, TRUE);

    g_signal_connect(texture, "button-release-event",
                     G_CALLBACK(buttonReleaseOnTexture), element);

    clutter_container_add_actor(CLUTTER_CONTAINER(stage), texture);
    clutter_actor_set_position(texture, 0, 0);
    clutter_actor_set_size(texture, 10, 10);

    clutter_actor_animate(texture, CLUTTER_LINEAR, 2000,
                          "width", 360.0,
                          "height", 360.0,
                          NULL);

    return TRUE;
}

int main(int argc, char** argv)
{
    if (clutter_init(&argc, &argv) != CLUTTER_INIT_SUCCESS) {
        g_error ("Failed to initialized clutter.");
        return 1;
    }

    webkit_init();

    stage = clutter_stage_get_default();
    clutter_actor_set_size(stage, WINDOW_WIDTH, WINDOW_HEIGHT);

    ClutterColor colorBlack;
    clutter_color_from_string(&colorBlack, "black");

    clutter_stage_set_color(CLUTTER_STAGE(stage), &colorBlack);

    // The main vertical box layout.
    ClutterLayoutManager* vboxManager = clutter_box_layout_new();
    clutter_box_layout_set_vertical(CLUTTER_BOX_LAYOUT(vboxManager), TRUE);

    ClutterActor* vbox = clutter_box_new(vboxManager);
    clutter_actor_set_size(CLUTTER_ACTOR(vbox), WINDOW_WIDTH, WINDOW_HEIGHT);
    clutter_container_add_actor(CLUTTER_CONTAINER(stage), vbox);

    // The toolbar.
    ClutterLayoutManager* hboxManager = clutter_box_layout_new();
    clutter_box_layout_set_vertical(CLUTTER_BOX_LAYOUT(hboxManager), FALSE);
    clutter_box_layout_set_spacing(CLUTTER_BOX_LAYOUT(hboxManager), 6);

    ClutterActor* hbox = clutter_box_new(hboxManager);
    clutter_container_add_actor(CLUTTER_CONTAINER(vbox), hbox);
    clutter_box_layout_set_expand(CLUTTER_BOX_LAYOUT(vboxManager), hbox, FALSE);
    clutter_box_layout_set_fill(CLUTTER_BOX_LAYOUT(vboxManager), hbox, TRUE, FALSE);

    // The view.
    ClutterActor* webActor = g_object_new(WEBKIT_TYPE_WEB_ACTOR, 0);
    clutter_container_add_actor(CLUTTER_CONTAINER(vbox), webActor);
    clutter_box_layout_set_expand(CLUTTER_BOX_LAYOUT(vboxManager), webActor, TRUE);
    clutter_box_layout_set_fill(CLUTTER_BOX_LAYOUT(vboxManager), webActor, TRUE, TRUE);

    g_signal_connect(webActor, "button-press-event",
                     G_CALLBACK(button_press_event_cb), NULL);

    g_signal_connect(webActor, "button-release-event",
                     G_CALLBACK(button_release_event_cb), NULL);

    g_signal_connect(webActor, "media-fullscreen",
                     G_CALLBACK(media_fullscreen_cb), NULL);

    g_signal_connect(webActor, "media-exit-fullscreen",
                     G_CALLBACK(media_exit_fullscreen_cb), NULL);

    WebKitWebPage* webPage = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(webActor));

    g_signal_connect(webActor, "scroll-event",
                     G_CALLBACK(scroll_event_cb), webPage);

    g_signal_connect(webPage, "create-window",
                     G_CALLBACK(createWindowCallback),
                     NULL);

    g_signal_connect(webPage, "input-method-state-set",
                     G_CALLBACK(input_method_state_set_cb),
                     NULL);

    webkit_web_page_set_zoom_level(webPage, 0.8);

    WebKitWebSettings* settings = webkit_web_page_get_settings(WEBKIT_WEB_PAGE(webPage));
    g_object_set(settings,
                 "javascript-can-open-windows-automatically", TRUE,
                 "auto-load-images", FALSE,
                 NULL);

    // Now the buttons.
    ClutterColor colorWhite;
    clutter_color_from_string(&colorWhite, "white");

    // The back "button".
    ClutterActor* backText = clutter_text_new_full("sans", "<", &colorWhite);
    g_object_set(backText,
                 "reactive", TRUE,
                 NULL);
    g_signal_connect(backText, "button-release-event",
                     G_CALLBACK(backCallback), webActor);
    clutter_container_add_actor(CLUTTER_CONTAINER(hbox), backText);
    clutter_box_layout_set_expand(CLUTTER_BOX_LAYOUT(hboxManager), backText, FALSE);
    clutter_box_layout_set_fill(CLUTTER_BOX_LAYOUT(hboxManager), backText, FALSE, FALSE);

    // The location bar.
    ClutterActor* locationBar = clutter_text_new_full("sans", "", &colorWhite);
    g_object_set(locationBar,
                 "editable", TRUE,
                 "activatable", TRUE,
                 "reactive", TRUE,
                 NULL);
    g_signal_connect(locationBar, "activate",
                     G_CALLBACK(locationChangeCallback), webActor);
    clutter_container_add_actor(CLUTTER_CONTAINER(hbox), locationBar);
    clutter_box_layout_set_expand(CLUTTER_BOX_LAYOUT(hboxManager), locationBar, TRUE);
    clutter_box_layout_set_fill(CLUTTER_BOX_LAYOUT(hboxManager), locationBar, TRUE, FALSE);

    g_object_connect(webPage,
                     "signal::notify::load-status", G_CALLBACK(load_status_changed_cb), NULL,
                     "signal::notify::title", G_CALLBACK(title_changed_cb), stage,
                     "signal::notify::uri", G_CALLBACK(uri_changed_cb), locationBar,
                     "signal::notify::progress", G_CALLBACK(progress_changed_cb), stage,
                     NULL);

    if (argc > 1)
        webkit_iweb_view_load_uri(WEBKIT_IWEB_VIEW(webActor), argv[1]);
    else
        webkit_iweb_view_load_uri(WEBKIT_IWEB_VIEW(webActor), "http://google.com/");

    clutter_actor_show_all(stage);

    clutter_main();

    return 0;
}
