/*
 * Copyright (C) 2010 Collabora Ltd.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "GraphicsLayerStage.h"

#include <mx/mx.h>
#include <webkit/webkitgeolocationpolicydecision.h>
#include <webkit/webkitiwebview.h>
#include <webkit/webkitwebinspector.h>
#include <webkit/webkitwebpolicydecision.h>
#include <webkit/mx/mx-web-view.h>

#define LIBSOUP_USE_UNSTABLE_REQUEST_API 1
#include <libsoup/soup-cache.h>

static guint debug = 0;
MxApplication* app;

typedef enum
{
  DEBUG_EVENTS = 1 << 0
} LauncherDebugFlags;

static const GDebugKey debug_keys[] = {
    { "events", DEBUG_EVENTS },
    { NULL, 0 }
};

#define DEBUG(flag, format, ...) \
    G_STMT_START { \
        if (flag & debug) \
            g_debug ("%s: " format, G_STRFUNC, ## __VA_ARGS__); \
    } G_STMT_END

extern void webkit_init();

static ClutterActor* createView(MxWindow*);

static gboolean optionSilent = FALSE;
static gboolean optionSmallScreenMode = FALSE;
static gboolean optionDisablePlugins = FALSE;
static gboolean optionDisableJavaScript = FALSE;
static gboolean optionDisableTiledBackingStore = FALSE;
static gboolean optionDebuggingBorders = FALSE;
static gchar*   optionDragMechanism = (gchar*)"native";
static gchar*   optionGeometry = NULL;
static gboolean optionLayerStage = FALSE;
static gboolean optionDisableAcceleratedCanvas = FALSE;
static gchar**  optionURIs = NULL;
static gboolean optionTimedQuit = FALSE;
static gboolean optionDisablePageCache = FALSE;
static gboolean optionDisableMediaPreload = FALSE;
static const GOptionEntry optionEntries[] =
{
    { "geometry", 'g', 0, G_OPTION_ARG_STRING, &optionGeometry,
      "Set the window size; for example: 1024x768 (WIDTHxHEIGHT)", "GEOMETRY" },
    { "silent", 's', 0, G_OPTION_ARG_NONE, &optionSilent,
      "Do not print JS console messages.", 0 },
    { "custom-stage", 'c', 0, G_OPTION_ARG_STRING, &optionLayerStage,
      "Use the custom GraphicsLayerStage", 0 },
    { "timed-quit", 'q', 0, G_OPTION_ARG_NONE, &optionTimedQuit,
      "Quit 5 seconds after finishing the first page load", 0 },
    { "disable-page-cache", 'M', 0, G_OPTION_ARG_NONE, &optionDisablePageCache,
      "Disable page cache.", 0 },
    { "disable-media-preload", 'P', 0, G_OPTION_ARG_NONE, &optionDisableMediaPreload,
      "Disable media preload.", 0 },
    { "debugging-borders", 'b', 0, G_OPTION_ARG_NONE, &optionDebuggingBorders,
      "Show debugging borders.", 0 },
    { "small-screen", 'S', 0, G_OPTION_ARG_NONE, &optionSmallScreenMode,
      "Use small screen mode. The window will resize to be as big as the screen size.", 0 },
    { "drag-mechanism", 'd', 0, G_OPTION_ARG_STRING, &optionDragMechanism,
      "Use 'scroll' or 'kinetic' to use an Mx wrapper around the web view or 'native' to let WebKit handle everything.", "SCROLLTYPE" },
    { "disable-plugins", 'p', 0, G_OPTION_ARG_NONE, &optionDisablePlugins,
      "Disable NPAPI plugins.", 0 },
    { "disable-javascript", 'j', 0, G_OPTION_ARG_NONE, &optionDisableJavaScript,
      "Disable JavaScript.", 0 },
    { "disable-tiled-backing-store", 't', 0, G_OPTION_ARG_NONE, &optionDisableTiledBackingStore,
      "Disable the Tiled Backing Store.", 0 },
    { "disable-accelerated-canvas", 'a', 0, G_OPTION_ARG_NONE, &optionDisableAcceleratedCanvas,
      "Disable accelerated 2D canvas.", 0 },
    { G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY, &optionURIs,
      0, 0 },
    { 0, 0, 0, 0, 0, 0, 0 }
};

unsigned currentURI = 0;

static gboolean
parse_geometry(char *geometry, int *width, int *height)
{
    g_return_val_if_fail(geometry != NULL, FALSE);
    double w = 0, h = 0;
    gchar** g = g_strsplit(geometry, "x", 2);
    if (g[0] != NULL && g[1] != NULL) {
        w = g_ascii_strtod(g[0], NULL);
        h = g_ascii_strtod(g[1], NULL);
    }
    g_strfreev(g);
    g_return_val_if_fail(w > 0 && h > 0, FALSE);
    *width = (int)w;
    *height = (int)h;
    return TRUE;
}

static const gchar *
get_event_type_name (const ClutterEvent *event)
{
  switch (event->type)
    {
    case CLUTTER_BUTTON_PRESS:
      return "BUTTON PRESS";

    case CLUTTER_BUTTON_RELEASE:
      return "BUTTON_RELEASE";

    case CLUTTER_KEY_PRESS:
      return "KEY PRESS";

    case CLUTTER_KEY_RELEASE:
      return "KEY RELEASE";

    case CLUTTER_ENTER:
      return "ENTER";

    case CLUTTER_LEAVE:
      return "LEAVE";

    case CLUTTER_MOTION:
      return "MOTION";

    case CLUTTER_DELETE:
      return "DELETE";

    default:
      return "EVENT";
    }
}

gboolean
debugCapturedEvent(ClutterActor *actor, ClutterEvent *event, gpointer data)
{
    DEBUG(DEBUG_EVENTS, "evt:%s type:%s",
        get_event_type_name (event),
        G_OBJECT_TYPE_NAME (actor));
    return FALSE;
}

static gboolean
debugButtonPressEvent(ClutterActor* actor, ClutterButtonEvent* event, gpointer data)
{
    DEBUG(DEBUG_EVENTS, "type:%s btn:%d x:%f y:%f",
        G_OBJECT_TYPE_NAME (actor), event->button, event->x, event->y);
    return FALSE;
}

static gboolean newWindowPolicyDecision(WebKitWebPage* page, WebKitWebFrame* frame, WebKitNetworkRequest* request, WebKitWebNavigationAction* action, WebKitWebPolicyDecision* policy, gpointer data)
{
    webkit_web_policy_decision_use(policy);
    return TRUE;
}

static void backCallback(ClutterActor* actor, MxWebView* view)
{
    WebKitWebPage* webPage = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(view));
    webkit_web_page_go_back(webPage);
}

static void locationChangeCallback(ClutterActor* actor, MxWebView* view)
{
    webkit_iweb_view_load_uri(WEBKIT_IWEB_VIEW(view), clutter_text_get_text(CLUTTER_TEXT(actor)));
}

static void zoomInCallback(ClutterActor* actor, MxWebView* view)
{
    WebKitWebPage* webPage = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(view));
    WebKitWebFrame* mainFrame = webkit_web_page_get_main_frame(webPage);

    float zoomFactor = webkit_web_frame_get_contents_scale(mainFrame);
    webkit_web_frame_set_contents_scale(mainFrame, zoomFactor + 0.2);
}

static void zoomOutCallback(ClutterActor* actor, MxWebView* view)
{
    WebKitWebPage* webPage = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(view));
    WebKitWebFrame* mainFrame = webkit_web_page_get_main_frame(webPage);

    float zoomFactor = webkit_web_frame_get_contents_scale(mainFrame);
    webkit_web_frame_set_contents_scale(mainFrame, zoomFactor - 0.2);
}

static gboolean blockButtonPressEvent(ClutterActor* actor, ClutterButtonEvent* event, gpointer data)
{
    // Prevent clicks to propagate to the stage as they're used to scroll and
    // should not reach the default MxWindow handler (which otherwise moves
    // the window if it has a toolbar)
    return event->button == 1;
}

static WebKitWebPage* inspectionRequestedCallback(WebKitWebPage* page, WebKitWebInspector* inspector, MxWindow* window)
{
    MxWindow* inspectorWindow = mx_window_new();
    ClutterActor* scrollView = mx_scroll_view_new();

    mx_window_set_child(inspectorWindow, scrollView);

    ClutterActor* webView = mx_web_view_new();
    clutter_container_add_actor(CLUTTER_CONTAINER(scrollView), webView);

    g_signal_connect_after(webView, "button-press-event",
                     G_CALLBACK(blockButtonPressEvent), 0);

    mx_window_set_window_size(inspectorWindow, 1000, 700);
    mx_window_show(inspectorWindow);

    return webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(webView));
}

static gboolean buttonPressEvent(ClutterActor* actor, ClutterButtonEvent* event, gpointer data)
{
    /* NOTE: this is just a proof of concept, it is not intended for
     * use in an actual browser..
     */
    DEBUG(DEBUG_EVENTS, "btn:%d x:%f y:%f", event->button, event->x, event->y);
   if (event->button == 3) {
        DEBUG(DEBUG_EVENTS, "launching ispector");
        WebKitWebPage* page = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(actor));
        WebKitWebInspector* inspector = webkit_web_page_get_inspector(page);
        webkit_web_inspector_inspect_coordinates(inspector, event->x, event->y);
        return TRUE;
    }

    return FALSE;
}

static void transitionsCompleted(ClutterActor* actor, gpointer data)
{
    WebKitWebPage* webPage = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(actor));
    WebKitWebFrame* mainFrame = webkit_web_page_get_main_frame(webPage);
    double zoomFactor;

    g_object_get(actor, "fast-zoom-level", &zoomFactor, NULL);

    zoomFactor *= webkit_web_frame_get_contents_scale(mainFrame);

    webkit_web_frame_set_contents_frozen(mainFrame, false);
    webkit_web_frame_set_contents_scale(mainFrame, zoomFactor);

    mx_web_view_set_fast_zoom_level(MX_WEB_VIEW(actor), 1.);
}

static WebKitDOMElement* chooseElementToZoomCb(GSList* candidates, void* userData)
{
    GSList* iter;

    if (candidates == NULL)
        return NULL;

    for (iter = candidates; iter; iter = g_slist_next(iter)) {
        WebKitDOMElement* element = WEBKIT_DOM_ELEMENT(iter->data);

        fprintf(stderr, "tag %s\n", webkit_dom_element_get_tag_name(element));
    }

    return WEBKIT_DOM_ELEMENT(candidates->data);
}

static gboolean buttonReleaseEvent(ClutterActor* actor, ClutterButtonEvent* event, gpointer data)
{
    WebKitWebPage* page = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(actor));
    WebKitWebFrame* mainFrame = webkit_web_page_get_main_frame(page);
    float x, y;
    int xZoomPosition, yZoomPosition;
    double zoomFactor;

    if (clutter_event_get_click_count ((ClutterEvent*)event) < 2)
        return FALSE;

    clutter_actor_transform_stage_point(actor, event->x, event->y, &x, &y);

    if (!webkit_web_page_get_zoom_target(page, x, y, &xZoomPosition, &yZoomPosition,
                                         &zoomFactor, chooseElementToZoomCb, actor))
        return FALSE;

    fprintf (stderr, "buttonReleaseEvent zoom %f scroll %d,%d\n",
             zoomFactor, xZoomPosition, yZoomPosition);

    webkit_web_frame_set_contents_frozen(mainFrame, true);

    // Take into account the existing scaling, as the fast zoom level is a
    // transformation that gets applied on top
    zoomFactor /= webkit_web_frame_get_contents_scale(mainFrame);

    clutter_actor_save_easing_state(actor);
    clutter_actor_set_easing_duration(actor, 2000);
    mx_web_view_set_x_scroll_offset(MX_WEB_VIEW(actor), xZoomPosition);
    mx_web_view_set_y_scroll_offset(MX_WEB_VIEW(actor), yZoomPosition);
    mx_web_view_set_fast_zoom_level(MX_WEB_VIEW(actor), zoomFactor);
    clutter_actor_restore_easing_state(actor);

    return TRUE;
}

static gboolean scrollEvent(ClutterActor* actor, ClutterEvent* event, WebKitWebPage* page)
{
    WebKitWebFrame* frame = webkit_web_page_get_focused_frame(page);
    if (!frame)
        return FALSE;

    ClutterScrollEvent* scroll = (ClutterScrollEvent*)event;

    DEBUG(DEBUG_EVENTS, "dir:%s modifier_state:%s",
        scroll->direction == CLUTTER_SCROLL_DOWN ? "down" :
        scroll->direction == CLUTTER_SCROLL_UP ? "up" :
        "other",
        scroll->modifier_state == 0 ? "none" :
        scroll->modifier_state & CLUTTER_CONTROL_MASK ? "ctrl" :
        "other"
        );

    if (scroll->direction != CLUTTER_SCROLL_DOWN && scroll->direction != CLUTTER_SCROLL_UP)
        return FALSE;

    if (!(scroll->modifier_state & CLUTTER_CONTROL_MASK))
        return FALSE;

    float zoomFactor = webkit_web_frame_get_contents_scale(frame);

    DEBUG(DEBUG_EVENTS, "zooming");
    if (scroll->direction == CLUTTER_SCROLL_UP)
        zoomFactor += 0.1;
    else if (scroll->direction == CLUTTER_SCROLL_DOWN)
        zoomFactor -= 0.1;

    webkit_web_frame_set_contents_scale(frame, zoomFactor);

    return TRUE;
}

static gboolean alertCallback(WebKitWebPage* webPage, WebKitWebFrame* frame, const char* message, gpointer data)
{
    g_message("ALERT: %s", message);
    return TRUE;
}

static gboolean consoleMessageCallback(WebKitWebPage* webPage, WebKitWebFrame* frame, const char* message, gpointer data)
{
    return optionSilent;
}

typedef struct {
    WebKitGeolocationPolicyDecision* decision;
    ClutterActor* dialog;
} GeoDecisionData;

static void
applyGeolocationPolicyDecision(MxAction* action, GeoDecisionData* data)
{
    WebKitGeolocationPolicyDecision* decision = data->decision;

    if (g_str_equal(mx_action_get_name(action), "allow"))
        webkit_geolocation_policy_allow(decision);
    else
        webkit_geolocation_policy_deny(decision);

    clutter_actor_destroy(data->dialog);
    g_object_unref(data->decision);
    g_slice_free(GeoDecisionData, data);
}

static gboolean
geolocation_policy_decision_cb(WebKitWebPage* webPage,
                               WebKitWebFrame* webFrame,
                               WebKitGeolocationPolicyDecision* decision,
                               gpointer user_data)
{
    ClutterActor* view = CLUTTER_ACTOR(webkit_web_page_get_view(webPage));

    ClutterActor* dialog = mx_dialog_new();
    mx_dialog_set_transient_parent(MX_DIALOG(dialog), clutter_actor_get_stage(view));

    char* text = g_strdup_printf("The site %s wants to know your location",
                                 webkit_web_page_get_uri(webPage));
    ClutterActor* label = mx_label_new_with_text(text);
    g_free(text);
    clutter_actor_add_child(dialog, label);

    GeoDecisionData* data = g_slice_new(GeoDecisionData);
    data->decision = g_object_ref(decision);
    data->dialog = dialog;

    MxAction* deny = mx_action_new_full("deny", "Deny",
                                         G_CALLBACK(applyGeolocationPolicyDecision),
                                         data);

    mx_dialog_add_action(MX_DIALOG(dialog), deny);

    MxAction* allow = mx_action_new_full("allow", "Allow",
                                         G_CALLBACK(applyGeolocationPolicyDecision),
                                         data);

    mx_dialog_add_action(MX_DIALOG(dialog), allow);

    clutter_actor_show_all(dialog);

    return TRUE;
}

static void
uriChanged(WebKitWebPage* page, GParamSpec* pspec, ClutterText* text)
{
    clutter_text_set_text(text, webkit_web_page_get_uri(page));
}

static WebKitWebPage* createWindow(ClutterStage* stage, const char* uri)
{
    if (!stage)
        stage = CLUTTER_STAGE(clutter_stage_new());

    MxWindow* window = mx_window_new_with_clutter_stage(CLUTTER_STAGE(stage));
    mx_window_set_title(window, "Clutter/WebKit");
    mx_application_add_window(MX_APPLICATION(app), window);

    if (optionSmallScreenMode)
        mx_window_set_small_screen(window, TRUE);

    mx_window_set_has_toolbar(window, TRUE);

    int width = 1024, height = 768;
    if (optionGeometry != NULL)
        parse_geometry(optionGeometry, &width, &height);
    mx_window_set_window_size(window, width, height);

    clutter_stage_set_color(CLUTTER_STAGE(stage), CLUTTER_COLOR_Black);

    // The toolbar.
    MxToolbar* toolbar = mx_window_get_toolbar(window);

    ClutterActor* toolbar_hbox = mx_box_layout_new();
    mx_box_layout_set_spacing(MX_BOX_LAYOUT(toolbar_hbox), 6);
    clutter_container_add_actor(CLUTTER_CONTAINER(toolbar), toolbar_hbox);

    // The view.
    ClutterActor* view = createView(window);
    clutter_actor_set_name (view, "MxWebView main");

    // Now the toolbar dwellers.
    ClutterColor colorWhite;
    clutter_color_from_string(&colorWhite, "white");

    // The + button.
    ClutterActor* zoomIn = mx_button_new_with_label("+");
    g_signal_connect(zoomIn, "clicked",
                     G_CALLBACK(zoomInCallback), view);
    mx_box_layout_insert_actor_with_properties(MX_BOX_LAYOUT(toolbar_hbox), zoomIn, 0,
        "y-fill", TRUE,
        NULL);


    // The - button.
    ClutterActor* zoomOut = mx_button_new_with_label("-");
    g_signal_connect(zoomOut, "clicked",
                     G_CALLBACK(zoomOutCallback), view);
    mx_box_layout_insert_actor_with_properties(MX_BOX_LAYOUT(toolbar_hbox), zoomOut, 0,
        "y-fill", TRUE,
        NULL);

    // The location bar.
    ClutterActor* locationBar = mx_entry_new();
    ClutterActor* locationTextActor = mx_entry_get_clutter_text(MX_ENTRY(locationBar));

    g_signal_connect(locationTextActor, "activate",
                     G_CALLBACK(locationChangeCallback), view);

    WebKitWebPage* page = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(view));

    g_signal_connect(page, "notify::uri",
                     G_CALLBACK(uriChanged), locationTextActor);

    mx_box_layout_insert_actor_with_properties(MX_BOX_LAYOUT(toolbar_hbox), locationBar, 0,
        "expand", TRUE,
        "x-fill", TRUE,
        "y-fill", TRUE,
        NULL);

    // The back button.
    ClutterActor* backText = mx_button_new_with_label("Back");
    g_signal_connect(backText, "clicked",
                     G_CALLBACK(backCallback), view);
    mx_box_layout_insert_actor_with_properties(MX_BOX_LAYOUT(toolbar_hbox), backText, 0,
        "y-fill", TRUE,
        NULL);

    mx_window_show(window);

    g_signal_connect(stage, "captured-event",
                     G_CALLBACK(debugCapturedEvent), 0);
    g_signal_connect(stage, "button-press-event",
                     G_CALLBACK(debugButtonPressEvent), 0);
    g_signal_connect(view, "captured-event",
                     G_CALLBACK(debugCapturedEvent), 0);
    g_signal_connect(view, "button-press-event",
                     G_CALLBACK(debugButtonPressEvent), 0);

    if (uri)
        webkit_iweb_view_load_uri(WEBKIT_IWEB_VIEW(view), uri);

    return page;
}

static void
startup(GApplication *app, const char *startup_uri)
{

    GOptionContext* optionContext = g_option_context_new("");

    g_option_context_set_description(optionContext,
        "Set the environment variable LAUNCHER_DEBUG to one or more of the "
        "following flags to enable debug messages, or just set it to 'all':\n"
        "  events - print event triggering and propagation");

    GOptionGroup* optionGroup = g_option_group_new("mxlauncher",
                                                   "Mx-based webkit-clutter test browser",
                                                   "Mx-based webkit-clutter test browser options",
                                                   0, 0);

    g_option_group_add_entries(optionGroup, optionEntries);

    g_option_context_set_main_group(optionContext, optionGroup);

    g_option_context_set_ignore_unknown_options(optionContext, TRUE);

    debug = g_parse_debug_string (g_getenv ("LAUNCHER_DEBUG"), debug_keys,
        G_N_ELEMENTS (debug_keys) - 1);

    webkit_init();

    ClutterActor* stage;
    if (optionLayerStage)
        stage = CLUTTER_ACTOR(graphics_layer_stage_new());
    else {
        stage = clutter_stage_new();
        clutter_stage_set_user_resizable(CLUTTER_STAGE(stage), TRUE);
    }

    clutter_actor_set_name(stage, "MxLauncher stage");

    char* cacheDir = g_build_filename(g_get_user_cache_dir(), "webkit-clutter", NULL);
    SoupCache* cache = soup_cache_new(cacheDir, SOUP_CACHE_SHARED);
    soup_session_add_feature(webkit_get_default_session(), SOUP_SESSION_FEATURE(cache));
    soup_cache_load(cache);
    g_object_unref(cache);

    /* Cookies */
    char* filename = g_build_filename(cacheDir, "cookies.txt", NULL);
    SoupCookieJar* jar = soup_cookie_jar_text_new(filename, FALSE);
    soup_session_add_feature(webkit_get_default_session(), SOUP_SESSION_FEATURE(jar));
    g_object_unref(jar);
    g_free(cacheDir);
    g_free(filename);

    const char* uri = "http://google.com/";
    if (optionURIs)
        uri = optionURIs[0];

    createWindow(CLUTTER_STAGE(stage), uri);
}

static WebKitWebPage* createWindowCallback(WebKitWebActor* actor, WebKitWebFrame* frame)
{
    return createWindow(NULL, NULL);
}

static void webLoadFinished(WebKitWebPage* page, WebKitWebFrame* frame, gpointer data)
{
    if (!optionURIs[++currentURI]) {
        g_timeout_add_full(G_PRIORITY_LOW, 5000, (GSourceFunc)g_application_quit, app, 0);
        return;
    }

    webkit_iweb_view_load_uri(webkit_web_page_get_view(page), optionURIs[currentURI]);
}

static ClutterActor*
createView(MxWindow *window)
{
    ClutterActor* view = mx_web_view_new();
    ClutterActor* scrollView = NULL;
    ClutterActor* hbox;

    hbox = mx_box_layout_new();
    mx_window_set_child(window, hbox);
    clutter_actor_set_x_expand(view, TRUE);
    clutter_actor_set_y_expand(view, TRUE);

    if (g_strcmp0("scroll", optionDragMechanism) == 0) {
        scrollView = mx_scroll_view_new();
    }
    else if (g_strcmp0("kinetic", optionDragMechanism) == 0) {
        MxAdjustment* hadjustment;
        MxAdjustment* vadjustment;
        mx_scrollable_get_adjustments(MX_SCROLLABLE(view), &hadjustment, &vadjustment);
        mx_adjustment_set_elastic(hadjustment, TRUE);
        mx_adjustment_set_elastic(vadjustment, TRUE);

        scrollView = mx_kinetic_scroll_view_new();
        mx_kinetic_scroll_view_set_use_captured(MX_KINETIC_SCROLL_VIEW(scrollView), TRUE);
    }

    if (scrollView != NULL) {
        clutter_actor_add_child(scrollView, view);
        mx_box_layout_insert_actor_with_properties(MX_BOX_LAYOUT(hbox), scrollView, 0,
                                                   "expand", TRUE,
                                                   "x-fill", TRUE,
                                                   "y-fill", TRUE,
                                                   NULL);
    }
    else {
        mx_box_layout_insert_actor_with_properties(MX_BOX_LAYOUT(hbox), view, 0,
                                                   "expand", TRUE,
                                                   "x-fill", TRUE,
                                                   "y-fill", TRUE,
                                                   NULL);
    }

    // Enabling a few features of WebKit
    WebKitWebPage* page = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(view));
    g_signal_connect(page, "inspection-requested",
                     G_CALLBACK(inspectionRequestedCallback), window);

    g_signal_connect(page, "geolocation-policy-decision-requested",
                     G_CALLBACK(geolocation_policy_decision_cb), NULL);

    g_signal_connect_after(view, "button-press-event",
                     G_CALLBACK(blockButtonPressEvent), 0);

    g_signal_connect(view, "button-press-event",
                     G_CALLBACK(buttonPressEvent), 0);

    g_signal_connect(view, "button-release-event",
                     G_CALLBACK(buttonReleaseEvent), 0);

    g_signal_connect(view, "scroll-event",
                     G_CALLBACK(scrollEvent), page);

    g_signal_connect(view, "transitions-completed",
                     G_CALLBACK(transitionsCompleted), NULL);

    g_signal_connect(page, "new-window-policy-decision-requested",
                     G_CALLBACK(newWindowPolicyDecision), NULL);

    g_signal_connect(page, "create-window",
                     G_CALLBACK(createWindowCallback),
                     NULL);

    g_signal_connect(page, "console-message",
                     G_CALLBACK(consoleMessageCallback), 0);

    g_signal_connect(page, "script-alert",
                     G_CALLBACK(alertCallback), window);

    if (optionTimedQuit)
        g_signal_connect(page, "web-load-finished",
                         G_CALLBACK(webLoadFinished), view);

    WebKitWebSettings* settings = webkit_web_page_get_settings(page);
    g_object_set(settings, "enable-developer-extras", TRUE, NULL);
    g_object_set(settings, "enable-webgl", TRUE, NULL);
    g_object_set(settings, "enable-accelerated-2d-canvas", !optionDisableAcceleratedCanvas, NULL);
    g_object_set(settings, "enable-universal-access-from-file-uris", TRUE, NULL);

    if (optionDisablePageCache)
        g_object_set(settings, "enable-page-cache", FALSE, NULL);

    if (!optionDisableTiledBackingStore)
        g_object_set(settings, "enable-tiled-backing-store", TRUE, NULL);

    if (optionDisableJavaScript)
        g_object_set(settings, "enable-scripts", FALSE, NULL);

    if (optionDisablePlugins)
        g_object_set(settings, "enable-plugins", FALSE, NULL);

    if (optionDisableMediaPreload)
        g_object_set(settings, "enable-media-preloading", FALSE, NULL);

    if (optionDebuggingBorders)
        g_object_set(settings, "enable-debugging-borders", TRUE, NULL);

    return view;
}


int main(int argc, char** argv)
{
    g_type_init();

    GOptionContext* optionContext = g_option_context_new("");

    g_option_context_set_description(optionContext,
        "Set the environment variable LAUNCHER_DEBUG to one or more of the "
        "following flags to enable debug messages, or just set it to 'all':\n"
        "  events - print event triggering and propagation");

    GOptionGroup* optionGroup = g_option_group_new("mxlauncher",
                                                   "Mx-based webkit-clutter test browser",
                                                   "Mx-based webkit-clutter test browser options",
                                                   0, 0);

    g_option_group_add_entries(optionGroup, optionEntries);

    g_option_context_set_main_group(optionContext, optionGroup);

    GError* error = 0;
    if (!g_option_context_parse(optionContext, &argc, &argv, &error)) {
        g_error("Failed to parse arguments: %s", error->message);
        return 1;
    }

    g_option_context_free(optionContext);

    clutter_x11_enable_xinput();

    if (clutter_init(&argc, &argv) != CLUTTER_INIT_SUCCESS) {
        g_error ("Failed to initialized clutter.");
        return 1;
    }

    debug = g_parse_debug_string (g_getenv ("LAUNCHER_DEBUG"), debug_keys,
        G_N_ELEMENTS (debug_keys) - 1);

    app = mx_application_new("com.collabora.WebKit.MxLauncher", 0);
    g_signal_connect(app, "startup", G_CALLBACK(startup), 0);
    g_application_run(G_APPLICATION(app), argc, argv);

    return 0;
}
