/*
 * Copyright (C) 2007 Eric Seidel <eric@webkit.org>
 * Copyright (C) 2008 Alp Toker <alp@nuanti.com>
 * Copyright (C) 2009 Jan Alonzo <jmalonzo@gmail.com>
 * Copyright (C) 2010 Igalia S.L.
 * Copyright (C) 2010 Collabora Ltd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "DumpRenderTree.h"

#include "EventSender.h"
#include "GCController.h"
#include "PlainTextController.h"
#include "TestRunner.h"
#include "WebCoreSupport/DumpRenderTreeSupportClutter.h"
#include "WebCoreTestSupport.h"
#include "WorkQueue.h"
#include "WorkQueueItem.h"
#include "webkit/vanilla/webkitwebactor.h"
#include "webkit/webkitclutter.h"
#include "webkit/webkitglobals.h"
#include "webkit/webkitiwebview.h"
#include "webkit/webkitwebbackforwardlist.h"
#include "webkit/webkitwebdatabase.h"
#include "webkit/webkitgeolocationpolicydecision.h"
#include "webkit/webkitgwebframe.h"
#include "webkit/webkitgwebframeprivate.h"
#include "webkit/webkitwebhistoryitemprivate.h"
#include "webkit/webkitwebpage.h"
#include <cstring>
#include <wtf/text/CString.h>

#include <fontconfig/fontconfig.h>

using namespace std;

static ClutterActor* stage = 0;
static WebKitWebActor* webActor = 0;
static WebKitWebPage* webPage = 0;

WebKitWebFrame* mainFrame = 0;
WebKitWebFrame* topLoadingFrame = 0;

// This is a list of opened webviews
GSList* webPageList = 0;

// current b/f item at the end of the previous test
static WebKitWebHistoryItem* prevTestBFItem = NULL;

const unsigned historyItemIndent = 8;

static GCController* gcController = 0;
static PlainTextController* plainTextController = 0;
RefPtr<TestRunner> gTestRunner;

volatile bool done;
static bool printSeparators;
static int dumpPixels;
static int dumpTree = 1;

guint waitToDumpWatchdog = 0;
bool waitForPolicy = false;

static bool shouldLogFrameLoadDelegates(const string& pathOrURL)
{
    return pathOrURL.find("loading/") != string::npos;
}

static void initializeFonts(const char* testURL = 0)
{
    // initializeGtkFontSettings(testURL);

    FcInit();

    // If a test resulted a font being added or removed via the @font-face rule, then
    // we want to reset the FontConfig configuration to prevent it from affecting other tests.
    static int numFonts = 0;
    FcFontSet* appFontSet = FcConfigGetFonts(0, FcSetApplication);
    if (appFontSet && numFonts && appFontSet->nfont == numFonts)
        return;

    // Load our configuration file, which sets up proper aliases for family
    // names like sans, serif and monospace.
    FcConfig* config = FcConfigCreate();
    GOwnPtr<gchar> fontConfigFilename(g_build_filename(FONTS_CONF_DIR, "fonts.conf", NULL));
    if (!g_file_test(fontConfigFilename.get(), G_FILE_TEST_EXISTS)) {
        fontConfigFilename.set(g_build_filename("/usr/share/webkit_clutter-0/testing/", "fonts.conf", NULL));
    }
    if (!FcConfigParseAndLoad(config, reinterpret_cast<FcChar8*>(fontConfigFilename.get()), true))
        g_error("Couldn't load font configuration file from: %s", fontConfigFilename.get());

    static const char *const fontPaths[][3] = {
        { "/usr/share/fonts/truetype/ttf-liberation/LiberationMono-BoldItalic.ttf",
          "/usr/share/fonts/truetype/liberation/LiberationMono-BoldItalic.ttf",
          "/usr/share/fonts/liberation/LiberationMono-BoldItalic.ttf", },
        { "/usr/share/fonts/truetype/ttf-liberation/LiberationMono-Bold.ttf",
          "/usr/share/fonts/truetype/liberation/LiberationMono-Bold.ttf",
          "/usr/share/fonts/liberation/LiberationMono-Bold.ttf", },
        { "/usr/share/fonts/truetype/ttf-liberation/LiberationMono-Italic.ttf",
          "/usr/share/fonts/truetype/liberation/LiberationMono-Italic.ttf",
          "/usr/share/fonts/liberation/LiberationMono-Italic.ttf", },
        { "/usr/share/fonts/truetype/ttf-liberation/LiberationMono-Regular.ttf",
          "/usr/share/fonts/truetype/liberation/LiberationMono-Regular.ttf",
          "/usr/share/fonts/liberation/LiberationMono-Regular.ttf", },
        { "/usr/share/fonts/truetype/ttf-liberation/LiberationSans-BoldItalic.ttf",
          "/usr/share/fonts/truetype/liberation/LiberationSans-BoldItalic.ttf",
          "/usr/share/fonts/liberation/LiberationSans-BoldItalic.ttf", },
        { "/usr/share/fonts/truetype/ttf-liberation/LiberationSans-Bold.ttf",
          "/usr/share/fonts/truetype/liberation/LiberationSans-Bold.ttf",
          "/usr/share/fonts/liberation/LiberationSans-Bold.ttf", },
        { "/usr/share/fonts/truetype/ttf-liberation/LiberationSans-Italic.ttf",
          "/usr/share/fonts/truetype/liberation/LiberationSans-Italic.ttf",
          "/usr/share/fonts/liberation/LiberationSans-Italic.ttf", },
        { "/usr/share/fonts/truetype/ttf-liberation/LiberationSans-Regular.ttf",
          "/usr/share/fonts/truetype/liberation/LiberationSans-Regular.ttf",
          "/usr/share/fonts/liberation/LiberationSans-Regular.ttf", },
        { "/usr/share/fonts/truetype/ttf-liberation/LiberationSerif-BoldItalic.ttf",
          "/usr/share/fonts/truetype/liberation/LiberationSerif-BoldItalic.ttf",
          "/usr/share/fonts/liberation/LiberationSerif-BoldItalic.ttf", },
        { "/usr/share/fonts/truetype/ttf-liberation/LiberationSerif-Bold.ttf",
          "/usr/share/fonts/truetype/liberation/LiberationSerif-Bold.ttf",
          "/usr/share/fonts/liberation/LiberationSerif-Bold.ttf", },
        { "/usr/share/fonts/truetype/ttf-liberation/LiberationSerif-Italic.ttf",
          "/usr/share/fonts/truetype/liberation/LiberationSerif-Italic.ttf",
          "/usr/share/fonts/liberation/LiberationSerif-Italic.ttf", },
        { "/usr/share/fonts/truetype/ttf-liberation/LiberationSerif-Regular.ttf",
          "/usr/share/fonts/truetype/liberation/LiberationSerif-Regular.ttf",
          "/usr/share/fonts/liberation/LiberationSerif-Regular.ttf", },
        { "/usr/share/fonts/truetype/ttf-dejavu/DejaVuSans.ttf",
          "/usr/share/fonts/dejavu/DejaVuSans.ttf",
          0, },
        { "/usr/share/fonts/truetype/ttf-dejavu/DejaVuSerif.ttf",
          "/usr/share/fonts/dejavu/DejaVuSerif.ttf",
          0, },

        // MathML tests require the STIX fonts.
        //{ "/usr/share/fonts/opentype/stix/STIXGeneral.otf",
        //  "/usr/share/fonts/stix/STIXGeneral.otf",
        //  0, },
        //{ "/usr/share/fonts/opentype/stix/STIXGeneralBolIta.otf",
        //  "/usr/share/fonts/stix/STIXGeneralBolIta.otf",
        //  0, },
        //{ "/usr/share/fonts/opentype/stix/STIXGeneralBol.otf",
        //  "/usr/share/fonts/stix/STIXGeneralBol.otf",
        //  0, },
        //{ "/usr/share/fonts/opentype/stix/STIXGeneralItalic.otf",
        //  "/usr/share/fonts/stix/STIXGeneralItalic.otf",
        //  0, },
    };

    // TODO: Some tests use Lucida. We should load these as well, once it becomes
    // clear how to install these fonts easily on Fedora.
    for (size_t font = 0; font < G_N_ELEMENTS(fontPaths); font++) {
        bool found = false;
        for (size_t path = 0; path < 3; path++) {
            if (!fontPaths[font][path])
                continue;

            if (g_file_test(fontPaths[font][path], G_FILE_TEST_EXISTS)) {
                found = true;
                if (!FcConfigAppFontAddFile(config, reinterpret_cast<const FcChar8*>(fontPaths[font][path])))
                    g_error("Could not load font at %s!", fontPaths[font][path]);
                else
                    break;
            }
        }

        if (!found)
            g_error("Could not find font at %s. Either install this font or file a bug "
                    "at http://bugs.webkit.org if it is installed in another location.",
                    fontPaths[font][0]);
    }

    // Ahem is used by many layout tests.
    GOwnPtr<gchar> ahemFontFilename(g_build_filename(FONTS_CONF_DIR, "AHEM____.TTF", NULL));
    if (!FcConfigAppFontAddFile(config, reinterpret_cast<FcChar8*>(ahemFontFilename.get())))
        g_error("Could not load font at %s!", ahemFontFilename.get());

    // A font with no valid Fontconfig encoding to test https://bugs.webkit.org/show_bug.cgi?id=47452
    GOwnPtr<gchar> fontWithNoValidEncodingFilename(g_build_filename(FONTS_CONF_DIR, "FontWithNoValidEncoding.fon", NULL));
    if (!FcConfigAppFontAddFile(config, reinterpret_cast<FcChar8*>(fontWithNoValidEncodingFilename.get())))
        g_error("Could not load font at %s!", fontWithNoValidEncodingFilename.get());

    if (!FcConfigSetCurrent(config))
        g_error("Could not set the current font configuration!");

    numFonts = FcConfigGetFonts(config, FcSetApplication)->nfont;
}

static gint compareHistoryItems(gpointer* item1, gpointer* item2)
{
    return g_ascii_strcasecmp(webkit_web_history_item_get_target(WEBKIT_WEB_HISTORY_ITEM(item1)),
                              webkit_web_history_item_get_target(WEBKIT_WEB_HISTORY_ITEM(item2)));
}

static void dumpHistoryItem(WebKitWebHistoryItem* item, int indent, bool current)
{
    ASSERT(item != NULL);
    int start = 0;
    g_object_ref(item);
    if (current) {
        printf("curr->");
        start = 6;
    }
    for (int i = start; i < indent; i++)
        putchar(' ');

    // normalize file URLs.
    const gchar* uri = webkit_web_history_item_get_uri(item);
    gchar* uriScheme = g_uri_parse_scheme(uri);
    if (g_strcmp0(uriScheme, "file") == 0) {
        gchar* pos = g_strstr_len(uri, -1, "/LayoutTests/");
        if (!pos)
            return;

        GString* result = g_string_sized_new(strlen(uri));
        result = g_string_append(result, "(file test):");
        result = g_string_append(result, pos + strlen("/LayoutTests/"));
        printf("%s", result->str);
        g_string_free(result, TRUE);
    } else
        printf("%s", uri);

    g_free(uriScheme);

    const gchar* target = webkit_web_history_item_get_target(item);
    if (target && strlen(target) > 0)
        printf(" (in frame \"%s\")", target);
    if (webkit_web_history_item_is_target_item(item))
        printf("  **nav target**");
    putchar('\n');
    GList* kids = webkit_web_history_item_get_children(item);
    if (kids) {
        // must sort to eliminate arbitrary result ordering which defeats reproducible testing
        kids = g_list_sort(kids, (GCompareFunc) compareHistoryItems);
        for (unsigned i = 0; i < g_list_length(kids); i++)
            dumpHistoryItem(WEBKIT_WEB_HISTORY_ITEM(g_list_nth_data(kids, i)), indent+4, FALSE);
    }
    g_object_unref(item);
}

static void dumpBackForwardListForWebPage(WebKitWebPage* page)
{
    printf("\n============== Back Forward List ==============\n");
    WebKitWebBackForwardList* bfList = webkit_web_page_get_back_forward_list(page);

    // Print out all items in the list after prevTestBFItem, which was from the previous test
    // Gather items from the end of the list, the print them out from oldest to newest
    GList* itemsToPrint = NULL;
    gint forwardListCount = webkit_web_back_forward_list_get_forward_length(bfList);
    for (int i = forwardListCount; i > 0; i--) {
        WebKitWebHistoryItem* item = webkit_web_back_forward_list_get_nth_item(bfList, i);
        // something is wrong if the item from the last test is in the forward part of the b/f list
        ASSERT(item != prevTestBFItem);
        g_object_ref(item);
        itemsToPrint = g_list_append(itemsToPrint, item);
    }

    WebKitWebHistoryItem* currentItem = webkit_web_back_forward_list_get_current_item(bfList);

    g_object_ref(currentItem);
    itemsToPrint = g_list_append(itemsToPrint, currentItem);

    gint currentItemIndex = g_list_length(itemsToPrint) - 1;
    gint backListCount = webkit_web_back_forward_list_get_back_length(bfList);
    for (int i = -1; i >= -(backListCount); i--) {
        WebKitWebHistoryItem* item = webkit_web_back_forward_list_get_nth_item(bfList, i);
        if (item == prevTestBFItem)
            break;
        g_object_ref(item);
        itemsToPrint = g_list_append(itemsToPrint, item);
    }

    for (int i = g_list_length(itemsToPrint) - 1; i >= 0; i--) {
        WebKitWebHistoryItem* item = WEBKIT_WEB_HISTORY_ITEM(g_list_nth_data(itemsToPrint, i));
        dumpHistoryItem(item, historyItemIndent, i == currentItemIndex);
        g_object_unref(item);
    }
    g_list_free(itemsToPrint);
    printf("===============================================\n");
}

static void dumpBackForwardListForAllWebPages()
{
    // Dump the back forward list of the main WebPage first
    dumpBackForwardListForWebPage(webPage);

    // The page list is prepended. Reverse the list so we get the order right.
    GSList* pageList = g_slist_reverse(webPageList);
    for (unsigned i = 0; i < g_slist_length(pageList); ++i)
        dumpBackForwardListForWebPage(WEBKIT_WEB_PAGE(g_slist_nth_data(pageList, i)));
}

static void invalidateAnyPreviousWaitToDumpWatchdog()
{
    if (waitToDumpWatchdog) {
        g_source_remove(waitToDumpWatchdog);
        waitToDumpWatchdog = 0;
    }

    waitForPolicy = false;
}

static void resetDefaultsToConsistentValues()
{
    WebKitWebSettings* settings = webkit_web_page_get_settings(webPage);
    g_object_set(G_OBJECT(settings),
                 "enable-private-browsing", FALSE,
                 "enable-developer-extras", FALSE,
                 "enable-spell-checking", TRUE,
                 "enable-html5-database", TRUE,
                 "enable-html5-local-storage", TRUE,
                 "enable-xss-auditor", FALSE,
                 "enable-spatial-navigation", FALSE,
                 "enable-frame-flattening", FALSE,
                 "javascript-can-access-clipboard", TRUE,
                 "javascript-can-open-windows-automatically", TRUE,
                 "enable-offline-web-application-cache", TRUE,
                 "enable-universal-access-from-file-uris", TRUE,
                 "enable-scripts", TRUE,
                 "enable-dom-paste", TRUE,
                 "default-font-family", "Times",
                 "monospace-font-family", "Courier",
                 "serif-font-family", "Times",
                 "sans-serif-font-family", "Helvetica",
                 "cursive-font-family", "cursive",
                 "fantasy-font-family", "fantasy",
                 "default-font-size", 16,
                 "default-monospace-font-size", 13,
                 "minimum-font-size", 1,
                 "enable-caret-browsing", FALSE,
                 "enable-page-cache", FALSE,
                 "auto-resize-window", TRUE,
                 "enable-java-applet", FALSE,
                 "enable-plugins", TRUE,
                 "enable-hyperlink-auditing", FALSE,
                 "editing-behavior", WEBKIT_EDITING_BEHAVIOR_UNIX,
                 "enable-fullscreen", TRUE,
                 NULL);
    webkit_web_page_set_settings(webPage, settings);

    webkit_web_frame_clear_main_frame_name(mainFrame);

    WebKitWebInspector* inspector = webkit_web_page_get_inspector(webPage);
    g_object_set(G_OBJECT(inspector), "javascript-profiling-enabled", FALSE, NULL);

    webkit_web_page_set_zoom_level(webPage, 1.0);

    //DumpRenderTreeSupportClutter::resetOriginAccessWhiteLists();

    WebKitWebBackForwardList* list = webkit_web_page_get_back_forward_list(webPage);
    webkit_web_back_forward_list_clear(list);

    SoupSession* session = webkit_get_default_session();
    SoupCookieJar* jar = reinterpret_cast<SoupCookieJar*>(soup_session_get_feature(session, SOUP_TYPE_COOKIE_JAR));

    // We only create the jar when the soup backend needs to do
    // HTTP. Should we initialize it earlier, perhaps?
    if (jar)
        g_object_set(G_OBJECT(jar), SOUP_COOKIE_JAR_ACCEPT_POLICY, SOUP_COOKIE_JAR_ACCEPT_NO_THIRD_PARTY, NULL);

    setlocale(LC_ALL, "");

    DumpRenderTreeSupportClutter::setLinksIncludedInFocusChain(true);

    DumpRenderTreeSupportClutter::removeAllUserContent(mainFrame);
    DumpRenderTreeSupportClutter::clearScriptWorlds();

    DumpRenderTreeSupportClutter::resetGeolocationClientMock(webPage);
}

void displayWebView()
{
    clutter_actor_queue_redraw(CLUTTER_ACTOR(webActor));
}

static void appendString(gchar*& target, gchar* string)
{
    gchar* oldString = target;
    target = g_strconcat(target, string, NULL);
    g_free(oldString);
}

static gchar* dumpFramesAsText(WebKitWebFrame* frame)
{
    gchar* result = 0;

    // Add header for all but the main frame.
    bool isMainFrame = (webkit_web_page_get_main_frame(webPage) == frame);

    CString innerText = DumpRenderTreeSupportClutter::getInnerText(frame);
    if (isMainFrame)
        result = g_strdup_printf("%s\n", innerText.data());
    else {
        const gchar* frameName = webkit_web_frame_get_name(frame);
        result = g_strdup_printf("\n--------\nFrame: '%s'\n--------\n%s\n", frameName, innerText.data());
    }

    if (gTestRunner->dumpChildFramesAsText()) {
        GSList* children = DumpRenderTreeSupportClutter::getFrameChildren(frame);
        for (GSList* child = children; child; child = g_slist_next(child))
            appendString(result, dumpFramesAsText(static_cast<WebKitWebFrame* >(child->data)));
        g_slist_free(children);
    }

    return result;
}

void dump()
{
    invalidateAnyPreviousWaitToDumpWatchdog();

    if (dumpTree) {
        char* result = 0;
        gchar* responseMimeType = webkit_web_frame_get_response_mime_type(mainFrame);

        if (g_str_equal(responseMimeType, "text/plain")) {
            gTestRunner->setDumpAsText(true);
            gTestRunner->setGeneratePixelResults(false);
        }
        g_free(responseMimeType);

        if (gTestRunner->dumpAsText())
            result = dumpFramesAsText(mainFrame);
        else {
            while (clutter_events_pending())
                g_main_context_iteration(NULL, TRUE);

            result = g_strdup(DumpRenderTreeSupportClutter::dumpRenderTree(mainFrame, gTestRunner->isPrinting()).data());
        }

        if (!result) {
            const char* errorMessage;
            if (gTestRunner->dumpAsText())
                errorMessage = "[documentElement innerText]";
            else if (gTestRunner->dumpDOMAsWebArchive())
                errorMessage = "[[mainFrame DOMDocument] webArchive]";
            else if (gTestRunner->dumpSourceAsWebArchive())
                errorMessage = "[[mainFrame dataSource] webArchive]";
            else
                errorMessage = "[mainFrame renderTreeAsExternalRepresentation]";
            printf("ERROR: nil result from %s", errorMessage);
        } else {
            printf("%s", result);
            g_free(result);
            if (!gTestRunner->dumpAsText() && !gTestRunner->dumpDOMAsWebArchive() && !gTestRunner->dumpSourceAsWebArchive()) { }

            if (gTestRunner->dumpBackForwardList())
                dumpBackForwardListForAllWebPages();
        }

        if (printSeparators) {
            puts("#EOF"); // terminate the content block
            fputs("#EOF\n", stderr);
            fflush(stdout);
            fflush(stderr);
        }
    }

    // No pixel dumps support yet.
    // if (dumpPixels
    //  && gTestRunner->generatePixelResults()
    //  && !gTestRunner->dumpDOMAsWebArchive()
    //  && !gTestRunner->dumpSourceAsWebArchive())
    //     dumpWebViewAsPixelsAndCompareWithExpected(gTestRunner->expectedPixelHash());

    // FIXME: call displayWebView here when we support --paint

    done = true;
    clutter_main_quit();
}

void webPageLoadStarted(WebKitWebPage* page, WebKitWebFrame* frame, void*)
{
    // Make sure we only set this once per test.  If it gets cleared, and then set again, we might
    // end up doing two dumps for one test.
    if (!topLoadingFrame && !done)
        topLoadingFrame = frame;
}

static gboolean processWork(void* data)
{
    // if we finish all the commands, we're ready to dump state
    if (WorkQueue::shared()->processWork() && !gTestRunner->waitToDump())
        dump();

    return FALSE;
}

static void webPageLoadFinished(WebKitWebPage* page, WebKitWebFrame* frame, void*)
{
    if (frame != topLoadingFrame)
        return;

    topLoadingFrame = 0;
    WorkQueue::shared()->setFrozen(true); // first complete load freezes the queue for the rest of this test
    if (gTestRunner->waitToDump())
        return;

    if (WorkQueue::shared()->count())
        g_timeout_add(0, processWork, 0);
    else
        dump();
}

static void webPageWindowObjectCleared(WebKitWebPage* page, WebKitWebFrame* frame, JSGlobalContextRef context, JSObjectRef windowObject, gpointer data)
{
    JSValueRef exception = 0;
    ASSERT(gTestRunner);

    gTestRunner->makeWindowObject(context, windowObject, &exception);
    ASSERT(!exception);

    gcController->makeWindowObject(context, windowObject, &exception);
    ASSERT(!exception);

    plainTextController->makeWindowObject(context, windowObject, &exception);
    ASSERT(!exception);

    // axController->makeWindowObject(context, windowObject, &exception);
    // ASSERT(!exception);

    JSStringRef eventSenderStr = JSStringCreateWithUTF8CString("eventSender");
    JSValueRef eventSender = makeEventSender(context, !webkit_web_frame_get_parent(frame));
    JSObjectSetProperty(context, windowObject, eventSenderStr, eventSender, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete, 0);
    JSStringRelease(eventSenderStr);
    WebCoreTestSupport::injectInternalsObject(context);
}

static gboolean webPageConsoleMessage(WebKitWebView* view, const gchar* message, unsigned int line, const gchar* sourceId, gpointer data)
{
    gchar* testMessage = 0;
    const gchar* uriScheme;

    // Tests expect only the filename part of local URIs
    uriScheme = g_strstr_len(message, -1, "file://");
    if (uriScheme) {
        GString* tempString = g_string_sized_new(strlen(message));
        gchar* filename = g_strrstr(uriScheme, G_DIR_SEPARATOR_S);

        if (filename) {
            filename += strlen(G_DIR_SEPARATOR_S);
            tempString = g_string_append_len(tempString, message, (uriScheme - message));
            tempString = g_string_append_len(tempString, filename, strlen(filename));
            testMessage = g_string_free(tempString, FALSE);
        }
    }

    fprintf(stdout, "CONSOLE MESSAGE: ");
    if (line)
        fprintf(stdout, "line %d: ", line);
    fprintf(stdout, "%s\n", testMessage ? testMessage : message);
    g_free(testMessage);

    return TRUE;
}

static void didReceiveResponseCallback(GObject* object, const gchar* file, const gchar* mimeType, void*)
{
    if (gTestRunner->dumpResourceResponseMIMETypes())
        printf("%s has MIME type %s\n", g_path_get_basename(file), mimeType);
}

static char* getFrameNameSuitableForTestResult(WebKitWebPage* view, WebKitWebFrame* frame)
{
    char* frameName = g_strdup(webkit_web_frame_get_name(frame));

    if (frame == webkit_web_page_get_main_frame(view)) {
        // This is a bit strange. Shouldn't web_frame_get_name return NULL?
        if (frameName && (frameName[0] != '\0')) {
            char* tmp = g_strdup_printf("main frame \"%s\"", frameName);
            g_free(frameName);
            frameName = tmp;
        } else {
            g_free(frameName);
            frameName = g_strdup("main frame");
        }
    } else if (!frameName || (frameName[0] == '\0')) {
        g_free(frameName);
        frameName = g_strdup("frame (anonymous)");
    } else {
        char* tmp = g_strdup_printf("frame \"%s\"", frameName);
        g_free(frameName);
        frameName = tmp;
    }

    return frameName;
}

static gboolean webPageScriptAlert(WebKitWebView* view, WebKitWebFrame* frame, const gchar* message, gpointer data)
{
    printf("ALERT: %s\n", message);
    return TRUE;
}

static gboolean webPageScriptPrompt(WebKitWebPage* webView, WebKitWebFrame* frame, const gchar* message, const gchar* defaultValue, gchar** value, gpointer data)
{
    fprintf(stdout, "PROMPT: %s, default text: %s\n", message, defaultValue);
    *value = g_strdup(defaultValue);
    return TRUE;
}

static gboolean webPageScriptConfirm(WebKitWebPage* view, WebKitWebFrame* frame, const gchar* message, gboolean* didConfirm, gpointer data)
{
    fprintf(stdout, "CONFIRM: %s\n", message);
    *didConfirm = TRUE;
    return TRUE;
}

static void webPageTitleChanged(WebKitWebPage* view, WebKitWebFrame* frame, const gchar* title, gpointer data)
{
    if (gTestRunner->dumpTitleChanges() && !done)
        printf("TITLE CHANGED: %s\n", title ? title : "");
}

static void webPageStatusBarTextChanged(WebKitWebPage* page, const gchar* message, gpointer data)
{
    // Are we doing anything wrong? One test that does not call
    // dumpStatusCallbacks gets true here
    if (gTestRunner->dumpStatusCallbacks()) {
        if (message && strcmp(message, ""))
            printf("UI DELEGATE STATUS CALLBACK: setStatusText:%s\n", message);
    }
}

static void databaseQuotaExceeded(WebKitWebPage* page, WebKitWebFrame* frame, WebKitWebDatabase *database)
{
    ASSERT(page);
    ASSERT(frame);
    ASSERT(database);

    WebKitSecurityOrigin* origin = webkit_web_database_get_security_origin(database);
    if (gTestRunner->dumpDatabaseCallbacks()) {
        printf("UI DELEGATE DATABASE CALLBACK: exceededDatabaseQuotaForSecurityOrigin:{%s, %s, %i} database:%s\n",
            webkit_security_origin_get_protocol(origin),
            webkit_security_origin_get_host(origin),
            webkit_security_origin_get_port(origin),
            webkit_web_database_get_name(database));
    }
    webkit_security_origin_set_web_database_quota(origin, 5 * 1024 * 1024);
}

static bool
geolocationPolicyDecisionRequested(WebKitWebPage*, WebKitWebFrame*, WebKitGeolocationPolicyDecision* decision)
{
    if (!gTestRunner->isGeolocationPermissionSet())
        return FALSE;
    if (gTestRunner->geolocationPermission())
        webkit_geolocation_policy_allow(decision);
    else
        webkit_geolocation_policy_deny(decision);

    return TRUE;
}

static void webPageDocumentLoadFinished(WebKitWebPage* view, WebKitWebFrame* frame, void*)
{
    if (!done && gTestRunner->dumpFrameLoadCallbacks()) {
        char* frameName = getFrameNameSuitableForTestResult(view, frame);
        printf("%s - didFinishDocumentLoadForFrame\n", frameName);
        g_free(frameName);
    } else if (!done) {
        guint pendingFrameUnloadEvents = DumpRenderTreeSupportClutter::getPendingUnloadEventCount(frame);
        if (pendingFrameUnloadEvents) {
            char* frameName = getFrameNameSuitableForTestResult(view, frame);
            printf("%s - has %u onunload handler(s)\n", frameName, pendingFrameUnloadEvents);
            g_free(frameName);
        }
    }
}

static void topLoadingFrameLoadFinished()
{
    topLoadingFrame = 0;
    WorkQueue::shared()->setFrozen(true); // first complete load freezes the queue for the rest of this test
    if (gTestRunner->waitToDump())
        return;

    if (WorkQueue::shared()->count())
        g_timeout_add(0, processWork, 0);
    else
        dump();
}

static void webFrameLoadStatusChanged(GObject*, WebKitWebFrame* frame, gpointer user_data)
{
    WebKitLoadStatus loadStatus = webkit_web_frame_get_load_status(frame);
    GOwnPtr<char> frameName(getFrameNameSuitableForTestResult(webkit_web_frame_get_page(frame), frame));

    if (gTestRunner->dumpFrameLoadCallbacks()) {
        switch (loadStatus) {
        case WEBKIT_LOAD_PROVISIONAL:
            if (!done)
                printf("%s - didStartProvisionalLoadForFrame\n", frameName.get());
            break;
        case WEBKIT_LOAD_COMMITTED:
            if (!done)
                printf("%s - didCommitLoadForFrame\n", frameName.get());
            break;
        case WEBKIT_LOAD_FINISHED:
            if (frame != topLoadingFrame || !done)
                printf("%s - didFinishLoadForFrame\n", frameName.get());
            break;
        default:
            break;
        }
    }

    if ((loadStatus == WEBKIT_LOAD_FINISHED || loadStatus == WEBKIT_LOAD_FAILED)
        && frame == topLoadingFrame)
        topLoadingFrameLoadFinished();
}

static void didChangeIconsCallback(GObject*, WebKitWebFrame* frame, gpointer user_data)
{
    if (gTestRunner->dumpIconChanges()) {
        GOwnPtr<char> frameName(getFrameNameSuitableForTestResult(webkit_web_frame_get_page(frame), frame));
        printf("%s - didChangeIcons\n", frameName.get());
    }
}

static void setDefaultsToConsistentStateValuesForTesting()
{
    clutter_backend_set_resolution(clutter_get_default_backend(), 72.0);

    resetDefaultsToConsistentValues();

#if PLATFORM(X11)
    //webkit_web_settings_add_extra_plugin_directory(webView, TEST_PLUGIN_DIR);
#endif

    gchar* databaseDirectory = g_build_filename(g_get_user_data_dir(), "clutterwebkitdrt", "databases", NULL);
    webkit_set_web_database_directory_path(databaseDirectory);
    g_free(databaseDirectory);
}

static WebKitWebActor* createWebActor()
{
    // It is important to declare DRT is running early so when creating
    // web view mock clients are used instead of proper ones.
    DumpRenderTreeSupportClutter::setDumpRenderTreeModeEnabled(true);

    WebKitWebActor* actor = WEBKIT_WEB_ACTOR(webkit_web_actor_new());
    WebKitWebPage* page = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(actor));

    g_object_connect(G_OBJECT(page),
                     "signal::web-load-started", webPageLoadStarted, 0,
                     "signal::web-load-finished", webPageLoadFinished, 0,
                     "signal::window-object-cleared", webPageWindowObjectCleared, 0,
                     "signal::console-message", webPageConsoleMessage, 0,
                     "signal::script-alert", webPageScriptAlert, 0,
                     "signal::script-prompt", webPageScriptPrompt, 0,
                     "signal::script-confirm", webPageScriptConfirm, 0,
                     "signal::title-changed", webPageTitleChanged, 0,
    //                  "signal::navigation-policy-decision-requested", webViewNavigationPolicyDecisionRequested, 0,
                     "signal::status-bar-text-changed", webPageStatusBarTextChanged, 0,
    //                  "signal::create-web-view", webViewCreate, 0,
    //                  "signal::close-web-view", webViewClose, 0,
                     "signal::database-quota-exceeded", databaseQuotaExceeded, 0,
                     "signal::document-load-finished", webPageDocumentLoadFinished, 0,
                     "signal::geolocation-policy-decision-requested", geolocationPolicyDecisionRequested, 0,
    //                  "signal::onload-event", webViewOnloadEvent, 0,
    //                  "signal::drag-begin", dragBeginCallback, 0,
    //                  "signal::drag-end", dragEndCallback, 0,
    //                  "signal::drag-failed", dragFailedCallback, 0,
                     NULL);
    // connectEditingCallbacks(view);

    // WebKitWebInspector* inspector = webkit_web_page_get_inspector(page);
    // g_object_connect(G_OBJECT(inspector),
    //                  "signal::inspect-web-view", webInspectorInspectWebView, 0,
    //                  "signal::show-window", webInspectorShowWindow, 0,
    //                  "signal::close-window", webInspectorCloseWindow, 0,
    //                  NULL);

    if (webPage) {
        WebKitWebSettings* settings = webkit_web_page_get_settings(webPage);
        webkit_web_page_set_settings(page, settings);
    }

    g_object_connect(G_OBJECT(DumpRenderTreeSupportClutter::gobject()),
                     "signal::did-receive-response", didReceiveResponseCallback, 0,
                     "signal::frame-load-status-changed", webFrameLoadStatusChanged, 0,
                     "signal::did-change-icons", didChangeIconsCallback, 0,
                     NULL);

    return actor;
}

static void sendPixelResultsEOF()
{
    puts("#EOF");

    fflush(stdout);
    fflush(stderr);
}

static void runTest(const string& testPathOrURL)
{
    ASSERT(!testPathOrURL.empty());

    // Look for "'" as a separator between the path or URL, and the pixel dump hash that follows.
    string testURL(testPathOrURL);
    string expectedPixelHash;
    size_t separatorPos = testURL.find("'");
    if (separatorPos != string::npos) {
        testURL = string(testPathOrURL, 0, separatorPos);
        expectedPixelHash = string(testPathOrURL, separatorPos + 1);
    }

    // Convert the path into a full file URL if it does not look
    // like an HTTP/S URL (doesn't start with http:// or https://).
    if (testURL.find("http://") && testURL.find("https://")) {
        GFile* testFile = g_file_new_for_path(testURL.c_str());
        gchar* testURLCString = g_file_get_uri(testFile);
        testURL = testURLCString;
        g_free(testURLCString);
        g_object_unref(testFile);
    }

    resetDefaultsToConsistentValues();

    gTestRunner = TestRunner::create(testURL, expectedPixelHash);
    topLoadingFrame = 0;
    done = false;

    gTestRunner->setIconDatabaseEnabled(false);

    if (shouldLogFrameLoadDelegates(testURL))
        gTestRunner->setDumpFrameLoadCallbacks(true);

    gTestRunner->setDeveloperExtrasEnabled(true);
    // if (shouldOpenWebInspector(testURL))
    //     gTestRunner->showWebInspector();

    // WorkQueue::shared()->clear();
    // WorkQueue::shared()->setFrozen(false);

    bool isSVGW3CTest = (testURL.find("svg/W3C-SVG-1.1") != string::npos);
    ClutterGeometry size;
    size.x = size.y = 0;
    size.width = isSVGW3CTest ? 480 : TestRunner::maxViewWidth;
    size.height = isSVGW3CTest ? 360 : TestRunner::maxViewHeight;

    // FIXME: this should apply to the window or the container.
    clutter_actor_set_size(CLUTTER_ACTOR(webActor), size.width, size.height);

    if (prevTestBFItem)
        g_object_unref(prevTestBFItem);
    WebKitWebBackForwardList* bfList = webkit_web_page_get_back_forward_list(webPage);
    prevTestBFItem = webkit_web_back_forward_list_get_current_item(bfList);
    if (prevTestBFItem)
        g_object_ref(prevTestBFItem);

    initializeFonts(testURL.c_str());

    // Focus the web view before loading the test to avoid focusing problems
    clutter_actor_grab_key_focus(CLUTTER_ACTOR(webActor));
    webkit_web_page_load_uri(webPage, testURL.c_str());

    clutter_main();

    // If developer extras enabled Web Inspector may have been open by the test.
    // gTestRunner->closeWebInspector();
    gTestRunner->setDeveloperExtrasEnabled(false);

    // Also check if we still have opened webViews and free them.
    // if (gTestRunner->closeRemainingWindowsWhenComplete() || webViewList) {
    //     while (webViewList) {
    //         g_object_unref(WEBKIT_WEB_VIEW(webViewList->data));
    //         webViewList = g_slist_next(webViewList);
    //     }
    //     g_slist_free(webViewList);
    //     webViewList = 0;
    // }

    WebCoreTestSupport::resetInternalsObject(webkit_web_frame_get_global_context(mainFrame));

    // A blank load seems to be necessary to reset state after certain tests.
    webkit_web_page_load_uri(webPage, "about:blank");

    gTestRunner.clear();

    // terminate the (possibly empty) pixels block after all the state reset
    sendPixelResultsEOF();
}

static bool useLongRunningServerMode(int argc, char *argv[])
{
    // This assumes you've already called getopt_long
    return (argc == optind+1 && !strcmp(argv[optind], "-"));
}

static void runTestingServerLoop()
{
    // When DumpRenderTree runs in server mode, we just wait around for file names
    // to be passed to us and read each in turn, passing the results back to the client
    char filenameBuffer[2048];
    while (fgets(filenameBuffer, sizeof(filenameBuffer), stdin)) {
        char* newLineCharacter = strchr(filenameBuffer, '\n');
        if (newLineCharacter)
            *newLineCharacter = '\0';

        if (!strlen(filenameBuffer))
            continue;

        runTest(filenameBuffer);
    }
}

static void logHandler(const gchar* domain, GLogLevelFlags level, const gchar* message, gpointer data)
{
    if (level < G_LOG_LEVEL_DEBUG)
        fprintf(stderr, "%s\n", message);
}

int main(int argc, char** argv)
{
    if (clutter_init(&argc, &argv) != CLUTTER_INIT_SUCCESS) {
        g_error ("Failed to initialized clutter.");
        return 1;
    }

    webkit_init();

    // Some plugins might try to use the GLib logger for printing debug
    // messages. This will cause tests to fail because of unexpected output.
    // We squelch all debug messages sent to the logger.
    g_log_set_default_handler(logHandler, 0);

    // FIXME: handle command line arguments. GTK+'s DRT has nice
    // functions we could share.

    stage = clutter_stage_get_default();
    clutter_stage_set_throttle_motion_events(CLUTTER_STAGE(stage), FALSE);
    clutter_actor_set_size(stage, 800, 600);
    clutter_actor_show(stage);

    webActor = createWebActor();
    clutter_container_add_actor(CLUTTER_CONTAINER(stage), CLUTTER_ACTOR(webActor));
    clutter_actor_show(CLUTTER_ACTOR(webActor));

    webPage = webkit_iweb_view_get_page(WEBKIT_IWEB_VIEW(webActor));
    mainFrame = webkit_web_page_get_main_frame(webPage);

    setDefaultsToConsistentStateValuesForTesting();

    gcController = new GCController();
    plainTextController = new PlainTextController();

    if (useLongRunningServerMode(argc, argv)) {
        printSeparators = true;
        runTestingServerLoop();
    } else {
        printSeparators = (optind < argc-1 || (dumpPixels && dumpTree));
        for (int i = optind; i != argc; ++i)
            runTest(argv[i]);
    }

    delete gcController;
    gcController = 0;

    delete plainTextController;
    plainTextController = 0;

    // delete axController;
    // axController = 0;

    clutter_actor_destroy(stage);

    return 0;
}
