/*
 * Copyright (C) 2007, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2009 Zan Dobersek <zandobersek@gmail.com>
 * Copyright (C) 2009 Holger Hans Peter Freyther
 * Copyright (C) 2010 Igalia S.L.
 * Copyright (C) 2011 ProFUSION Embedded Systems
 * Copyright (C) 2011, 2012 Samsung Electronics
 * Copyright (C) 2012 Collabora Ltd.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "EventSender.h"

#include "DumpRenderTree.h"
#include "WebCoreSupport/DumpRenderTreeSupportClutter.h"
#include "IntPoint.h"
#include "PlatformTouchPoint.h"
#include <JavaScriptCore/JSBase.h>
#include <JavaScriptCore/JSObjectRef.h>
#include <JavaScriptCore/JSRetainPtr.h>
#include <JavaScriptCore/JSStringRef.h>
#include <webkit/webkitgwebframe.h>
#include <webkit/webkitwebpage.h>
#include <wtf/ASCIICType.h>

#include <clutter/clutter.h>

#include <stdio.h>


static bool dragMode;
static int timeOffset = 0;

static int lastMousePositionX;
static int lastMousePositionY;
static int lastClickPositionX;
static int lastClickPositionY;
static int lastClickTimeOffset;
static int lastClickButton;
static int buttonCurrentlyDown;
static int clickCount;

static void sendOrQueueEvent(ClutterEvent*, bool = true);
static void dispatchEvent(ClutterEvent* event);

// Key event location code defined in DOM Level 3.
enum KeyLocationCode {
    DOM_KEY_LOCATION_STANDARD      = 0x00,
    DOM_KEY_LOCATION_LEFT          = 0x01,
    DOM_KEY_LOCATION_RIGHT         = 0x02,
    DOM_KEY_LOCATION_NUMPAD        = 0x03
};

bool prepareMouseButtonEvent(ClutterEvent* event, int eventSenderButtonNumber, guint modifiers)
{
    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    if (!page)
        return false;

    // The logic for mapping EventSender button numbers to GDK button
    // numbers originates from the Windows EventSender.
    int clutterButtonNumber = 0;
    if (eventSenderButtonNumber >= 0 && eventSenderButtonNumber <= 2)
        clutterButtonNumber = eventSenderButtonNumber + 1;

    // fast/events/mouse-click-events expects the 4th button
    // to be event->button = 1, so send a middle-button event.
    else if (eventSenderButtonNumber == 3)
        clutterButtonNumber = 2;

    event->button.time = g_get_monotonic_time()/1000;
    event->button.flags = CLUTTER_EVENT_NONE;

    ClutterActor* actor = CLUTTER_ACTOR(webkit_web_page_get_view(page));
    event->button.stage = CLUTTER_STAGE(clutter_actor_get_stage(actor));

    event->button.x = lastMousePositionX;
    event->button.y = lastMousePositionY;
    event->button.modifier_state = (ClutterModifierType)modifiers;
    event->button.button = clutterButtonNumber;
    event->button.click_count = 0;
    event->button.axes = 0;
    event->button.device = clutter_device_manager_get_core_device(clutter_device_manager_get_default(),
                                                                  CLUTTER_POINTER_DEVICE);

    return true;
}

static void updateClickCount(int button)
{
    if (lastClickPositionX != lastMousePositionX
        || lastClickPositionY != lastMousePositionY
        || lastClickButton != button
        || timeOffset - lastClickTimeOffset >= 1)
        clickCount = 1;
    else
        clickCount++;
}

static guint clutterModifersFromJSValue(JSContextRef context, const JSValueRef modifiers)
{
    JSObjectRef modifiersArray = JSValueToObject(context, modifiers, 0);
    if (!modifiersArray)
        return 0;

    guint clutterModifiers = 0;
    int modifiersCount = JSValueToNumber(context, JSObjectGetProperty(context, modifiersArray, JSStringCreateWithUTF8CString("length"), 0), 0);
    for (int i = 0; i < modifiersCount; ++i) {
        JSValueRef value = JSObjectGetPropertyAtIndex(context, modifiersArray, i, 0);
        JSStringRef string = JSValueToStringCopy(context, value, 0);
        if (JSStringIsEqualToUTF8CString(string, "ctrlKey")
            || JSStringIsEqualToUTF8CString(string, "addSelectionKey"))
            clutterModifiers |= CLUTTER_CONTROL_MASK;
        else if (JSStringIsEqualToUTF8CString(string, "shiftKey")
                 || JSStringIsEqualToUTF8CString(string, "rangeSelectionKey"))
            clutterModifiers |= CLUTTER_SHIFT_MASK;
        else if (JSStringIsEqualToUTF8CString(string, "altKey"))
            clutterModifiers |= CLUTTER_MOD1_MASK;
        else if (JSStringIsEqualToUTF8CString(string, "metaKey"))
            clutterModifiers |= CLUTTER_META_MASK;

        JSStringRelease(string);
    }
    return clutterModifiers;
}

static JSValueRef mouseScrollByCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
/*
    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    if (!page)
        return JSValueMakeUndefined(context);

    if (argumentCount < 2)
        return JSValueMakeUndefined(context);

    int horizontal = (int)JSValueToNumber(context, arguments[0], exception);
    g_return_val_if_fail((!exception || !*exception), JSValueMakeUndefined(context));
    int vertical = (int)JSValueToNumber(context, arguments[1], exception);
    g_return_val_if_fail((!exception || !*exception), JSValueMakeUndefined(context));

    // GTK+ doesn't support multiple direction scrolls in the same event!
    g_return_val_if_fail((!vertical || !horizontal), JSValueMakeUndefined(context));

    GdkEvent* event = gdk_event_new(GDK_SCROLL);
    event->scroll.x = lastMousePositionX;
    event->scroll.y = lastMousePositionY;
    event->scroll.time = GDK_CURRENT_TIME;
    event->scroll.window = gtk_widget_get_window(GTK_WIDGET(view));
    g_object_ref(event->scroll.window);

    if (horizontal < 0)
        event->scroll.direction = GDK_SCROLL_RIGHT;
    else if (horizontal > 0)
        event->scroll.direction = GDK_SCROLL_LEFT;
    else if (vertical < 0)
        event->scroll.direction = GDK_SCROLL_DOWN;
    else if (vertical > 0)
        event->scroll.direction = GDK_SCROLL_UP;
    else
        g_assert_not_reached();

    sendOrQueueEvent(event);
*/
    return JSValueMakeUndefined(context);
}

static JSValueRef continuousMouseScrollByCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    // GTK doesn't support continuous scroll events.
    return JSValueMakeUndefined(context);
}

static JSValueRef contextClickCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
/*
    GdkEvent* pressEvent = gdk_event_new(GDK_BUTTON_PRESS);

    if (!prepareMouseButtonEvent(pressEvent, 2, 0))
        return JSObjectMakeArray(context, 0, 0, 0);

    GdkEvent* releaseEvent = gdk_event_copy(pressEvent);
    sendOrQueueEvent(pressEvent);

    JSValueRef valueRef = JSObjectMakeArray(context, 0, 0, 0);
    WebKitWebView* view = webkit_web_frame_get_web_view(mainFrame);
    GtkMenu* gtkMenu = webkit_web_view_get_context_menu(view);
    if (gtkMenu) {
        GList* items = gtk_container_get_children(GTK_CONTAINER(gtkMenu));
        JSValueRef arrayValues[g_list_length(items)];
        int index = 0;
        for (GList* item = g_list_first(items); item; item = g_list_next(item)) {
            arrayValues[index] = JSObjectMake(context, getMenuItemClass(), item->data);
            index++;
        }
        if (index)
            valueRef = JSObjectMakeArray(context, index - 1, arrayValues, 0);
    }

    releaseEvent->type = GDK_BUTTON_RELEASE;
    sendOrQueueEvent(releaseEvent);
    return valueRef;
*/
    return JSValueMakeUndefined(context);
}

static JSValueRef mouseDownCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    int button = 0;
    if (argumentCount == 1) {
        button = static_cast<int>(JSValueToNumber(context, arguments[0], exception));
        g_return_val_if_fail((!exception || !*exception), JSValueMakeUndefined(context));
    }
    guint modifiers = argumentCount >= 2 ? clutterModifersFromJSValue(context, arguments[1]) : 0;

    ClutterEvent* event = clutter_event_new(CLUTTER_BUTTON_PRESS);
    if (!prepareMouseButtonEvent(event, button, modifiers))
        return JSValueMakeUndefined(context);

    buttonCurrentlyDown = event->button.button;

    updateClickCount(event->button.button);
    event->button.click_count = clickCount;

    sendOrQueueEvent(event);

    return JSValueMakeUndefined(context);
}

static JSValueRef mouseUpCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    int button = 0;
    if (argumentCount == 1) {
        button = static_cast<int>(JSValueToNumber(context, arguments[0], exception));
        g_return_val_if_fail((!exception || !*exception), JSValueMakeUndefined(context));
    }
    guint modifiers = argumentCount >= 2 ? clutterModifersFromJSValue(context, arguments[1]) : 0;

    ClutterEvent* event = clutter_event_new(CLUTTER_BUTTON_RELEASE);
    if (!prepareMouseButtonEvent(event, button, modifiers))
        return JSValueMakeUndefined(context);

    lastClickPositionX = lastMousePositionX;
    lastClickPositionY = lastMousePositionY;
    lastClickButton = buttonCurrentlyDown;
    lastClickTimeOffset = timeOffset;
    buttonCurrentlyDown = 0;

    sendOrQueueEvent(event);

    return JSValueMakeUndefined(context);
}

static guint getStateFlags()
{
    if (buttonCurrentlyDown == 1)
        return CLUTTER_BUTTON1_MASK;
    if (buttonCurrentlyDown == 2)
        return CLUTTER_BUTTON2_MASK;
    if (buttonCurrentlyDown == 3)
        return CLUTTER_BUTTON3_MASK;
    return 0;
}

static JSValueRef mouseMoveToCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    if (!page)
        return JSValueMakeUndefined(context);

    if (argumentCount < 2)
        return JSValueMakeUndefined(context);

    lastMousePositionX = (int)JSValueToNumber(context, arguments[0], exception);
    g_return_val_if_fail((!exception || !*exception), JSValueMakeUndefined(context));
    lastMousePositionY = (int)JSValueToNumber(context, arguments[1], exception);
    g_return_val_if_fail((!exception || !*exception), JSValueMakeUndefined(context));

    ClutterEvent* event = clutter_event_new(CLUTTER_MOTION);
    event->motion.time = g_get_monotonic_time()/1000;
    event->motion.flags = CLUTTER_EVENT_NONE;

    ClutterActor* actor = CLUTTER_ACTOR(webkit_web_page_get_view(page));
    event->motion.stage = CLUTTER_STAGE(clutter_actor_get_stage(actor));

    event->motion.x = lastMousePositionX;
    event->motion.y = lastMousePositionY;
    event->motion.modifier_state = (ClutterModifierType)getStateFlags();
    event->motion.axes = 0;
    event->motion.device = clutter_device_manager_get_core_device(clutter_device_manager_get_default(),
                                                                  CLUTTER_POINTER_DEVICE);

    sendOrQueueEvent(event, false);

    return JSValueMakeUndefined(context);
}

static JSValueRef beginDragWithFilesCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    if (argumentCount < 1)
        return JSValueMakeUndefined(context);

    // FIXME: Implement this completely once WebCore has complete drag and drop support
    return JSValueMakeUndefined(context);
}

static void spinMainLoopToProcessEvents()
{
    // Force clutter to go through the queued events: we need to do two
    // iteration because ClutterClockSource has lower priority and it gets
    // kicked out by higher priority sources in the first one
    // FIXME: find a less magical solution, this may break if other higher
    // priority sources at different priority level are added
    g_main_context_iteration(NULL, FALSE);
    g_main_context_iteration(NULL, FALSE);
}

static void sendOrQueueEvent(ClutterEvent* event, bool shouldReplaySavedEvents)
{
/*
    // Mouse move events are queued if the previous event was queued or if a
    // delay was set up by leapForward().
    if ((dragMode && buttonCurrentlyDown) || endOfQueue != startOfQueue || msgQueue[endOfQueue].delay) {
        msgQueue[endOfQueue++].event = event;

        if (shouldReplaySavedEvents)
            replaySavedEvents();

        return;
    }
*/
    dispatchEvent(event);
}

static void dispatchEvent(ClutterEvent* event)
{
    // DumpRenderTreeSupportClutter::layoutFrame(mainFrame);
    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    if (!page) {
        clutter_event_free(event);
        return;
    }

    ClutterDeviceManager* manager = clutter_device_manager_get_default();
    ClutterInputDevice* device = clutter_device_manager_get_core_device(manager, CLUTTER_POINTER_DEVICE);
    clutter_input_device_update_from_event (device, event, TRUE);

    clutter_event_put(event);

    spinMainLoopToProcessEvents();

    clutter_event_free(event);
}

/*
void replaySavedEvents()
{
    // First send all the events that are ready to be sent
    while (startOfQueue < endOfQueue) {
        if (msgQueue[startOfQueue].delay) {
            g_usleep(msgQueue[startOfQueue].delay * 1000);
            msgQueue[startOfQueue].delay = 0;
        }

        dispatchEvent(msgQueue[startOfQueue++].event);
    }

    startOfQueue = 0;
    endOfQueue = 0;
}
*/

static JSValueRef getDragModeCallback(JSContextRef context, JSObjectRef object, JSStringRef propertyName, JSValueRef* exception)
{
    return JSValueMakeBoolean(context, dragMode);
}

static bool setDragModeCallback(JSContextRef context, JSObjectRef object, JSStringRef propertyName, JSValueRef value, JSValueRef* exception)
{
    // dragMode = JSValueToBoolean(context, value);
    return true;
}

static JSValueRef leapForwardCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    if (argumentCount < 1)
        return JSValueMakeUndefined(context);

    ASSERT(!exception || !*exception);

    spinMainLoopToProcessEvents();

    int milliseconds = JSValueToNumber(context, arguments[0], exception);
    timeOffset += milliseconds;
    g_usleep (milliseconds * 1000);

    return JSValueMakeUndefined(context);
}

static JSValueRef keyDownCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    if (argumentCount < 1)
        return JSValueMakeUndefined(context);
    guint modifiers = argumentCount >= 2 ? clutterModifersFromJSValue(context, arguments[1]) : 0;

    // handle location argument.
    int location = DOM_KEY_LOCATION_STANDARD;
    if (argumentCount > 2)
        location = (int)JSValueToNumber(context, arguments[2], exception);

    JSStringRef character = JSValueToStringCopy(context, arguments[0], exception);
    g_return_val_if_fail((!exception || !*exception), JSValueMakeUndefined(context));
    int clutterKeySym = CLUTTER_VoidSymbol;
    if (location == DOM_KEY_LOCATION_NUMPAD) {
        if (JSStringIsEqualToUTF8CString(character, "leftArrow"))
            clutterKeySym = CLUTTER_KP_Left;
        else if (JSStringIsEqualToUTF8CString(character, "rightArrow"))
            clutterKeySym = CLUTTER_KP_Right;
        else if (JSStringIsEqualToUTF8CString(character, "upArrow"))
            clutterKeySym = CLUTTER_KP_Up;
        else if (JSStringIsEqualToUTF8CString(character, "downArrow"))
            clutterKeySym = CLUTTER_KP_Down;
        else if (JSStringIsEqualToUTF8CString(character, "pageUp"))
            clutterKeySym = CLUTTER_KP_Page_Up;
        else if (JSStringIsEqualToUTF8CString(character, "pageDown"))
            clutterKeySym = CLUTTER_KP_Page_Down;
        else if (JSStringIsEqualToUTF8CString(character, "home"))
            clutterKeySym = CLUTTER_KP_Home;
        else if (JSStringIsEqualToUTF8CString(character, "end"))
            clutterKeySym = CLUTTER_KP_End;
        else if (JSStringIsEqualToUTF8CString(character, "insert"))
            clutterKeySym = CLUTTER_KP_Insert;
        else if (JSStringIsEqualToUTF8CString(character, "delete"))
            clutterKeySym = CLUTTER_KP_Delete;
        else
            // If we get some other key specified with the numpad location,
            // crash here, so we add it sooner rather than later.
            g_assert_not_reached();
    } else {
        if (JSStringIsEqualToUTF8CString(character, "leftArrow"))
            clutterKeySym = CLUTTER_Left;
        else if (JSStringIsEqualToUTF8CString(character, "rightArrow"))
            clutterKeySym = CLUTTER_Right;
        else if (JSStringIsEqualToUTF8CString(character, "upArrow"))
            clutterKeySym = CLUTTER_Up;
        else if (JSStringIsEqualToUTF8CString(character, "downArrow"))
            clutterKeySym = CLUTTER_Down;
        else if (JSStringIsEqualToUTF8CString(character, "pageUp"))
            clutterKeySym = CLUTTER_Page_Up;
        else if (JSStringIsEqualToUTF8CString(character, "pageDown"))
            clutterKeySym = CLUTTER_Page_Down;
        else if (JSStringIsEqualToUTF8CString(character, "home"))
            clutterKeySym = CLUTTER_Home;
        else if (JSStringIsEqualToUTF8CString(character, "end"))
            clutterKeySym = CLUTTER_End;
        else if (JSStringIsEqualToUTF8CString(character, "insert"))
            clutterKeySym = CLUTTER_Insert;
        else if (JSStringIsEqualToUTF8CString(character, "delete"))
            clutterKeySym = CLUTTER_Delete;
        else if (JSStringIsEqualToUTF8CString(character, "printScreen"))
            clutterKeySym = CLUTTER_Print;
        else if (JSStringIsEqualToUTF8CString(character, "F1"))
            clutterKeySym = CLUTTER_F1;
        else if (JSStringIsEqualToUTF8CString(character, "F2"))
            clutterKeySym = CLUTTER_F2;
        else if (JSStringIsEqualToUTF8CString(character, "F3"))
            clutterKeySym = CLUTTER_F3;
        else if (JSStringIsEqualToUTF8CString(character, "F4"))
            clutterKeySym = CLUTTER_F4;
        else if (JSStringIsEqualToUTF8CString(character, "F5"))
            clutterKeySym = CLUTTER_F5;
        else if (JSStringIsEqualToUTF8CString(character, "F6"))
            clutterKeySym = CLUTTER_F6;
        else if (JSStringIsEqualToUTF8CString(character, "F7"))
            clutterKeySym = CLUTTER_F7;
        else if (JSStringIsEqualToUTF8CString(character, "F8"))
            clutterKeySym = CLUTTER_F8;
        else if (JSStringIsEqualToUTF8CString(character, "F9"))
            clutterKeySym = CLUTTER_F9;
        else if (JSStringIsEqualToUTF8CString(character, "F10"))
            clutterKeySym = CLUTTER_F10;
        else if (JSStringIsEqualToUTF8CString(character, "F11"))
            clutterKeySym = CLUTTER_F11;
        else if (JSStringIsEqualToUTF8CString(character, "F12"))
            clutterKeySym = CLUTTER_F12;
        else {
            int charCode = JSStringGetCharactersPtr(character)[0];
            if (charCode == '\n' || charCode == '\r')
                clutterKeySym = CLUTTER_Return;
            else if (charCode == '\t')
                clutterKeySym = CLUTTER_Tab;
            else if (charCode == '\x8')
                clutterKeySym = CLUTTER_BackSpace;
            else {
                clutterKeySym = clutter_keysym_to_unicode(charCode);
                if (WTF::isASCIIUpper(charCode))
                    modifiers |= CLUTTER_SHIFT_MASK;
            }
        }
    }
    JSStringRelease(character);

    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    if (!page)
        return JSValueMakeUndefined(context);

    // create and send the event
    ClutterEvent* pressEvent = clutter_event_new(CLUTTER_KEY_PRESS);
    pressEvent->key.time = g_get_monotonic_time()/1000;
    pressEvent->key.flags = CLUTTER_EVENT_NONE;

    ClutterActor* actor = CLUTTER_ACTOR(webkit_web_page_get_view(page));
    pressEvent->key.stage = CLUTTER_STAGE(clutter_actor_get_stage(actor));
    pressEvent->key.source = actor;

    pressEvent->key.modifier_state = (ClutterModifierType)modifiers;
    pressEvent->key.keyval = clutterKeySym;
    pressEvent->key.hardware_keycode = 0;
    pressEvent->key.unicode_value = 0;
    pressEvent->key.device = clutter_device_manager_get_core_device(clutter_device_manager_get_default(),
                                                                    CLUTTER_KEYBOARD_DEVICE);

    ClutterEvent* releaseEvent = clutter_event_copy(pressEvent);
    dispatchEvent(pressEvent);
    releaseEvent->key.type = CLUTTER_KEY_RELEASE;
    dispatchEvent(releaseEvent);

    return JSValueMakeUndefined(context);
}

/*
static void zoomIn(gboolean fullContentsZoom)
{
    WebKitWebView* view = webkit_web_frame_get_web_view(mainFrame);
    if (!view)
        return;

    webkit_web_view_set_full_content_zoom(view, fullContentsZoom);
    gfloat currentZoom = webkit_web_view_get_zoom_level(view);
    webkit_web_view_set_zoom_level(view, currentZoom * zoomMultiplierRatio);
}

static void zoomOut(gboolean fullContentsZoom)
{
    WebKitWebView* view = webkit_web_frame_get_web_view(mainFrame);
    if (!view)
        return;

    webkit_web_view_set_full_content_zoom(view, fullContentsZoom);
    gfloat currentZoom = webkit_web_view_get_zoom_level(view);
    webkit_web_view_set_zoom_level(view, currentZoom / zoomMultiplierRatio);
}
*/
static JSValueRef textZoomInCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    // zoomIn(FALSE);
    return JSValueMakeUndefined(context);
}

static JSValueRef textZoomOutCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    // zoomOut(FALSE);
    return JSValueMakeUndefined(context);
}

static JSValueRef zoomPageInCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    // zoomIn(TRUE);
    return JSValueMakeUndefined(context);
}

static JSValueRef zoomPageOutCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    // zoomOut(TRUE);
    return JSValueMakeUndefined(context);
}

struct TouchEventInfo {
    TouchEventInfo(unsigned id, WebCore::PlatformTouchPoint::State state, const WebCore::IntPoint& point)
        : id(id)
        , state(state)
        , point(point)
    {
    }

    unsigned id;
    WebCore::PlatformTouchPoint::State state;
    WebCore::IntPoint point;
};

static unsigned touchModifiers;

WTF::Vector<TouchEventInfo>& touchPointList()
{
    DEFINE_STATIC_LOCAL(WTF::Vector<TouchEventInfo>, staticTouchPointList, ());
    return staticTouchPointList;
}

static ClutterEventType getEventTypeFromTouchState(WebCore::PlatformTouchPoint::State state)
{
    switch (state)
    {
    case WebCore::PlatformTouchPoint::TouchPressed:
        return CLUTTER_TOUCH_BEGIN;
    case WebCore::PlatformTouchPoint::TouchMoved:
        return CLUTTER_TOUCH_UPDATE;
    case WebCore::PlatformTouchPoint::TouchReleased:
        return CLUTTER_TOUCH_END;
    case WebCore::PlatformTouchPoint::TouchCancelled:
        return CLUTTER_TOUCH_CANCEL;
     default:
        ASSERT_NOT_REACHED();
        return CLUTTER_NOTHING;
   }
}

static void sendTouchEvent()
{
    for (unsigned i = 0; i < touchPointList().size(); ++i) {
        ClutterEventType type = getEventTypeFromTouchState(touchPointList().at(i).state);
        WebCore::IntPoint point = touchPointList().at(i).point;
        ClutterTouchEvent *event = (ClutterTouchEvent*)clutter_event_new(type);
        WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
        ClutterActor* actor = CLUTTER_ACTOR(webkit_web_page_get_view(page));
        
        event->time = g_get_monotonic_time() / 1000;
        event->flags = CLUTTER_EVENT_NONE;
        event->stage = CLUTTER_STAGE(clutter_actor_get_stage(actor));
        event->source = actor;
        event->x = point.x();
        event->y = point.y();
        event->sequence = (ClutterEventSequence*)GUINT_TO_POINTER(touchPointList().at(i).id);
        event->modifier_state = (ClutterModifierType)0;
        event->device = clutter_device_manager_get_core_device(clutter_device_manager_get_default(), CLUTTER_POINTER_DEVICE);

        dispatchEvent((ClutterEvent*)event);
    }

    for (unsigned i = 0; i < touchPointList().size(); ) {
        if (touchPointList().at(i).state == WebCore::PlatformTouchPoint::TouchReleased)
            touchPointList().remove(i);
        else {
            touchPointList().at(i).state = WebCore::PlatformTouchPoint::TouchStationary;
            ++i;
        }
    }
}

static JSValueRef addTouchPointCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    if (argumentCount != 2)
        return JSValueMakeUndefined(context);

    int x = static_cast<int>(JSValueToNumber(context, arguments[0], exception));
    ASSERT(!exception || !*exception);
    int y = static_cast<int>(JSValueToNumber(context, arguments[1], exception));
    ASSERT(!exception || !*exception);

    const WebCore::IntPoint point(x, y);
    const unsigned id = touchPointList().isEmpty() ? 0 : touchPointList().last().id + 1;
    TouchEventInfo eventInfo(id, WebCore::PlatformTouchPoint::TouchPressed, point);
    touchPointList().append(eventInfo);

    return JSValueMakeUndefined(context);
}

static JSValueRef touchStartCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    sendTouchEvent();
    return JSValueMakeUndefined(context);
}

static JSValueRef updateTouchPointCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    if (argumentCount != 3)
        return JSValueMakeUndefined(context);

    int index = static_cast<int>(JSValueToNumber(context, arguments[0], exception));
    ASSERT(!exception || !*exception);
    int x = static_cast<int>(JSValueToNumber(context, arguments[1], exception));
    ASSERT(!exception || !*exception);
    int y = static_cast<int>(JSValueToNumber(context, arguments[2], exception));
    ASSERT(!exception || !*exception);

    if (index < 0 || index >= static_cast<int>(touchPointList().size()))
        return JSValueMakeUndefined(context);

    WebCore::IntPoint& point = touchPointList().at(index).point;
    point.setX(x);
    point.setY(y);
    touchPointList().at(index).state = WebCore::PlatformTouchPoint::TouchMoved;

    return JSValueMakeUndefined(context);
}

static JSValueRef touchMoveCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    sendTouchEvent();
    return JSValueMakeUndefined(context);
}

static JSValueRef cancelTouchPointCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    if (argumentCount != 1)
        return JSValueMakeUndefined(context);

    int index = static_cast<int>(JSValueToNumber(context, arguments[0], exception));
    ASSERT(!exception || !*exception);
    if (index < 0 || index >= static_cast<int>(touchPointList().size()))
        return JSValueMakeUndefined(context);

    touchPointList().at(index).state = WebCore::PlatformTouchPoint::TouchCancelled;
    return JSValueMakeUndefined(context);
}

static JSValueRef touchCancelCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    sendTouchEvent();
    return JSValueMakeUndefined(context);
}

static JSValueRef releaseTouchPointCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    if (argumentCount != 1)
        return JSValueMakeUndefined(context);

    int index = static_cast<int>(JSValueToNumber(context, arguments[0], exception));
    ASSERT(!exception || !*exception);
    if (index < 0 || index >= static_cast<int>(touchPointList().size()))
        return JSValueMakeUndefined(context);

    touchPointList().at(index).state = WebCore::PlatformTouchPoint::TouchReleased;
    return JSValueMakeUndefined(context);
}

static JSValueRef touchEndCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    sendTouchEvent();
    touchModifiers = 0;
    return JSValueMakeUndefined(context);
}

static JSValueRef clearTouchPointsCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    touchPointList().clear();
    return JSValueMakeUndefined(context);
}

static JSValueRef sendRawTouchEvent(ClutterEventType type, JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);

    g_return_val_if_fail(page, JSValueMakeUndefined(context));
    g_return_val_if_fail(argumentCount >= 2, JSValueMakeUndefined(context));

    int x = static_cast<int>(JSValueToNumber(context, arguments[0], exception));
    g_return_val_if_fail((!exception || !*exception), JSValueMakeUndefined(context));
    int y = static_cast<int>(JSValueToNumber(context, arguments[1], exception));
    g_return_val_if_fail((!exception || !*exception), JSValueMakeUndefined(context));

    ClutterActor* actor = CLUTTER_ACTOR(webkit_web_page_get_view(page));
    const unsigned sequence = touchPointList().isEmpty() ? 0 : touchPointList().last().id + 1;

    ClutterTouchEvent *event = (ClutterTouchEvent*)clutter_event_new(type);
    event->time = g_get_monotonic_time() / 1000;
    event->flags = CLUTTER_EVENT_NONE;
    event->stage = CLUTTER_STAGE(clutter_actor_get_stage(actor));
    event->source = actor;
    event->x = x;
    event->y = y;
    event->sequence = (ClutterEventSequence*)GUINT_TO_POINTER(sequence);
    event->modifier_state = (ClutterModifierType)getStateFlags();
    event->device = clutter_device_manager_get_core_device(clutter_device_manager_get_default(), CLUTTER_POINTER_DEVICE);

    sendOrQueueEvent((ClutterEvent*)event, false);

    return JSValueMakeUndefined(context);
}

static JSValueRef gestureTapDownCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    return sendRawTouchEvent(CLUTTER_TOUCH_BEGIN, context, function, thisObject, argumentCount, arguments, exception);
}

static JSValueRef gestureTapCancelCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    return sendRawTouchEvent(CLUTTER_TOUCH_CANCEL, context, function, thisObject, argumentCount, arguments, exception);
}

static JSValueRef gestureTapCallback(JSContextRef context, JSObjectRef function, JSObjectRef thisObject, size_t argumentCount, const JSValueRef arguments[], JSValueRef* exception)
{
    gestureTapDownCallback(context, function, thisObject, argumentCount, arguments, exception);
    return sendRawTouchEvent(CLUTTER_TOUCH_END, context, function, thisObject, argumentCount, arguments, exception);
}

static JSStaticFunction staticFunctions[] = {
    { "mouseScrollBy", mouseScrollByCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "continuousMouseScrollBy", continuousMouseScrollByCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "contextClick", contextClickCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "mouseDown", mouseDownCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "mouseUp", mouseUpCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "mouseMoveTo", mouseMoveToCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "beginDragWithFiles", beginDragWithFilesCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "leapForward", leapForwardCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "keyDown", keyDownCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "textZoomIn", textZoomInCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "textZoomOut", textZoomOutCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "zoomPageIn", zoomPageInCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "zoomPageOut", zoomPageOutCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "addTouchPoint", addTouchPointCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "touchStart", touchStartCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "updateTouchPoint", updateTouchPointCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "touchMove", touchMoveCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "releaseTouchPoint", releaseTouchPointCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "touchEnd", touchEndCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "cancelTouchPoint", cancelTouchPointCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "touchCancel", touchCancelCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "clearTouchPoints", clearTouchPointsCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "gestureTapDown", gestureTapDownCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "gestureTapCancel", gestureTapCancelCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { "gestureTap", gestureTapCallback, kJSPropertyAttributeReadOnly | kJSPropertyAttributeDontDelete },
    { 0, 0, 0 }
};

static JSStaticValue staticValues[] = {
    { "dragMode", getDragModeCallback, setDragModeCallback, kJSPropertyAttributeNone },
    { 0, 0, 0, 0 }
};

static JSClassRef getClass(JSContextRef context)
{
    static JSClassRef eventSenderClass = 0;

    if (!eventSenderClass) {
        JSClassDefinition classDefinition = {
                0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
        classDefinition.staticFunctions = staticFunctions;
        classDefinition.staticValues = staticValues;

        eventSenderClass = JSClassCreate(&classDefinition);
    }

    return eventSenderClass;
}

JSObjectRef makeEventSender(JSContextRef context, bool isTopFrame)
{
    if (isTopFrame) {
        dragMode = true;

        // Fly forward in time one second when the main frame loads. This will
        // ensure that when a test begins clicking in the same location as
        // a previous test, those clicks won't be interpreted as continuations
        // of the previous test's click sequences.
        timeOffset += 1000;

        lastMousePositionX = lastMousePositionY = 0;
        lastClickPositionX = lastClickPositionY = 0;
        lastClickTimeOffset = 0;
        lastClickButton = 0;
        buttonCurrentlyDown = 0;
        clickCount = 0;
    }

    return JSObjectMake(context, getClass(context), 0);
}
