/*
 * Copyright (C) 2007 Apple Inc. All rights reserved.
 * Copyright (C) 2007 Eric Seidel <eric@webkit.org>
 * Copyright (C) 2008 Nuanti Ltd.
 * Copyright (C) 2009 Jan Michael Alonzo <jmalonzo@gmail.com>
 * Copyright (C) 2009, 2010 Collabora Ltd.
 * Copyright (C) 2010 Joone Hur <joone@kldp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "TestRunner.h"

#include "DumpRenderTree.h"
#include "WebCoreSupport/DumpRenderTreeSupportClutter.h"
#include "WorkQueue.h"
#include "WorkQueueItem.h"
#include "webkit/webkitclutter.h"
#include "webkit/webkitglobals.h"
#include "webkit/webkitiwebview.h"
#include "webkit/webkitwebbackforwardlist.h"
#include "webkit/webkitwebdatabase.h"
#include "webkit/webkitgwebframe.h"
#include "webkit/webkitwebinspector.h"
#include "webkit/webkitwebpage.h"
#include "webkit/webkitsecurityorigin.h"
#include <JavaScriptCore/JSRetainPtr.h>
#include <JavaScriptCore/JSStringRef.h>
#include <cstring>
#include <iostream>
#include <sstream>
#include <stdio.h>
#include <glib.h>
#include <libsoup/soup.h>
#include <wtf/gobject/GOwnPtr.h>

extern WebKitWebFrame* mainFrame;
extern guint waitToDumpWatchdog;
extern bool waitForPolicy;
extern WebKitWebFrame* topLoadingFrame;

extern "C" {
void webkit_application_cache_set_maximum_size(unsigned long long size);
void webkit_web_inspector_execute_script(WebKitWebInspector* inspector, long callId, const gchar* script);
}

gchar* JSStringCopyUTF8CString(JSStringRef jsString);

TestRunner::~TestRunner()
{
    // FIXME: implement
}

void TestRunner::addDisallowedURL(JSStringRef url)
{
    // FIXME: implement
}

void TestRunner::clearBackForwardList()
{
    WebKitWebPage* webPage = webkit_web_frame_get_page(mainFrame);
    WebKitWebBackForwardList* list = webkit_web_page_get_back_forward_list(webPage);
    WebKitWebHistoryItem* item = webkit_web_back_forward_list_get_current_item(list);
    g_object_ref(item);

    // We clear the history by setting the back/forward list's capacity to 0
    // then restoring it back and adding back the current item.
    gint limit = webkit_web_back_forward_list_get_limit(list);
    webkit_web_back_forward_list_set_limit(list, 0);
    webkit_web_back_forward_list_set_limit(list, limit);
    webkit_web_back_forward_list_add_item(list, item);
    webkit_web_back_forward_list_go_to_item(list, item);
    g_object_unref(item);
}

JSStringRef TestRunner::copyDecodedHostName(JSStringRef name)
{
    // FIXME: implement
    return 0;
}

JSStringRef TestRunner::copyEncodedHostName(JSStringRef name)
{
    // FIXME: implement
    return 0;
}

void TestRunner::dispatchPendingLoadRequests()
{
    // FIXME: Implement for testing fix for 6727495
}

void TestRunner::display()
{
    displayWebView();
}

void TestRunner::keepWebHistory()
{
    // FIXME: implement
}

JSValueRef TestRunner::computedStyleIncludingVisitedInfo(JSContextRef context, JSValueRef value)
{
    return DumpRenderTreeSupportClutter::computedStyleIncludingVisitedInfo(context, value);
}

size_t TestRunner::webHistoryItemCount()
{
    WebKitWebPage* webPage = webkit_web_frame_get_page(mainFrame);
    WebKitWebBackForwardList* list = webkit_web_page_get_back_forward_list(webPage);

    if (!list)
        return -1;

    // We do not add the current page to the total count as it's not
    // considered in DRT tests
    return webkit_web_back_forward_list_get_back_length(list) +
            webkit_web_back_forward_list_get_forward_length(list);
}

void TestRunner::notifyDone()
{
    if (m_waitToDump && !topLoadingFrame && !WorkQueue::shared()->count())
        dump();
    m_waitToDump = false;
    waitForPolicy = false;
}

JSStringRef TestRunner::pathToLocalResource(JSContextRef context, JSStringRef url)
{
    // Function introduced in r28690. This may need special-casing on Windows.
    return JSStringRetain(url); // Do nothing on Unix.
}

void TestRunner::queueLoad(JSStringRef url, JSStringRef target)
{
    gchar* relativeURL = JSStringCopyUTF8CString(url);
    SoupURI* baseURI = soup_uri_new(webkit_web_frame_get_uri(mainFrame));

    SoupURI* absoluteURI = soup_uri_new_with_base(baseURI, relativeURL);
    soup_uri_free(baseURI);
    g_free(relativeURL);

    gchar* absoluteCString;
    if (absoluteURI) {
        absoluteCString = soup_uri_to_string(absoluteURI, FALSE);
        soup_uri_free(absoluteURI);
    } else
        absoluteCString = JSStringCopyUTF8CString(url);

    JSRetainPtr<JSStringRef> absoluteURL(Adopt, JSStringCreateWithUTF8CString(absoluteCString));
    g_free(absoluteCString);

    WorkQueue::shared()->queue(new LoadItem(absoluteURL.get(), target));
}

void TestRunner::setAcceptsEditing(bool acceptsEditing)
{
    WebKitWebPage* webPage = webkit_web_frame_get_page(mainFrame);
    webkit_web_page_set_editable(webPage, acceptsEditing);
}

void TestRunner::setAlwaysAcceptCookies(bool alwaysAcceptCookies)
{
#ifdef HAVE_LIBSOUP_2_29_90
    SoupSession* session = webkit_get_default_session();
    SoupCookieJar* jar = reinterpret_cast<SoupCookieJar*>(soup_session_get_feature(session, SOUP_TYPE_COOKIE_JAR));

    /* If the jar was not created - we create it on demand, i.e, just
       in case we have HTTP requests - then we must create it here in
       order to set the proper accept policy */
    if (!jar) {
        jar = soup_cookie_jar_new();
        soup_session_add_feature(session, SOUP_SESSION_FEATURE(jar));
        g_object_unref(jar);
    }

    SoupCookieJarAcceptPolicy policy;

    if (alwaysAcceptCookies)
        policy = SOUP_COOKIE_JAR_ACCEPT_ALWAYS;
    else
        policy = SOUP_COOKIE_JAR_ACCEPT_NO_THIRD_PARTY;

    g_object_set(G_OBJECT(jar), SOUP_COOKIE_JAR_ACCEPT_POLICY, policy, NULL);
#endif
}

void TestRunner::setCustomPolicyDelegate(bool setDelegate, bool permissive)
{
    // FIXME: implement
}

void TestRunner::waitForPolicyDelegate()
{
    waitForPolicy = true;
    setWaitToDump(true);
}

void TestRunner::setScrollbarPolicy(JSStringRef orientation, JSStringRef policy)
{
    // FIXME: implement
}

void TestRunner::addOriginAccessWhitelistEntry(JSStringRef sourceOrigin, JSStringRef protocol, JSStringRef host, bool includeSubdomains)
{
    // gchar* sourceOriginGChar = JSStringCopyUTF8CString(sourceOrigin);
    // gchar* protocolGChar = JSStringCopyUTF8CString(protocol);
    // gchar* hostGChar = JSStringCopyUTF8CString(host);
    // DumpRenderTreeSupportClutter::whiteListAccessFromOrigin(sourceOriginGChar, protocolGChar, hostGChar, includeSubdomains);
    // g_free(sourceOriginGChar);
    // g_free(protocolGChar);
    // g_free(hostGChar);
}

void TestRunner::removeOriginAccessWhitelistEntry(JSStringRef sourceOrigin, JSStringRef protocol, JSStringRef host, bool includeSubdomains)
{
    // FIXME: implement
}

void TestRunner::setMainFrameIsFirstResponder(bool flag)
{
    // FIXME: implement
}

void TestRunner::setTabKeyCyclesThroughElements(bool cycles)
{
    WebKitWebPage* webPage = webkit_web_frame_get_page(mainFrame);
    WebKitWebSettings* settings = webkit_web_page_get_settings(webPage);
    g_object_set(G_OBJECT(settings), "tab-key-cycles-through-elements", cycles, NULL);
}

void TestRunner::setUseDashboardCompatibilityMode(bool flag)
{
    // FIXME: implement
}

static gchar* userStyleSheet = NULL;
static gboolean userStyleSheetEnabled = TRUE;

void TestRunner::setUserStyleSheetEnabled(bool flag)
{
    userStyleSheetEnabled = flag;

    WebKitWebPage* webPage = webkit_web_frame_get_page(mainFrame);
    WebKitWebSettings* settings = webkit_web_page_get_settings(webPage);
    if (flag && userStyleSheet)
        g_object_set(G_OBJECT(settings), "user-stylesheet-uri", userStyleSheet, NULL);
    else
        g_object_set(G_OBJECT(settings), "user-stylesheet-uri", "", NULL);
}

void TestRunner::setUserStyleSheetLocation(JSStringRef path)
{
    g_free(userStyleSheet);
    userStyleSheet = JSStringCopyUTF8CString(path);
    if (userStyleSheetEnabled)
        setUserStyleSheetEnabled(true);
}

void TestRunner::setValueForUser(JSContextRef context, JSValueRef nodeObject, JSStringRef value)
{
    // FIXME: implement
}

void TestRunner::setViewModeMediaFeature(JSStringRef mode)
{
    // WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    // ASSERT(page);

    // char* viewMode = JSStringCopyUTF8CString(mode);

    // if (!g_strcmp0(viewMode, "windowed"))
    //     webkit_web_page_set_view_mode(page, WEBKIT_WEB_PAGE_VIEW_MODE_WINDOWED);
    // else if (!g_strcmp0(viewMode, "floating"))
    //     webkit_web_page_set_view_mode(page, WEBKIT_WEB_PAGE_VIEW_MODE_FLOATING);
    // else if (!g_strcmp0(viewMode, "fullscreen"))
    //     webkit_web_page_set_view_mode(page, WEBKIT_WEB_PAGE_VIEW_MODE_FULLSCREEN);
    // else if (!g_strcmp0(viewMode, "maximized"))
    //     webkit_web_page_set_view_mode(page, WEBKIT_WEB_PAGE_VIEW_MODE_MAXIMIZED);
    // else if (!g_strcmp0(viewMode, "minimized"))
    //     webkit_web_page_set_view_mode(page, WEBKIT_WEB_PAGE_VIEW_MODE_MINIMIZED);

    // g_free(viewMode);
}

void TestRunner::setWindowIsKey(bool windowIsKey)
{
    // FIXME: implement
}

void TestRunner::setSmartInsertDeleteEnabled(bool flag)
{
    // FIXME: implement
}

static gboolean waitToDumpWatchdogFired(void*)
{
    waitToDumpWatchdog = 0;
    gTestRunner->waitToDumpWatchdogTimerFired();
    return FALSE;
}

void TestRunner::setWaitToDump(bool waitUntilDone)
{
    static const int timeoutSeconds = 30;

    m_waitToDump = waitUntilDone;
    if (m_waitToDump && !waitToDumpWatchdog)
        waitToDumpWatchdog = g_timeout_add_seconds(timeoutSeconds, waitToDumpWatchdogFired, 0);
}

int TestRunner::windowCount()
{
    // +1 -> including the main view
    //return g_slist_length(webViewList) + 1;
    return 1;
}

void TestRunner::setPrivateBrowsingEnabled(bool flag)
{
    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    ASSERT(page);

    WebKitWebSettings* settings = webkit_web_page_get_settings(page);
    g_object_set(G_OBJECT(settings), "enable-private-browsing", flag, NULL);
}

void TestRunner::setJavaScriptCanAccessClipboard(bool flag)
{
    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    ASSERT(page);

    WebKitWebSettings* settings = webkit_web_page_get_settings(page);
    g_object_set(G_OBJECT(settings), "javascript-can-access-clipboard", flag, NULL);
}

void TestRunner::setXSSAuditorEnabled(bool flag)
{
    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    ASSERT(page);

    WebKitWebSettings* settings = webkit_web_page_get_settings(page);
    g_object_set(G_OBJECT(settings), "enable-xss-auditor", flag, NULL);
}

void TestRunner::setSpatialNavigationEnabled(bool flag)
{
    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    ASSERT(page);

    WebKitWebSettings* settings = webkit_web_page_get_settings(page);
    g_object_set(G_OBJECT(settings), "enable-spatial-navigation", flag, NULL);
}

void TestRunner::setAllowUniversalAccessFromFileURLs(bool flag)
{
    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    ASSERT(page);

    WebKitWebSettings* settings = webkit_web_page_get_settings(page);
    g_object_set(G_OBJECT(settings), "enable-universal-access-from-file-uris", flag, NULL);
}

void TestRunner::setAllowFileAccessFromFileURLs(bool flag)
{
    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    ASSERT(page);

    WebKitWebSettings* settings = webkit_web_page_get_settings(page);
    g_object_set(G_OBJECT(settings), "enable-file-access-from-file-uris", flag, NULL);
}

void TestRunner::setAuthorAndUserStylesEnabled(bool flag)
{
    // FIXME: implement
}

void TestRunner::setAutofilled(JSContextRef context, JSValueRef nodeObject, bool isAutofilled)
{
    // FIXME: implement
}

void TestRunner::setMockDeviceOrientation(bool canProvideAlpha, double alpha, bool canProvideBeta, double beta, bool canProvideGamma, double gamma)
{
    // FIXME: Implement for DeviceOrientation layout tests.
    // See https://bugs.webkit.org/show_bug.cgi?id=30335.
}

void TestRunner::setMockGeolocationPosition(double latitude, double longitude, double accuracy, bool, double, bool, double, bool, double, bool, double)
{
    WebKitWebPage* webPage = webkit_web_frame_get_page(mainFrame);
    ASSERT(webPage);

    DumpRenderTreeSupportClutter::setMockGeolocationPosition(webPage, latitude, longitude, accuracy);
}

void TestRunner::setMockGeolocationPositionUnavailableError(JSStringRef message)
{
    WebKitWebPage* webPage = webkit_web_frame_get_page(mainFrame);
    ASSERT(webPage);

    GOwnPtr<gchar> cMessage(JSStringCopyUTF8CString(message));
    DumpRenderTreeSupportClutter::setMockGeolocationPositionUnavailableError(webPage, cMessage.get());
}

void TestRunner::setGeolocationPermission(bool allow)
{
    setGeolocationPermissionCommon(allow);

    WebKitWebPage* webPage = webkit_web_frame_get_page(mainFrame);
    ASSERT(webPage);

    DumpRenderTreeSupportClutter::setMockGeolocationPermission(webPage, allow);
}

int TestRunner::numberOfPendingGeolocationPermissionRequests()
{
    WebKitWebPage* webPage = webkit_web_frame_get_page(mainFrame);
    ASSERT(webPage);

    return DumpRenderTreeSupportClutter::numberOfPendingGeolocationPermissionRequests(webPage);
}

void TestRunner::addMockSpeechInputResult(JSStringRef result, double confidence, JSStringRef language)
{
    // FIXME: Implement for speech input layout tests.
    // See https://bugs.webkit.org/show_bug.cgi?id=39485.
}

void TestRunner::setMockSpeechInputDumpRect(bool flag)
{
    // FIXME: Implement for speech input layout tests.
    // See https://bugs.webkit.org/show_bug.cgi?id=39485.
}

void TestRunner::startSpeechInput(JSContextRef inputElement)
{
    // FIXME: Implement for speech input layout tests.
    // See https://bugs.webkit.org/show_bug.cgi?id=39485.
}

void TestRunner::setIconDatabaseEnabled(bool flag)
{
    // FIXME: implement
}

void TestRunner::setSelectTrailingWhitespaceEnabled(bool flag)
{
    // FIXME: implement
}

void TestRunner::setPopupBlockingEnabled(bool flag)
{
    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    ASSERT(page);

    WebKitWebSettings* settings = webkit_web_page_get_settings(page);
    g_object_set(G_OBJECT(settings), "javascript-can-open-windows-automatically", !flag, NULL);

}

void TestRunner::setPluginsEnabled(bool flag)
{
    // FIXME: Implement
}

bool TestRunner::elementDoesAutoCompleteForElementWithId(JSStringRef id) 
{
    // FIXME: Use GOwnPtr
    gchar* name = JSStringCopyUTF8CString(id);
    bool returnValue = DumpRenderTreeSupportClutter::elementDoesAutoCompleteForElementWithId(mainFrame, name);
    g_free(name);
    return returnValue;
}

void TestRunner::execCommand(JSStringRef name, JSStringRef value)
{
    // WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    // ASSERT(page);

    // gchar* cName = JSStringCopyUTF8CString(name);
    // gchar* cValue = JSStringCopyUTF8CString(value);
    // DumpRenderTreeSupportClutter::executeCoreCommandByName(page, cName, cValue);
    // g_free(cName);
    // g_free(cValue);
}

bool TestRunner::findString(JSContextRef context, JSStringRef target, JSObjectRef optionsArray)
{
    JSRetainPtr<JSStringRef> lengthPropertyName(Adopt, JSStringCreateWithUTF8CString("length"));
    JSValueRef lengthValue = JSObjectGetProperty(context, optionsArray, lengthPropertyName.get(), 0);
    if (!JSValueIsNumber(context, lengthValue))
        return false;

    GList* optionsList = 0;

    size_t length = static_cast<size_t>(JSValueToNumber(context, lengthValue, 0));
    for (size_t i = 0; i < length; ++i) {
        JSValueRef value = JSObjectGetPropertyAtIndex(context, optionsArray, i, 0);
        if (!JSValueIsString(context, value))
            continue;

        JSRetainPtr<JSStringRef> option(Adopt, JSValueToStringCopy(context, value, 0));
        gchar* optionName = JSStringCopyUTF8CString(option.get());
        optionsList = g_list_append(optionsList, optionName);
    }

    // FIXME: Use OwnPtr
    gchar* targetString = JSStringCopyUTF8CString(target);
    bool returnValue = DumpRenderTreeSupportClutter::findString(mainFrame, targetString, optionsList);
    g_free(targetString);
    g_list_free_full(optionsList, g_free);

    return returnValue;
}

bool TestRunner::isCommandEnabled(JSStringRef name)
{
    // WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    // ASSERT(page);

    // gchar* cName = JSStringCopyUTF8CString(name);
    // bool result = DumpRenderTreeSupportClutter::isCommandEnabled(page, cName);
    // g_free(cName);
    // return result;
    return true;
}

void TestRunner::setCacheModel(int cacheModel)
{
    if (!cacheModel) // WebCacheModelDocumentViewer
        webkit_set_cache_model(WEBKIT_CACHE_MODEL_DOCUMENT_VIEWER); 
    else 
        webkit_set_cache_model(WEBKIT_CACHE_MODEL_WEB_BROWSER); 
}

void TestRunner::setPersistentUserStyleSheetLocation(JSStringRef jsURL)
{
    // FIXME: implement
}

void TestRunner::clearPersistentUserStyleSheet()
{
    // FIXME: implement
}

void TestRunner::clearAllApplicationCaches()
{
    // FIXME: implement to support Application Cache quotas.
}

void TestRunner::setApplicationCacheOriginQuota(unsigned long long quota)
{
    // FIXME: implement to support Application Cache quotas.
}

void TestRunner::clearApplicationCacheForOrigin(OpaqueJSString*)
{
    // FIXME: Implement to support deleting all application caches for an origin.
}

long long TestRunner::localStorageDiskUsageForOrigin(JSStringRef originIdentifier)
{
    // FIXME: Implement to support getting disk usage in bytes for an origin.
    return 0;
}

JSValueRef TestRunner::originsWithApplicationCache(JSContextRef context)
{
    // FIXME: Implement to get origins that contain application caches.
    return JSValueMakeUndefined(context);
}

long long TestRunner::applicationCacheDiskUsageForOrigin(JSStringRef name)
{
    // FIXME: implement
    return 0;
}

void TestRunner::clearAllDatabases()
{
    webkit_remove_all_web_databases();
}
 
void TestRunner::setDatabaseQuota(unsigned long long quota)
{
    WebKitSecurityOrigin* origin = webkit_web_frame_get_security_origin(mainFrame);
    webkit_security_origin_set_web_database_quota(origin, quota);
}

JSValueRef TestRunner::originsWithLocalStorage(JSContextRef context)
{
    // FIXME: implement
    return JSValueMakeUndefined(context);
}

void TestRunner::deleteAllLocalStorage()
{
        // FIXME: implement
}

void TestRunner::deleteLocalStorageForOrigin(JSStringRef originIdentifier)
{
        // FIXME: implement
}

void TestRunner::observeStorageTrackerNotifications(unsigned number)
{
        // FIXME: implement
}

void TestRunner::syncLocalStorage()
{
    // FIXME: implement
}

void TestRunner::setDomainRelaxationForbiddenForURLScheme(bool, JSStringRef)
{
    // FIXME: implement
}

void TestRunner::goBack()
{
    // FIXME: implement to enable loader/navigation-while-deferring-loads.html
}

void TestRunner::setDefersLoading(bool)
{
    // FIXME: implement to enable loader/navigation-while-deferring-loads.html
}

void TestRunner::setAppCacheMaximumSize(unsigned long long size)
{
    webkit_application_cache_set_maximum_size(size);
}

void TestRunner::overridePreference(JSStringRef key, JSStringRef value)
{
    GOwnPtr<gchar> originalName(JSStringCopyUTF8CString(key));
    GOwnPtr<gchar> valueAsString(JSStringCopyUTF8CString(value));

    WebKitWebPage* page = webkit_web_frame_get_page(mainFrame);
    ASSERT(page);

    // This transformation could be handled by a hash table (and it once was), but
    // having it prominent, makes it easier for people from other ports to keep the
    // list up to date.
    const gchar* propertyName = 0;
    if (g_str_equal(originalName.get(), "WebKitJavaScriptEnabled"))
        propertyName = "enable-scripts";
    else if (g_str_equal(originalName.get(), "WebKitDefaultFontSize"))
        propertyName = "default-font-size";
    else if (g_str_equal(originalName.get(), "WebKitEnableCaretBrowsing"))
        propertyName = "enable-caret-browsing";
    else if (g_str_equal(originalName.get(), "WebKitUsesPageCachePreferenceKey"))
        propertyName = "enable-page-cache";
    else if (g_str_equal(originalName.get(), "WebKitPluginsEnabled"))
        propertyName = "enable-plugins";
    else if (g_str_equal(originalName.get(), "WebKitHyperlinkAuditingEnabled"))
        propertyName = "enable-hyperlink-auditing";
    else if (g_str_equal(originalName.get(), "WebKitWebGLEnabled"))
        propertyName = "enable-webgl";
    else if (g_str_equal(originalName.get(), "WebKitTabToLinksPreferenceKey")) {
        DumpRenderTreeSupportClutter::setLinksIncludedInFocusChain(!g_ascii_strcasecmp(valueAsString.get(), "true")
                                                                   || !g_ascii_strcasecmp(valueAsString.get(), "1"));
        return;
    } else {
        fprintf(stderr, "TestRunner::overridePreference tried to override "
                "unknown preference '%s'.\n", originalName.get());
        return;
    }

    WebKitWebSettings* settings = webkit_web_page_get_settings(page);
    GParamSpec* pspec = g_object_class_find_property(G_OBJECT_CLASS(
        WEBKIT_WEB_SETTINGS_GET_CLASS(settings)), propertyName);
    GValue currentPropertyValue = { 0, { { 0 } } };
    g_value_init(&currentPropertyValue, pspec->value_type);

    if (G_VALUE_HOLDS_STRING(&currentPropertyValue))
        g_object_set(settings, propertyName, valueAsString.get(), NULL);
    else if (G_VALUE_HOLDS_BOOLEAN(&currentPropertyValue))
        g_object_set(G_OBJECT(settings), propertyName, !g_ascii_strcasecmp(valueAsString.get(), "true")
                        || !g_ascii_strcasecmp(valueAsString.get(), "1"), NULL);
    else if (G_VALUE_HOLDS_INT(&currentPropertyValue))
        g_object_set(G_OBJECT(settings), propertyName, atoi(valueAsString.get()), NULL);
    else if (G_VALUE_HOLDS_FLOAT(&currentPropertyValue)) {
        gfloat newValue = g_ascii_strtod(valueAsString.get(), 0);
        g_object_set(G_OBJECT(settings), propertyName, newValue, NULL);
    } else
        fprintf(stderr, "TestRunner::overridePreference failed to override "
                "preference '%s'.\n", originalName.get());
}

void TestRunner::addUserScript(JSStringRef source, bool runAtStart, bool allFrames)
{
    gchar* sourceStr = JSStringCopyUTF8CString(source);
    DumpRenderTreeSupportClutter::addUserScript(mainFrame, sourceStr, runAtStart, allFrames);
    g_free(sourceStr);
}

void TestRunner::addUserStyleSheet(JSStringRef source, bool allFrames)
{
    gchar* sourceStr = JSStringCopyUTF8CString(source);
    DumpRenderTreeSupportClutter::addUserStyleSheet(mainFrame, sourceStr, allFrames);
    g_free(sourceStr);
}

void TestRunner::setDeveloperExtrasEnabled(bool enabled)
{
    WebKitWebPage* webPage = webkit_web_frame_get_page(mainFrame);
    WebKitWebSettings* webSettings = webkit_web_page_get_settings(webPage);

    g_object_set(webSettings, "enable-developer-extras", enabled, NULL);
}

void TestRunner::setAsynchronousSpellCheckingEnabled(bool)
{
    // FIXME: Implement this.
}

void TestRunner::showWebInspector()
{
    WebKitWebPage* webPage = webkit_web_frame_get_page(mainFrame);
    WebKitWebInspector* inspector = webkit_web_page_get_inspector(webPage);

    webkit_web_inspector_show(inspector);
}

void TestRunner::closeWebInspector()
{
    WebKitWebPage* webPage = webkit_web_frame_get_page(mainFrame);
    WebKitWebInspector* inspector = webkit_web_page_get_inspector(webPage);

    webkit_web_inspector_close(inspector);
}

void TestRunner::evaluateInWebInspector(long callId, JSStringRef script)
{
    WebKitWebPage* webPage = webkit_web_frame_get_page(mainFrame);
    WebKitWebInspector* inspector = webkit_web_page_get_inspector(webPage);
    char* scriptString = JSStringCopyUTF8CString(script);

    webkit_web_inspector_execute_script(inspector, callId, scriptString);
    g_free(scriptString);
}

void TestRunner::evaluateScriptInIsolatedWorld(unsigned worldID, JSObjectRef globalObject, JSStringRef script)
{
    GOwnPtr<gchar> scriptString(JSStringCopyUTF8CString(script));
    DumpRenderTreeSupportClutter::evaluateScriptInIsolatedWorld(mainFrame, worldID, scriptString.get());
}

void TestRunner::evaluateScriptInIsolatedWorldAndReturnValue(unsigned worldID, JSObjectRef globalObject, JSStringRef script)
{
    // FIXME: Implement this.
}

void TestRunner::removeAllVisitedLinks()
{
    // FIXME: Implement this.
}

bool TestRunner::callShouldCloseOnWebView()
{
    // FIXME: Implement for testing fix for https://bugs.webkit.org/show_bug.cgi?id=27481
    return false;
}

void TestRunner::apiTestNewWindowDataLoadBaseURL(JSStringRef utf8Data, JSStringRef baseURL)
{

}

void TestRunner::apiTestGoToCurrentBackForwardItem()
{

}

void TestRunner::setWebViewEditable(bool)
{
}

JSRetainPtr<JSStringRef> TestRunner::markerTextForListItem(JSContextRef context, JSValueRef nodeObject) const
{
    CString markerTextGChar = DumpRenderTreeSupportClutter::markerTextForListItem(mainFrame, context, nodeObject);
    if (markerTextGChar.isNull())
        return 0;

    JSRetainPtr<JSStringRef> markerText(Adopt, JSStringCreateWithUTF8CString(markerTextGChar.data()));
    return markerText;
}

void TestRunner::authenticateSession(JSStringRef, JSStringRef, JSStringRef)
{
}

void TestRunner::abortModal()
{
}

void TestRunner::setSerializeHTTPLoads(bool)
{
    // FIXME: Implement if needed for https://bugs.webkit.org/show_bug.cgi?id=50758.
}

void TestRunner::setTextDirection(JSStringRef direction)
{
    // FIXME: Implement.
}

void TestRunner::addChromeInputField()
{
}

void TestRunner::removeChromeInputField()
{
}

void TestRunner::focusWebView()
{
}

void TestRunner::setBackingScaleFactor(double)
{
}

void TestRunner::grantWebNotificationPermission(JSStringRef origin)
{
}

void TestRunner::denyWebNotificationPermission(JSStringRef jsOrigin)
{
}

void TestRunner::removeAllWebNotificationPermissions()
{
}

void TestRunner::simulateWebNotificationClick(JSValueRef jsNotification)
{
}

void TestRunner::simulateLegacyWebNotificationClick(JSStringRef title)
{
}

void TestRunner::resetPageVisibility()
{
    // FIXME: Implement this.
}

void TestRunner::setPageVisibility(const char*)
{
    // FIXME: Implement this.
}

void TestRunner::setAutomaticLinkDetectionEnabled(bool)
{
    // FIXME: Implement this.
}

void TestRunner::setStorageDatabaseIdleInterval(double)
{
    // FIXME: Implement this.
}

void TestRunner::closeIdleLocalStorageDatabases()
{
}
